/*
Copyright (C) 2018-2019 de4dot@gmail.com

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

using System;
using System.Collections.Generic;
using System.Linq;
using Generator.Enums;
using Generator.Enums.Encoder;

namespace Generator.Encoder {
	static class OpCodeInfoTable {
		public static readonly OpCodeInfo[] Data = GetData();

		static OpCodeInfo[] GetData() {
			var data = new OpCodeInfo[] {
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.INVALID)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x00, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.NoInstruction, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.DeclareByte)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x00, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.NoInstruction, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.DeclareWord)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x00, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.NoInstruction, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.DeclareDword)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x00, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.NoInstruction, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.DeclareQword)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x00, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.NoInstruction, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Add_rm8_r8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x00, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eb, LegacyOpKind.Gb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Add_rm16_r16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x01, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Add_rm32_r32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x01, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Add_rm64_r64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x01, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq, LegacyOpKind.Gq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Add_r8_rm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x02, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gb, LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Add_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x03, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Add_r32_rm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x03, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Add_r64_rm64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x03, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Add_AL_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x04, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.AL, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Add_AX_imm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x05, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.AX, LegacyOpKind.Iw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Add_EAX_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x05, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.EAX, LegacyOpKind.Id }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Add_RAX_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x05, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.RAX, LegacyOpKind.Id64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pushw_ES)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x06, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.ES }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pushd_ES)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x06, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.ES }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Popw_ES)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x07, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.ES }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Popd_ES)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x07, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.ES }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Or_rm8_r8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x08, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eb, LegacyOpKind.Gb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Or_rm16_r16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x09, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Or_rm32_r32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x09, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Or_rm64_r64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x09, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq, LegacyOpKind.Gq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Or_r8_rm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x0A, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gb, LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Or_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x0B, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Or_r32_rm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x0B, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Or_r64_rm64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x0B, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Or_AL_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x0C, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.AL, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Or_AX_imm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x0D, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.AX, LegacyOpKind.Iw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Or_EAX_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x0D, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.EAX, LegacyOpKind.Id }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Or_RAX_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x0D, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.RAX, LegacyOpKind.Id64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pushw_CS)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x0E, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.CS }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pushd_CS)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x0E, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.CS }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Popw_CS)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x0F, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.CS }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Adc_rm8_r8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x10, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eb, LegacyOpKind.Gb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Adc_rm16_r16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x11, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Adc_rm32_r32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x11, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Adc_rm64_r64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x11, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq, LegacyOpKind.Gq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Adc_r8_rm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x12, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gb, LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Adc_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x13, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Adc_r32_rm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x13, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Adc_r64_rm64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x13, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Adc_AL_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x14, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.AL, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Adc_AX_imm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x15, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.AX, LegacyOpKind.Iw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Adc_EAX_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x15, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.EAX, LegacyOpKind.Id }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Adc_RAX_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x15, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.RAX, LegacyOpKind.Id64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pushw_SS)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x16, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.SS }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pushd_SS)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x16, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.SS }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Popw_SS)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x17, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.SS }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Popd_SS)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x17, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.SS }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sbb_rm8_r8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x18, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eb, LegacyOpKind.Gb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sbb_rm16_r16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x19, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sbb_rm32_r32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x19, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sbb_rm64_r64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x19, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq, LegacyOpKind.Gq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sbb_r8_rm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x1A, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gb, LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sbb_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x1B, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sbb_r32_rm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x1B, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sbb_r64_rm64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x1B, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sbb_AL_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x1C, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.AL, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sbb_AX_imm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x1D, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.AX, LegacyOpKind.Iw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sbb_EAX_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x1D, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.EAX, LegacyOpKind.Id }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sbb_RAX_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x1D, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.RAX, LegacyOpKind.Id64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pushw_DS)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x1E, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.DS }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pushd_DS)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x1E, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.DS }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Popw_DS)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x1F, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.DS }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Popd_DS)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x1F, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.DS }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.And_rm8_r8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x20, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eb, LegacyOpKind.Gb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.And_rm16_r16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x21, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.And_rm32_r32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x21, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.And_rm64_r64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x21, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq, LegacyOpKind.Gq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.And_r8_rm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x22, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gb, LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.And_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x23, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.And_r32_rm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x23, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.And_r64_rm64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x23, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.And_AL_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x24, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.AL, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.And_AX_imm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x25, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.AX, LegacyOpKind.Iw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.And_EAX_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x25, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.EAX, LegacyOpKind.Id }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.And_RAX_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x25, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.RAX, LegacyOpKind.Id64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Daa)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x27, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sub_rm8_r8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x28, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eb, LegacyOpKind.Gb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sub_rm16_r16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x29, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sub_rm32_r32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x29, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sub_rm64_r64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x29, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq, LegacyOpKind.Gq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sub_r8_rm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x2A, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gb, LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sub_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x2B, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sub_r32_rm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x2B, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sub_r64_rm64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x2B, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sub_AL_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x2C, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.AL, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sub_AX_imm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x2D, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.AX, LegacyOpKind.Iw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sub_EAX_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x2D, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.EAX, LegacyOpKind.Id }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sub_RAX_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x2D, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.RAX, LegacyOpKind.Id64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Das)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x2F, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xor_rm8_r8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x30, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eb, LegacyOpKind.Gb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xor_rm16_r16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x31, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xor_rm32_r32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x31, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xor_rm64_r64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x31, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq, LegacyOpKind.Gq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xor_r8_rm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x32, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gb, LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xor_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x33, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xor_r32_rm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x33, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xor_r64_rm64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x33, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xor_AL_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x34, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.AL, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xor_AX_imm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x35, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.AX, LegacyOpKind.Iw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xor_EAX_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x35, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.EAX, LegacyOpKind.Id }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xor_RAX_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x35, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.RAX, LegacyOpKind.Id64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Aaa)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x37, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmp_rm8_r8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x38, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb, LegacyOpKind.Gb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmp_rm16_r16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x39, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmp_rm32_r32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x39, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmp_rm64_r64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x39, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Gq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmp_r8_rm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x3A, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gb, LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmp_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x3B, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmp_r32_rm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x3B, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmp_r64_rm64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x3B, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmp_AL_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x3C, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.AL, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmp_AX_imm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x3D, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.AX, LegacyOpKind.Iw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmp_EAX_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x3D, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.EAX, LegacyOpKind.Id }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmp_RAX_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x3D, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.RAX, LegacyOpKind.Id64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Aas)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x3F, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Inc_r16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x40, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.r16_rw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Inc_r32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x40, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.r32_rd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Dec_r16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x48, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.r16_rw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Dec_r32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x48, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.r32_rd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Push_r16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x50, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.r16_rw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Push_r32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x50, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.r32_rd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Push_r64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x50, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.r64_ro }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pop_r16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x58, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.r16_rw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pop_r32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x58, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.r32_rd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pop_r64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x58, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.r64_ro }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pushaw)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x60, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pushad)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x60, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Popaw)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x61, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Popad)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x61, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bound_r16_m1616)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x62, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Gw, LegacyOpKind.Mw2 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bound_r32_m3232)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x62, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Gd, LegacyOpKind.Md2 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Arpl_rm16_r16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x63, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Arpl_r32m16_r32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x63, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.RdMw, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movsxd_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x63, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movsxd_r32_rm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x63, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movsxd_r64_rm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x63, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Push_imm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x68, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Iw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pushd_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x68, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Id }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pushq_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x68, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Id64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Imul_r16_rm16_imm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x69, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew, LegacyOpKind.Iw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Imul_r32_rm32_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x69, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed, LegacyOpKind.Id }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Imul_r64_rm64_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x69, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq, LegacyOpKind.Id64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pushw_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x6A, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ib16 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pushd_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x6A, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Ib32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pushq_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x6A, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Ib64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Imul_r16_rm16_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x6B, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew, LegacyOpKind.Ib16 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Imul_r32_rm32_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x6B, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed, LegacyOpKind.Ib32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Imul_r64_rm64_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x6B, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq, LegacyOpKind.Ib64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Insb_m8_DX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x6C, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, new[] { LegacyOpKind.Yb, LegacyOpKind.DX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Insw_m16_DX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x6D, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, new[] { LegacyOpKind.Yw, LegacyOpKind.DX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Insd_m32_DX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x6D, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, new[] { LegacyOpKind.Yd, LegacyOpKind.DX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Outsb_DX_m8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x6E, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, new[] { LegacyOpKind.DX, LegacyOpKind.Xb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Outsw_DX_m16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x6F, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, new[] { LegacyOpKind.DX, LegacyOpKind.Xw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Outsd_DX_m32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x6F, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, new[] { LegacyOpKind.DX, LegacyOpKind.Xd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jo_rel8_16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x70, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.wJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jo_rel8_32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x70, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.dJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jo_rel8_64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x70, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.qJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jno_rel8_16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x71, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.wJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jno_rel8_32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x71, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.dJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jno_rel8_64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x71, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.qJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jb_rel8_16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x72, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.wJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jb_rel8_32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x72, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.dJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jb_rel8_64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x72, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.qJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jae_rel8_16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x73, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.wJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jae_rel8_32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x73, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.dJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jae_rel8_64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x73, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.qJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Je_rel8_16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x74, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.wJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Je_rel8_32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x74, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.dJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Je_rel8_64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x74, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.qJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jne_rel8_16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x75, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.wJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jne_rel8_32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x75, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.dJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jne_rel8_64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x75, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.qJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jbe_rel8_16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x76, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.wJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jbe_rel8_32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x76, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.dJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jbe_rel8_64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x76, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.qJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ja_rel8_16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x77, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.wJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ja_rel8_32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x77, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.dJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ja_rel8_64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x77, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.qJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Js_rel8_16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x78, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.wJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Js_rel8_32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x78, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.dJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Js_rel8_64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x78, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.qJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jns_rel8_16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x79, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.wJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jns_rel8_32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x79, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.dJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jns_rel8_64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x79, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.qJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jp_rel8_16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x7A, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.wJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jp_rel8_32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x7A, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.dJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jp_rel8_64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x7A, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.qJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jnp_rel8_16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x7B, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.wJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jnp_rel8_32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x7B, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.dJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jnp_rel8_64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x7B, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.qJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jl_rel8_16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x7C, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.wJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jl_rel8_32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x7C, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.dJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jl_rel8_64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x7C, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.qJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jge_rel8_16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x7D, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.wJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jge_rel8_32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x7D, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.dJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jge_rel8_64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x7D, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.qJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jle_rel8_16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x7E, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.wJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jle_rel8_32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x7E, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.dJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jle_rel8_64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x7E, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.qJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jg_rel8_16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x7F, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.wJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jg_rel8_32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x7F, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.dJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jg_rel8_64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x7F, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.qJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Add_rm8_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x80, 0, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eb, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Or_rm8_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x80, 1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eb, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Adc_rm8_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x80, 2, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eb, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sbb_rm8_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x80, 3, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eb, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.And_rm8_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x80, 4, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eb, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sub_rm8_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x80, 5, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eb, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xor_rm8_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x80, 6, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eb, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmp_rm8_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x80, 7, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Add_rm16_imm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x81, 0, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew, LegacyOpKind.Iw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Add_rm32_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x81, 0, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed, LegacyOpKind.Id }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Add_rm64_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x81, 0, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq, LegacyOpKind.Id64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Or_rm16_imm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x81, 1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew, LegacyOpKind.Iw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Or_rm32_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x81, 1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed, LegacyOpKind.Id }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Or_rm64_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x81, 1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq, LegacyOpKind.Id64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Adc_rm16_imm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x81, 2, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew, LegacyOpKind.Iw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Adc_rm32_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x81, 2, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed, LegacyOpKind.Id }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Adc_rm64_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x81, 2, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq, LegacyOpKind.Id64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sbb_rm16_imm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x81, 3, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew, LegacyOpKind.Iw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sbb_rm32_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x81, 3, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed, LegacyOpKind.Id }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sbb_rm64_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x81, 3, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq, LegacyOpKind.Id64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.And_rm16_imm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x81, 4, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew, LegacyOpKind.Iw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.And_rm32_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x81, 4, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed, LegacyOpKind.Id }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.And_rm64_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x81, 4, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq, LegacyOpKind.Id64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sub_rm16_imm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x81, 5, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew, LegacyOpKind.Iw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sub_rm32_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x81, 5, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed, LegacyOpKind.Id }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sub_rm64_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x81, 5, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq, LegacyOpKind.Id64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xor_rm16_imm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x81, 6, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew, LegacyOpKind.Iw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xor_rm32_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x81, 6, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed, LegacyOpKind.Id }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xor_rm64_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x81, 6, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq, LegacyOpKind.Id64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmp_rm16_imm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x81, 7, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Iw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmp_rm32_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x81, 7, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Id }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmp_rm64_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x81, 7, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Id64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Add_rm8_imm8_82)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x82, 0, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eb, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Or_rm8_imm8_82)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x82, 1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eb, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Adc_rm8_imm8_82)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x82, 2, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eb, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sbb_rm8_imm8_82)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x82, 3, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eb, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.And_rm8_imm8_82)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x82, 4, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eb, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sub_rm8_imm8_82)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x82, 5, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eb, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xor_rm8_imm8_82)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x82, 6, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eb, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmp_rm8_imm8_82)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x82, 7, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Eb, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Add_rm16_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x83, 0, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew, LegacyOpKind.Ib16 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Add_rm32_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x83, 0, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed, LegacyOpKind.Ib32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Add_rm64_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x83, 0, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq, LegacyOpKind.Ib64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Or_rm16_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x83, 1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew, LegacyOpKind.Ib16 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Or_rm32_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x83, 1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed, LegacyOpKind.Ib32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Or_rm64_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x83, 1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq, LegacyOpKind.Ib64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Adc_rm16_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x83, 2, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew, LegacyOpKind.Ib16 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Adc_rm32_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x83, 2, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed, LegacyOpKind.Ib32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Adc_rm64_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x83, 2, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq, LegacyOpKind.Ib64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sbb_rm16_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x83, 3, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew, LegacyOpKind.Ib16 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sbb_rm32_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x83, 3, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed, LegacyOpKind.Ib32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sbb_rm64_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x83, 3, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq, LegacyOpKind.Ib64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.And_rm16_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x83, 4, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew, LegacyOpKind.Ib16 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.And_rm32_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x83, 4, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed, LegacyOpKind.Ib32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.And_rm64_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x83, 4, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq, LegacyOpKind.Ib64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sub_rm16_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x83, 5, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew, LegacyOpKind.Ib16 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sub_rm32_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x83, 5, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed, LegacyOpKind.Ib32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sub_rm64_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x83, 5, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq, LegacyOpKind.Ib64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xor_rm16_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x83, 6, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew, LegacyOpKind.Ib16 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xor_rm32_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x83, 6, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed, LegacyOpKind.Ib32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xor_rm64_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x83, 6, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq, LegacyOpKind.Ib64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmp_rm16_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x83, 7, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Ib16 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmp_rm32_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x83, 7, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Ib32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmp_rm64_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x83, 7, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Ib64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Test_rm8_r8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x84, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb, LegacyOpKind.Gb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Test_rm16_r16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x85, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Test_rm32_r32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x85, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Test_rm64_r64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x85, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Gq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xchg_rm8_r8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x86, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eb, LegacyOpKind.Gb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xchg_rm16_r16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x87, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xchg_rm32_r32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x87, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xchg_rm64_r64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x87, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq, LegacyOpKind.Gq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_rm8_r8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x88, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eb, LegacyOpKind.Gb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_rm16_r16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x89, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_rm32_r32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x89, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_rm64_r64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x89, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq, LegacyOpKind.Gq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_r8_rm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x8A, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gb, LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x8B, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_r32_rm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x8B, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_r64_rm64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x8B, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_rm16_Sreg)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x8C, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Sw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_r32m16_Sreg)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x8C, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Sw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_r64m16_Sreg)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x8C, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Sw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lea_r16_m)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x8D, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lea_r32_m)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x8D, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lea_r64_m)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x8D, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_Sreg_rm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x8E, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Sw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_Sreg_r32m16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x8E, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Sw, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_Sreg_r64m16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x8E, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Sw, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pop_rm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x8F, 0, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pop_rm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x8F, 0, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pop_rm64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x8F, 0, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Nopw)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x90, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Nopd)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x90, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Nopq)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x90, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xchg_r16_AX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x90, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.r16_rw, LegacyOpKind.AX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xchg_r32_EAX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x90, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.r32_rd, LegacyOpKind.EAX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xchg_r64_RAX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x90, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.r64_ro, LegacyOpKind.RAX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pause)], MandatoryPrefix.PF3, OpCodeTableKind.Normal, 0x90, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cbw)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x98, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cwde)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x98, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cdqe)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x98, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cwd)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x99, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cdq)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x99, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cqo)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x99, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Call_ptr1616)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x9A, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Aww }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Call_ptr1632)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x9A, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Adw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Wait)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x9B, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pushfw)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x9C, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pushfd)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x9C, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pushfq)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x9C, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Popfw)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x9D, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Popfd)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x9D, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Popfq)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x9D, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sahf)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x9E, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lahf)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0x9F, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_AL_moffs8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xA0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.AL, LegacyOpKind.Ob }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_AX_moffs16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xA1, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.AX, LegacyOpKind.Ow }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_EAX_moffs32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xA1, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.EAX, LegacyOpKind.Od }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_RAX_moffs64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xA1, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.RAX, LegacyOpKind.Oq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_moffs8_AL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xA2, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ob, LegacyOpKind.AL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_moffs16_AX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xA3, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ow, LegacyOpKind.AX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_moffs32_EAX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xA3, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Od, LegacyOpKind.EAX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_moffs64_RAX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xA3, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Oq, LegacyOpKind.RAX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movsb_m8_m8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xA4, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, new[] { LegacyOpKind.Yb, LegacyOpKind.Xb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movsw_m16_m16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xA5, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, new[] { LegacyOpKind.Yw, LegacyOpKind.Xw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movsd_m32_m32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xA5, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, new[] { LegacyOpKind.Yd, LegacyOpKind.Xd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movsq_m64_m64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xA5, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, new[] { LegacyOpKind.Yq, LegacyOpKind.Xq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmpsb_m8_m8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xA6, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix | OpCodeFlags.RepnePrefix, new[] { LegacyOpKind.Xb, LegacyOpKind.Yb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmpsw_m16_m16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xA7, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix | OpCodeFlags.RepnePrefix, new[] { LegacyOpKind.Xw, LegacyOpKind.Yw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmpsd_m32_m32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xA7, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix | OpCodeFlags.RepnePrefix, new[] { LegacyOpKind.Xd, LegacyOpKind.Yd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmpsq_m64_m64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xA7, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix | OpCodeFlags.RepnePrefix, new[] { LegacyOpKind.Xq, LegacyOpKind.Yq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Test_AL_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xA8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.AL, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Test_AX_imm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xA9, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.AX, LegacyOpKind.Iw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Test_EAX_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xA9, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.EAX, LegacyOpKind.Id }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Test_RAX_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xA9, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.RAX, LegacyOpKind.Id64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Stosb_m8_AL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xAA, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, new[] { LegacyOpKind.Yb, LegacyOpKind.AL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Stosw_m16_AX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xAB, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, new[] { LegacyOpKind.Yw, LegacyOpKind.AX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Stosd_m32_EAX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xAB, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, new[] { LegacyOpKind.Yd, LegacyOpKind.EAX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Stosq_m64_RAX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xAB, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, new[] { LegacyOpKind.Yq, LegacyOpKind.RAX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lodsb_AL_m8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xAC, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, new[] { LegacyOpKind.AL, LegacyOpKind.Xb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lodsw_AX_m16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xAD, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, new[] { LegacyOpKind.AX, LegacyOpKind.Xw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lodsd_EAX_m32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xAD, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, new[] { LegacyOpKind.EAX, LegacyOpKind.Xd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lodsq_RAX_m64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xAD, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, new[] { LegacyOpKind.RAX, LegacyOpKind.Xq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Scasb_AL_m8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xAE, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix | OpCodeFlags.RepnePrefix, new[] { LegacyOpKind.AL, LegacyOpKind.Yb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Scasw_AX_m16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xAF, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix | OpCodeFlags.RepnePrefix, new[] { LegacyOpKind.AX, LegacyOpKind.Yw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Scasd_EAX_m32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xAF, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix | OpCodeFlags.RepnePrefix, new[] { LegacyOpKind.EAX, LegacyOpKind.Yd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Scasq_RAX_m64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xAF, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix | OpCodeFlags.RepnePrefix, new[] { LegacyOpKind.RAX, LegacyOpKind.Yq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_r8_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xB0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.r8_rb, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_r16_imm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xB8, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.r16_rw, LegacyOpKind.Iw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_r32_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xB8, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.r32_rd, LegacyOpKind.Id }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_r64_imm64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xB8, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.r64_ro, LegacyOpKind.Iq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rol_rm8_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC0, 0, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ror_rm8_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC0, 1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rcl_rm8_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC0, 2, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rcr_rm8_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC0, 3, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shl_rm8_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC0, 4, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shr_rm8_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC0, 5, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sal_rm8_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC0, 6, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sar_rm8_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC0, 7, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rol_rm16_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC1, 0, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rol_rm32_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC1, 0, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rol_rm64_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC1, 0, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ror_rm16_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC1, 1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ror_rm32_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC1, 1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ror_rm64_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC1, 1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rcl_rm16_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC1, 2, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rcl_rm32_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC1, 2, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rcl_rm64_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC1, 2, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rcr_rm16_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC1, 3, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rcr_rm32_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC1, 3, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rcr_rm64_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC1, 3, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shl_rm16_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC1, 4, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shl_rm32_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC1, 4, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shl_rm64_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC1, 4, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shr_rm16_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC1, 5, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shr_rm32_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC1, 5, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shr_rm64_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC1, 5, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sal_rm16_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC1, 6, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sal_rm32_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC1, 6, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sal_rm64_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC1, 6, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sar_rm16_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC1, 7, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sar_rm32_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC1, 7, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sar_rm64_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC1, 7, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Retnw_imm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC2, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix, new[] { LegacyOpKind.Iw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Retnd_imm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC2, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix, new[] { LegacyOpKind.Iw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Retnq_imm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC2, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix, new[] { LegacyOpKind.Iw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Retnw)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC3, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Retnd)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC3, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Retnq)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC3, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Les_r16_m1616)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC4, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Gw, LegacyOpKind.Mp }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Les_r32_m1632)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC4, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Gd, LegacyOpKind.Mp }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lds_r16_m1616)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC5, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Gw, LegacyOpKind.Mp }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lds_r32_m1632)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC5, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Gd, LegacyOpKind.Mp }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_rm8_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC6, 0, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eb, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xabort_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC6F8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_rm16_imm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC7, 0, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew, LegacyOpKind.Iw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_rm32_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC7, 0, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed, LegacyOpKind.Id }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_rm64_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC7, 0, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq, LegacyOpKind.Id64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xbegin_rel16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC7F8, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Jxw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xbegin_rel32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC7F8, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Jxd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Enterw_imm16_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC8, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Iw, LegacyOpKind.Ib21 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Enterd_imm16_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC8, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Iw, LegacyOpKind.Ib21 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Enterq_imm16_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Iw, LegacyOpKind.Ib21 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Leavew)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC9, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Leaved)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC9, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Leaveq)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xC9, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Retfw_imm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xCA, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Iw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Retfd_imm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xCA, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Iw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Retfq_imm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xCA, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Iw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Retfw)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xCB, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Retfd)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xCB, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Retfq)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xCB, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Int3)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xCC, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Int_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xCD, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Into)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xCE, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Iretw)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xCF, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Iretd)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xCF, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Iretq)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xCF, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rol_rm8_1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD0, 0, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb, LegacyOpKind.Imm1 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ror_rm8_1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD0, 1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb, LegacyOpKind.Imm1 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rcl_rm8_1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD0, 2, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb, LegacyOpKind.Imm1 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rcr_rm8_1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD0, 3, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb, LegacyOpKind.Imm1 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shl_rm8_1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD0, 4, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb, LegacyOpKind.Imm1 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shr_rm8_1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD0, 5, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb, LegacyOpKind.Imm1 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sal_rm8_1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD0, 6, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb, LegacyOpKind.Imm1 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sar_rm8_1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD0, 7, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb, LegacyOpKind.Imm1 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rol_rm16_1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD1, 0, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Imm1 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rol_rm32_1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD1, 0, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Imm1 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rol_rm64_1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD1, 0, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Imm1 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ror_rm16_1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD1, 1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Imm1 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ror_rm32_1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD1, 1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Imm1 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ror_rm64_1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD1, 1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Imm1 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rcl_rm16_1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD1, 2, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Imm1 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rcl_rm32_1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD1, 2, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Imm1 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rcl_rm64_1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD1, 2, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Imm1 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rcr_rm16_1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD1, 3, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Imm1 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rcr_rm32_1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD1, 3, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Imm1 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rcr_rm64_1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD1, 3, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Imm1 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shl_rm16_1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD1, 4, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Imm1 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shl_rm32_1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD1, 4, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Imm1 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shl_rm64_1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD1, 4, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Imm1 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shr_rm16_1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD1, 5, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Imm1 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shr_rm32_1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD1, 5, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Imm1 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shr_rm64_1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD1, 5, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Imm1 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sal_rm16_1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD1, 6, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Imm1 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sal_rm32_1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD1, 6, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Imm1 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sal_rm64_1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD1, 6, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Imm1 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sar_rm16_1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD1, 7, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Imm1 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sar_rm32_1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD1, 7, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Imm1 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sar_rm64_1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD1, 7, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Imm1 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rol_rm8_CL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD2, 0, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ror_rm8_CL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD2, 1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rcl_rm8_CL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD2, 2, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rcr_rm8_CL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD2, 3, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shl_rm8_CL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD2, 4, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shr_rm8_CL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD2, 5, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sal_rm8_CL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD2, 6, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sar_rm8_CL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD2, 7, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rol_rm16_CL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD3, 0, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rol_rm32_CL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD3, 0, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rol_rm64_CL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD3, 0, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ror_rm16_CL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD3, 1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ror_rm32_CL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD3, 1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ror_rm64_CL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD3, 1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rcl_rm16_CL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD3, 2, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rcl_rm32_CL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD3, 2, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rcl_rm64_CL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD3, 2, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rcr_rm16_CL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD3, 3, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rcr_rm32_CL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD3, 3, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rcr_rm64_CL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD3, 3, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shl_rm16_CL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD3, 4, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shl_rm32_CL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD3, 4, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shl_rm64_CL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD3, 4, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shr_rm16_CL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD3, 5, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shr_rm32_CL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD3, 5, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shr_rm64_CL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD3, 5, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sal_rm16_CL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD3, 6, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sal_rm32_CL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD3, 6, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sal_rm64_CL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD3, 6, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sar_rm16_CL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD3, 7, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sar_rm32_CL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD3, 7, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sar_rm64_CL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD3, 7, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Aam_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD4, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Aad_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD5, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Salc)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD6, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xlat_m8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD7, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.MRBX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fadd_m32fp)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD8, 0, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mf32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fmul_m32fp)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD8, 1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mf32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fcom_m32fp)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD8, 2, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mf32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fcomp_m32fp)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD8, 3, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mf32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fsub_m32fp)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD8, 4, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mf32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fsubr_m32fp)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD8, 5, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mf32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fdiv_m32fp)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD8, 6, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mf32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fdivr_m32fp)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD8, 7, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mf32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fadd_st0_sti)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD8C0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.ST, LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fmul_st0_sti)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD8C8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.ST, LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fcom_st0_sti)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD8D0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.ST, LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fcomp_st0_sti)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD8D8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.ST, LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fsub_st0_sti)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD8E0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.ST, LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fsubr_st0_sti)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD8E8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.ST, LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fdiv_st0_sti)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD8F0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.ST, LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fdivr_st0_sti)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD8F8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.ST, LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fld_m32fp)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9, 0, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mf32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fst_m32fp)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9, 2, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mf32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fstp_m32fp)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9, 3, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mf32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fldenv_m14byte)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9, 4, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.M14 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fldenv_m28byte)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9, 4, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.M28 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fldcw_m2byte)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9, 5, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fnstenv_m14byte)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9, 6, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.M14 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fstenv_m14byte)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9, 6, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Fwait, new[] { LegacyOpKind.M14 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fnstenv_m28byte)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9, 6, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.M28 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fstenv_m28byte)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9, 6, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Fwait, new[] { LegacyOpKind.M28 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fnstcw_m2byte)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9, 7, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fstcw_m2byte)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9, 7, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Fwait, new[] { LegacyOpKind.Mw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fld_st0_sti)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9C0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.ST, LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fxch_st0_sti)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9C8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.ST, LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fnop)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9D0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fstpnce_sti)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9D8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fchs)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9E0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fabs)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9E1, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ftst)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9E4, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fxam)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9E5, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fld1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9E8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fldl2t)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9E9, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fldl2e)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9EA, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fldpi)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9EB, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fldlg2)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9EC, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fldln2)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9ED, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fldz)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9EE, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.F2xm1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9F0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fyl2x)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9F1, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fptan)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9F2, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fpatan)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9F3, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fxtract)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9F4, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fprem1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9F5, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fdecstp)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9F6, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fincstp)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9F7, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fprem)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9F8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fyl2xp1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9F9, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fsqrt)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9FA, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fsincos)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9FB, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Frndint)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9FC, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fscale)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9FD, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fsin)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9FE, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fcos)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xD9FF, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fiadd_m32int)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDA, 0, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mfi32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fimul_m32int)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDA, 1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mfi32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ficom_m32int)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDA, 2, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mfi32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ficomp_m32int)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDA, 3, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mfi32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fisub_m32int)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDA, 4, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mfi32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fisubr_m32int)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDA, 5, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mfi32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fidiv_m32int)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDA, 6, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mfi32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fidivr_m32int)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDA, 7, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mfi32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fcmovb_st0_sti)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDAC0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.ST, LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fcmove_st0_sti)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDAC8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.ST, LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fcmovbe_st0_sti)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDAD0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.ST, LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fcmovu_st0_sti)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDAD8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.ST, LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fucompp)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDAE9, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fild_m32int)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDB, 0, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mfi32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fisttp_m32int)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDB, 1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mfi32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fist_m32int)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDB, 2, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mfi32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fistp_m32int)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDB, 3, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mfi32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fld_m80fp)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDB, 5, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mf80 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fstp_m80fp)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDB, 7, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mf80 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fcmovnb_st0_sti)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDBC0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.ST, LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fcmovne_st0_sti)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDBC8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.ST, LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fcmovnbe_st0_sti)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDBD0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.ST, LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fcmovnu_st0_sti)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDBD8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.ST, LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fneni)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDBE0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Feni)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDBE0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Fwait, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fndisi)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDBE1, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fdisi)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDBE1, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Fwait, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fnclex)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDBE2, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fclex)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDBE2, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Fwait, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fninit)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDBE3, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Finit)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDBE3, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Fwait, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fnsetpm)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDBE4, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fsetpm)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDBE4, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Fwait, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Frstpm)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDBE5, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fucomi_st0_sti)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDBE8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.ST, LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fcomi_st0_sti)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDBF0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.ST, LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fadd_m64fp)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDC, 0, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mf64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fmul_m64fp)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDC, 1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mf64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fcom_m64fp)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDC, 2, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mf64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fcomp_m64fp)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDC, 3, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mf64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fsub_m64fp)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDC, 4, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mf64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fsubr_m64fp)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDC, 5, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mf64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fdiv_m64fp)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDC, 6, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mf64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fdivr_m64fp)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDC, 7, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mf64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fadd_sti_st0)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDCC0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.STi, LegacyOpKind.ST }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fmul_sti_st0)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDCC8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.STi, LegacyOpKind.ST }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fcom_st0_sti_DCD0)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDCD0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.ST, LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fcomp_st0_sti_DCD8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDCD8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.ST, LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fsubr_sti_st0)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDCE0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.STi, LegacyOpKind.ST }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fsub_sti_st0)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDCE8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.STi, LegacyOpKind.ST }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fdivr_sti_st0)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDCF0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.STi, LegacyOpKind.ST }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fdiv_sti_st0)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDCF8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.STi, LegacyOpKind.ST }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fld_m64fp)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDD, 0, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mf64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fisttp_m64int)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDD, 1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mf64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fst_m64fp)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDD, 2, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mf64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fstp_m64fp)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDD, 3, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mf64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Frstor_m94byte)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDD, 4, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.M98 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Frstor_m108byte)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDD, 4, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.M108 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fnsave_m94byte)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDD, 6, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.M98 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fsave_m94byte)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDD, 6, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Fwait, new[] { LegacyOpKind.M98 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fnsave_m108byte)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDD, 6, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.M108 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fsave_m108byte)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDD, 6, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Fwait, new[] { LegacyOpKind.M108 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fnstsw_m2byte)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDD, 7, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fstsw_m2byte)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDD, 7, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Fwait, new[] { LegacyOpKind.Mw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ffree_sti)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDDC0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fxch_st0_sti_DDC8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDDC8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.ST, LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fst_sti)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDDD0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fstp_sti)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDDD8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fucom_st0_sti)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDDE0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.ST, LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fucomp_st0_sti)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDDE8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.ST, LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fiadd_m16int)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDE, 0, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mfi16 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fimul_m16int)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDE, 1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mfi16 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ficom_m16int)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDE, 2, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mfi16 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ficomp_m16int)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDE, 3, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mfi16 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fisub_m16int)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDE, 4, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mfi16 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fisubr_m16int)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDE, 5, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mfi16 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fidiv_m16int)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDE, 6, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mfi16 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fidivr_m16int)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDE, 7, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mfi16 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Faddp_sti_st0)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDEC0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.STi, LegacyOpKind.ST }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fmulp_sti_st0)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDEC8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.STi, LegacyOpKind.ST }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fcomp_st0_sti_DED0)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDED0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.ST, LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fcompp)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDED9, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fsubrp_sti_st0)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDEE0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.STi, LegacyOpKind.ST }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fsubp_sti_st0)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDEE8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.STi, LegacyOpKind.ST }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fdivrp_sti_st0)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDEF0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.STi, LegacyOpKind.ST }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fdivp_sti_st0)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDEF8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.STi, LegacyOpKind.ST }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fild_m16int)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDF, 0, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mfi16 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fisttp_m16int)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDF, 1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mfi16 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fist_m16int)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDF, 2, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mfi16 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fistp_m16int)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDF, 3, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mfi16 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fbld_m80bcd)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDF, 4, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mfbcd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fild_m64int)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDF, 5, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mfi64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fbstp_m80bcd)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDF, 6, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mfbcd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fistp_m64int)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDF, 7, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mfi64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ffreep_sti)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDFC0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fxch_st0_sti_DFC8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDFC8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.ST, LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fstp_sti_DFD0)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDFD0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fstp_sti_DFD8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDFD8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fnstsw_AX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDFE0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.AX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fstsw_AX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDFE0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Fwait, new[] { LegacyOpKind.AX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fstdw_AX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDFE1, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.AX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fstsg_AX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDFE2, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.AX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fucomip_st0_sti)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDFE8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.ST, LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fcomip_st0_sti)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xDFF0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.ST, LegacyOpKind.STi }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Loopne_rel8_16_CX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE0, -1, OperandSize.Size16, AddressSize.Size16, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.wJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Loopne_rel8_32_CX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE0, -1, OperandSize.Size32, AddressSize.Size16, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.dJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Loopne_rel8_16_ECX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE0, -1, OperandSize.Size16, AddressSize.Size32, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.wJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Loopne_rel8_32_ECX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE0, -1, OperandSize.Size32, AddressSize.Size32, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.dJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Loopne_rel8_64_ECX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE0, -1, OperandSize.None, AddressSize.Size32, OpCodeFlags.Mode64, new[] { LegacyOpKind.qJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Loopne_rel8_16_RCX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE0, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.wJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Loopne_rel8_64_RCX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.qJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Loope_rel8_16_CX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE1, -1, OperandSize.Size16, AddressSize.Size16, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.wJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Loope_rel8_32_CX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE1, -1, OperandSize.Size32, AddressSize.Size16, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.dJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Loope_rel8_16_ECX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE1, -1, OperandSize.Size16, AddressSize.Size32, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.wJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Loope_rel8_32_ECX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE1, -1, OperandSize.Size32, AddressSize.Size32, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.dJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Loope_rel8_64_ECX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE1, -1, OperandSize.None, AddressSize.Size32, OpCodeFlags.Mode64, new[] { LegacyOpKind.qJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Loope_rel8_16_RCX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE1, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.wJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Loope_rel8_64_RCX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE1, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.qJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Loop_rel8_16_CX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE2, -1, OperandSize.Size16, AddressSize.Size16, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.wJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Loop_rel8_32_CX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE2, -1, OperandSize.Size32, AddressSize.Size16, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.dJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Loop_rel8_16_ECX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE2, -1, OperandSize.Size16, AddressSize.Size32, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.wJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Loop_rel8_32_ECX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE2, -1, OperandSize.Size32, AddressSize.Size32, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.dJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Loop_rel8_64_ECX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE2, -1, OperandSize.None, AddressSize.Size32, OpCodeFlags.Mode64, new[] { LegacyOpKind.qJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Loop_rel8_16_RCX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE2, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.wJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Loop_rel8_64_RCX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE2, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.qJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jcxz_rel8_16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE3, -1, OperandSize.Size16, AddressSize.Size16, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.wJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jcxz_rel8_32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE3, -1, OperandSize.Size32, AddressSize.Size16, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.dJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jecxz_rel8_16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE3, -1, OperandSize.Size16, AddressSize.Size32, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.wJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jecxz_rel8_32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE3, -1, OperandSize.Size32, AddressSize.Size32, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.dJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jecxz_rel8_64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE3, -1, OperandSize.None, AddressSize.Size32, OpCodeFlags.Mode64, new[] { LegacyOpKind.qJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jrcxz_rel8_16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE3, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.wJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jrcxz_rel8_64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE3, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.qJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.In_AL_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE4, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.AL, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.In_AX_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE5, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.AX, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.In_EAX_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE5, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.EAX, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Out_imm8_AL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE6, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ib, LegacyOpKind.AL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Out_imm8_AX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE7, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ib, LegacyOpKind.AX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Out_imm8_EAX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE7, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ib, LegacyOpKind.EAX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Call_rel16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE8, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix, new[] { LegacyOpKind.Jw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Call_rel32_32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE8, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix, new[] { LegacyOpKind.dJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Call_rel32_64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix, new[] { LegacyOpKind.qJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jmp_rel16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE9, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix, new[] { LegacyOpKind.Jw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jmp_rel32_32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE9, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix, new[] { LegacyOpKind.dJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jmp_rel32_64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xE9, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix, new[] { LegacyOpKind.qJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jmp_ptr1616)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xEA, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Aww }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jmp_ptr1632)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xEA, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Adw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jmp_rel8_16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xEB, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.wJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jmp_rel8_32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xEB, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.dJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jmp_rel8_64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xEB, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.qJb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.In_AL_DX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xEC, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.AL, LegacyOpKind.DX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.In_AX_DX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xED, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.AX, LegacyOpKind.DX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.In_EAX_DX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xED, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.EAX, LegacyOpKind.DX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Out_DX_AL)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xEE, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.DX, LegacyOpKind.AL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Out_DX_AX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xEF, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.DX, LegacyOpKind.AX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Out_DX_EAX)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xEF, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.DX, LegacyOpKind.EAX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Int1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF1, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Hlt)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF4, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmc)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF5, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Test_rm8_imm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF6, 0, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Test_rm8_imm8_F6r1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF6, 1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Not_rm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF6, 2, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Neg_rm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF6, 3, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mul_rm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF6, 4, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Imul_rm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF6, 5, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Div_rm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF6, 6, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Idiv_rm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF6, 7, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Test_rm16_imm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF7, 0, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Iw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Test_rm32_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF7, 0, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Id }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Test_rm64_imm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF7, 0, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Id64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Test_rm16_imm16_F7r1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF7, 1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Iw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Test_rm32_imm32_F7r1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF7, 1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Id }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Test_rm64_imm32_F7r1)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF7, 1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Id64 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Not_rm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF7, 2, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Not_rm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF7, 2, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Not_rm64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF7, 2, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Neg_rm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF7, 3, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Neg_rm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF7, 3, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Neg_rm64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF7, 3, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mul_rm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF7, 4, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mul_rm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF7, 4, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mul_rm64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF7, 4, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Imul_rm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF7, 5, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Imul_rm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF7, 5, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Imul_rm64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF7, 5, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Div_rm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF7, 6, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Div_rm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF7, 6, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Div_rm64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF7, 6, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Idiv_rm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF7, 7, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Idiv_rm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF7, 7, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Idiv_rm64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF7, 7, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Clc)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Stc)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xF9, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cli)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xFA, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sti)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xFB, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cld)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xFC, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Std)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xFD, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Inc_rm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xFE, 0, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Dec_rm8)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xFE, 1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Inc_rm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xFF, 0, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Inc_rm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xFF, 0, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Inc_rm64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xFF, 0, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Dec_rm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xFF, 1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Dec_rm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xFF, 1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Dec_rm64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xFF, 1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Call_rm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xFF, 2, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.NotrackPrefix, new[] { LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Call_rm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xFF, 2, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.NotrackPrefix, new[] { LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Call_rm64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xFF, 2, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.NotrackPrefix, new[] { LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Call_m1616)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xFF, 3, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eww }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Call_m1632)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xFF, 3, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Edw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Call_m1664)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xFF, 3, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eqw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jmp_rm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xFF, 4, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.NotrackPrefix, new[] { LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jmp_rm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xFF, 4, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.NotrackPrefix, new[] { LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jmp_rm64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xFF, 4, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.NotrackPrefix, new[] { LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jmp_m1616)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xFF, 5, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eww }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jmp_m1632)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xFF, 5, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Edw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jmp_m1664)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xFF, 5, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eqw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Push_rm16)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xFF, 6, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Push_rm32)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xFF, 6, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Push_rm64)], MandatoryPrefix.None, OpCodeTableKind.Normal, 0xFF, 6, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sldt_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x00, 0, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sldt_r32m16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x00, 0, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew_d }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sldt_r64m16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x00, 0, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew_q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Str_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x00, 1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Str_r32m16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x00, 1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew_d }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Str_r64m16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x00, 1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew_q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lldt_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x00, 2, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lldt_r32m16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x00, 2, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew_d }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lldt_r64m16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x00, 2, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew_q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ltr_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x00, 3, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ltr_r32m16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x00, 3, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew_d }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ltr_r64m16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x00, 3, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew_q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Verr_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x00, 4, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Verr_r32m16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x00, 4, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew_d }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Verr_r64m16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x00, 4, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew_q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Verw_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x00, 5, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Verw_r32m16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x00, 5, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew_d }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Verw_r64m16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x00, 5, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew_q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jmpe_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x00, 6, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jmpe_rm32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x00, 6, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sgdt_m1632_16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01, 0, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Ms }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sgdt_m1632)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01, 0, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Ms }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sgdt_m1664)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01, 0, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Ms }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sidt_m1632_16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01, 1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Ms }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sidt_m1632)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01, 1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Ms }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sidt_m1664)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01, 1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Ms }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lgdt_m1632_16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01, 2, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Ms }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lgdt_m1632)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01, 2, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Ms }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lgdt_m1664)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01, 2, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Ms }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lidt_m1632_16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01, 3, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Ms }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lidt_m1632)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01, 3, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Ms }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lidt_m1664)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01, 3, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Ms }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Smsw_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01, 4, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Smsw_r32m16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01, 4, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew_d }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Smsw_r64m16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01, 4, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew_q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rstorssp_m64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x01, 5, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lmsw_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01, 6, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lmsw_r32m16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01, 6, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew_d }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lmsw_r64m16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01, 6, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew_q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Invlpg_m)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01, 7, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Enclv)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x01C0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Vmcall)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x01C1, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Vmlaunch)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x01C2, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Vmresume)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x01C3, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Vmxoff)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x01C4, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pconfig)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x01C5, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Monitorw)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x01C8, -1, OperandSize.None, AddressSize.Size16, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Monitord)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x01C8, -1, OperandSize.None, AddressSize.Size32, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Monitorq)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x01C8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mwait)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x01C9, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Clac)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x01CA, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Stac)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x01CB, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Encls)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x01CF, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xgetbv)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x01D0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xsetbv)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x01D1, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Vmfunc)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x01D4, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xend)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x01D5, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xtest)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x01D6, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Enclu)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x01D7, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Vmrunw)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01D8, -1, OperandSize.None, AddressSize.Size16, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Vmrund)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01D8, -1, OperandSize.None, AddressSize.Size32, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Vmrunq)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01D8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Vmmcall)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01D9, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Vmloadw)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01DA, -1, OperandSize.None, AddressSize.Size16, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Vmloadd)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01DA, -1, OperandSize.None, AddressSize.Size32, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Vmloadq)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01DA, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Vmsavew)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01DB, -1, OperandSize.None, AddressSize.Size16, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Vmsaved)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01DB, -1, OperandSize.None, AddressSize.Size32, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Vmsaveq)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01DB, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Stgi)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01DC, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Clgi)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01DD, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Skinit)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01DE, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Invlpgaw)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01DF, -1, OperandSize.None, AddressSize.Size16, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Invlpgad)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01DF, -1, OperandSize.None, AddressSize.Size32, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Invlpgaq)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01DF, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Setssbsy)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x01E8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Saveprevssp)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x01EA, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rdpkru)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x01EE, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Wrpkru)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x01EF, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Swapgs)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01F8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rdtscp)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01F9, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Monitorxw)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x01FA, -1, OperandSize.None, AddressSize.Size16, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Monitorxd)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x01FA, -1, OperandSize.None, AddressSize.Size32, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Monitorxq)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x01FA, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mcommit)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x01FA, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mwaitx)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01FB, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Clzerow)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01FC, -1, OperandSize.None, AddressSize.Size16, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Clzerod)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01FC, -1, OperandSize.None, AddressSize.Size32, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Clzeroq)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01FC, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rdpru)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x01FD, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lar_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x02, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lar_r32_r32m16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x02, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lar_r64_r64m16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x02, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lsl_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x03, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lsl_r32_r32m16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x03, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lsl_r64_r64m16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x03, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Loadallreset286)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x04, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Loadall286)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x05, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Syscall)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x05, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Clts)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x06, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Loadall386)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x07, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sysretd)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x07, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sysretq)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x07, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Invd)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x08, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Wbinvd)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x09, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Wbnoinvd)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x09, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cl1invmb)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x0A, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ud2)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x0B, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.ReservedNop_rm16_r16_0F0D)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x0D, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.ReservedNop_rm32_r32_0F0D)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x0D, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.ReservedNop_rm64_r64_0F0D)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x0D, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Gq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Prefetch_m8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x0D, 0, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Prefetchw_m8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x0D, 1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Prefetchwt1_m8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x0D, 2, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Femms)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x0E, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Umov_rm8_r8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x10, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Eb, LegacyOpKind.Gb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Umov_rm16_r16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x11, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Umov_rm32_r32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x11, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Umov_r8_rm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x12, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Gb, LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Umov_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x13, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Umov_r32_rm32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x13, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movups_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x10, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovups_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x10, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovups_ymm_ymmm256)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x10, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovups_xmm_k1z_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x10, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovups_ymm_k1z_ymmm256)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x10, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovups_zmm_k1z_zmmm512)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x10, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movupd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x10, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovupd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x10, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovupd_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x10, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovupd_xmm_k1z_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x10, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovupd_ymm_k1z_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x10, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovupd_zmm_k1z_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x10, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movss_xmm_xmmm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x10, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovss_xmm_xmm_xmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x10, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.RX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovss_xmm_m32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x10, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.M }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovss_xmm_k1z_xmm_xmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x10, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.RX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovss_xmm_k1z_m32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x10, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movsd_xmm_xmmm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x10, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovsd_xmm_xmm_xmm)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x10, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.RX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovsd_xmm_m64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x10, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.M }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovsd_xmm_k1z_xmm_xmm)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x10, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.RX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovsd_xmm_k1z_m64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x10, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movups_xmmm128_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x11, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.WX, LegacyOpKind.VX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovups_xmmm128_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x11, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.WX, VexOpKind.VX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovups_ymmm256_ymm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x11, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.WY, VexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovups_xmmm128_k1z_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x11, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovups_ymmm256_k1z_ymm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x11, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WY, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovups_zmmm512_k1z_zmm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x11, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WZ, EvexOpKind.VZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movupd_xmmm128_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x11, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.WX, LegacyOpKind.VX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovupd_xmmm128_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x11, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.WX, VexOpKind.VX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovupd_ymmm256_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x11, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.WY, VexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovupd_xmmm128_k1z_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x11, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovupd_ymmm256_k1z_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x11, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WY, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovupd_zmmm512_k1z_zmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x11, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WZ, EvexOpKind.VZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movss_xmmm32_xmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x11, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.WX, LegacyOpKind.VX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovss_xmm_xmm_xmm_0F11)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x11, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.RX, VexOpKind.HX, VexOpKind.VX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovss_m32_xmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x11, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.M, VexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovss_xmm_k1z_xmm_xmm_0F11)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x11, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.RX, EvexOpKind.HX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovss_m32_k1_xmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x11, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.M, EvexOpKind.VX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movsd_xmmm64_xmm)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x11, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.WX, LegacyOpKind.VX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovsd_xmm_xmm_xmm_0F11)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x11, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.RX, VexOpKind.HX, VexOpKind.VX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovsd_m64_xmm)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x11, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.M, VexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovsd_xmm_k1z_xmm_xmm_0F11)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x11, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.RX, EvexOpKind.HX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovsd_m64_k1_xmm)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x11, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.M, EvexOpKind.VX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movhlps_xmm_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x12, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.RX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movlps_xmm_m64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x12, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.M }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovhlps_xmm_xmm_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x12, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.RX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovlps_xmm_xmm_m64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x12, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.M }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovhlps_xmm_xmm_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x12, -1, EvexVectorLength.L128, TupleType.Tuple2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.RX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovlps_xmm_xmm_m64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x12, -1, EvexVectorLength.L128, TupleType.Tuple2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movlpd_xmm_m64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x12, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.M }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovlpd_xmm_xmm_m64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x12, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.M }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovlpd_xmm_xmm_m64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x12, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movsldup_xmm_xmmm128)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x12, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovsldup_xmm_xmmm128)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x12, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovsldup_ymm_ymmm256)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x12, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovsldup_xmm_k1z_xmmm128)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x12, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovsldup_ymm_k1z_ymmm256)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x12, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovsldup_zmm_k1z_zmmm512)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x12, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movddup_xmm_xmmm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x12, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovddup_xmm_xmmm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x12, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovddup_ymm_ymmm256)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x12, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovddup_xmm_k1z_xmmm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x12, -1, EvexVectorLength.L128, TupleType.MOVDDUP_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovddup_ymm_k1z_ymmm256)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x12, -1, EvexVectorLength.L256, TupleType.MOVDDUP_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovddup_zmm_k1z_zmmm512)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x12, -1, EvexVectorLength.L512, TupleType.MOVDDUP_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movlps_m64_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x13, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.M, LegacyOpKind.VX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovlps_m64_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x13, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.M, VexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovlps_m64_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x13, -1, EvexVectorLength.L128, TupleType.Tuple2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { EvexOpKind.M, EvexOpKind.VX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movlpd_m64_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x13, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.M, LegacyOpKind.VX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovlpd_m64_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x13, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.M, VexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovlpd_m64_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x13, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.M, EvexOpKind.VX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Unpcklps_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x14, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vunpcklps_xmm_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x14, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vunpcklps_ymm_ymm_ymmm256)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x14, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vunpcklps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x14, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vunpcklps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x14, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vunpcklps_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x14, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Unpcklpd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x14, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vunpcklpd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x14, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vunpcklpd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x14, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vunpcklpd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x14, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vunpcklpd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x14, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vunpcklpd_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x14, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Unpckhps_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x15, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vunpckhps_xmm_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x15, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vunpckhps_ymm_ymm_ymmm256)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x15, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vunpckhps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x15, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vunpckhps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x15, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vunpckhps_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x15, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Unpckhpd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x15, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vunpckhpd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x15, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vunpckhpd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x15, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vunpckhpd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x15, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vunpckhpd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x15, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vunpckhpd_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x15, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movlhps_xmm_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x16, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.RX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovlhps_xmm_xmm_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x16, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.RX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovlhps_xmm_xmm_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x16, -1, EvexVectorLength.L128, TupleType.Tuple2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.RX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movhps_xmm_m64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x16, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.M }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovhps_xmm_xmm_m64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x16, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.M }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovhps_xmm_xmm_m64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x16, -1, EvexVectorLength.L128, TupleType.Tuple2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movhpd_xmm_m64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x16, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.M }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovhpd_xmm_xmm_m64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x16, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.M }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovhpd_xmm_xmm_m64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x16, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movshdup_xmm_xmmm128)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x16, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovshdup_xmm_xmmm128)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x16, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovshdup_ymm_ymmm256)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x16, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovshdup_xmm_k1z_xmmm128)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x16, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovshdup_ymm_k1z_ymmm256)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x16, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovshdup_zmm_k1z_zmmm512)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x16, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movhps_m64_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x17, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.M, LegacyOpKind.VX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovhps_m64_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x17, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.M, VexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovhps_m64_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x17, -1, EvexVectorLength.L128, TupleType.Tuple2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { EvexOpKind.M, EvexOpKind.VX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movhpd_m64_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x17, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.M, LegacyOpKind.VX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovhpd_m64_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x17, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.M, VexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovhpd_m64_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x17, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.M, EvexOpKind.VX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.ReservedNop_rm16_r16_0F18)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x18, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.ReservedNop_rm32_r32_0F18)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x18, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.ReservedNop_rm64_r64_0F18)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x18, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Gq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.ReservedNop_rm16_r16_0F19)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x19, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.ReservedNop_rm32_r32_0F19)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x19, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.ReservedNop_rm64_r64_0F19)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x19, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Gq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.ReservedNop_rm16_r16_0F1A)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x1A, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.ReservedNop_rm32_r32_0F1A)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x1A, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.ReservedNop_rm64_r64_0F1A)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x1A, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Gq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.ReservedNop_rm16_r16_0F1B)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x1B, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.ReservedNop_rm32_r32_0F1B)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x1B, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.ReservedNop_rm64_r64_0F1B)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x1B, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Gq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.ReservedNop_rm16_r16_0F1C)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x1C, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.ReservedNop_rm32_r32_0F1C)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x1C, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.ReservedNop_rm64_r64_0F1C)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x1C, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Gq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.ReservedNop_rm16_r16_0F1D)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x1D, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.ReservedNop_rm32_r32_0F1D)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x1D, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.ReservedNop_rm64_r64_0F1D)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x1D, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Gq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.ReservedNop_rm16_r16_0F1E)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x1E, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.ReservedNop_rm32_r32_0F1E)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x1E, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.ReservedNop_rm64_r64_0F1E)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x1E, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Gq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.ReservedNop_rm16_r16_0F1F)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x1F, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.ReservedNop_rm32_r32_0F1F)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x1F, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.ReservedNop_rm64_r64_0F1F)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x1F, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Gq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Prefetchnta_m8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x18, 0, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Prefetcht0_m8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x18, 1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Prefetcht1_m8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x18, 2, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Prefetcht2_m8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x18, 3, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bndldx_bnd_mib)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x1A, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.B, LegacyOpKind.MIB }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bndmov_bnd_bndm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x1A, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.B, LegacyOpKind.BMq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bndmov_bnd_bndm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x1A, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.B, LegacyOpKind.BMo }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bndcl_bnd_rm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x1A, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.B, LegacyOpKind.Ed_MPX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bndcl_bnd_rm64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x1A, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.B, LegacyOpKind.Eq_MPX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bndcu_bnd_rm32)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x1A, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.B, LegacyOpKind.Ed_MPX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bndcu_bnd_rm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x1A, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.B, LegacyOpKind.Eq_MPX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bndstx_mib_bnd)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x1B, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.MIB, LegacyOpKind.B }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bndmov_bndm64_bnd)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x1B, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.BMq, LegacyOpKind.B }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bndmov_bndm128_bnd)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x1B, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.BMo, LegacyOpKind.B }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bndmk_bnd_m32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x1B, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.B, LegacyOpKind.Md_MPX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bndmk_bnd_m64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x1B, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.B, LegacyOpKind.Mq_MPX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bndcn_bnd_rm32)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x1B, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.B, LegacyOpKind.Ed_MPX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bndcn_bnd_rm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x1B, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.B, LegacyOpKind.Eq_MPX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cldemote_m8)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x1C, 0, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rdsspd_r32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x1E, 1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Rd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rdsspq_r64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x1E, 1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Rq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Endbr64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x1EFA, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Endbr32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x1EFB, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Nop_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x1F, 0, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Nop_rm32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x1F, 0, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Nop_rm64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x1F, 0, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_r32_cr)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x20, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Rd, LegacyOpKind.Cd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_r64_cr)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x20, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Rq, LegacyOpKind.Cq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_r32_dr)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x21, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Rd, LegacyOpKind.Dd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_r64_dr)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x21, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Rq, LegacyOpKind.Dq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_cr_r32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x22, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Cd, LegacyOpKind.Rd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_cr_r64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x22, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Cq, LegacyOpKind.Rq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_dr_r32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x23, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Dd, LegacyOpKind.Rd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_dr_r64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x23, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Dq, LegacyOpKind.Rq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_r32_tr)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x24, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Rd, LegacyOpKind.Td }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mov_tr_r32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x26, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Td, LegacyOpKind.Rd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movaps_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x28, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovaps_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x28, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovaps_ymm_ymmm256)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x28, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovaps_xmm_k1z_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x28, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovaps_ymm_k1z_ymmm256)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x28, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovaps_zmm_k1z_zmmm512)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x28, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movapd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x28, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovapd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x28, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovapd_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x28, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovapd_xmm_k1z_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x28, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovapd_ymm_k1z_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x28, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovapd_zmm_k1z_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x28, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movaps_xmmm128_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x29, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.WX, LegacyOpKind.VX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovaps_xmmm128_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x29, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.WX, VexOpKind.VX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovaps_ymmm256_ymm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x29, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.WY, VexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovaps_xmmm128_k1z_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x29, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovaps_ymmm256_k1z_ymm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x29, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WY, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovaps_zmmm512_k1z_zmm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x29, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WZ, EvexOpKind.VZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movapd_xmmm128_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x29, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.WX, LegacyOpKind.VX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovapd_xmmm128_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x29, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.WX, VexOpKind.VX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovapd_ymmm256_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x29, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.WY, VexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovapd_xmmm128_k1z_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x29, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovapd_ymmm256_k1z_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x29, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WY, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovapd_zmmm512_k1z_zmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x29, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WZ, EvexOpKind.VZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cvtpi2ps_xmm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x2A, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cvtpi2pd_xmm_mmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x2A, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cvtsi2ss_xmm_rm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x2A, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cvtsi2ss_xmm_rm64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x2A, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.Eq }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvtsi2ss_xmm_xmm_rm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x2A, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG32, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.Ed }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvtsi2ss_xmm_xmm_rm64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x2A, -1, VexVectorLength.LIG, OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.Eq }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtsi2ss_xmm_xmm_rm32_er)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x2A, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG32 | OpCodeFlags.RoundingControl, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.Ed }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtsi2ss_xmm_xmm_rm64_er)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x2A, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.RoundingControl, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cvtsi2sd_xmm_rm32)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x2A, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cvtsi2sd_xmm_rm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x2A, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.Eq }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvtsi2sd_xmm_xmm_rm32)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x2A, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG32, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.Ed }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvtsi2sd_xmm_xmm_rm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x2A, -1, VexVectorLength.LIG, OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.Eq }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtsi2sd_xmm_xmm_rm32_er)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x2A, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG32 | OpCodeFlags.RoundingControl, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.Ed }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtsi2sd_xmm_xmm_rm64_er)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x2A, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.RoundingControl, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movntps_m128_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x2B, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.M, LegacyOpKind.VX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovntps_m128_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x2B, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.M, VexOpKind.VX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovntps_m256_ymm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x2B, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.M, VexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovntps_m128_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x2B, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { EvexOpKind.M, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovntps_m256_ymm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x2B, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { EvexOpKind.M, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovntps_m512_zmm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x2B, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { EvexOpKind.M, EvexOpKind.VZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movntpd_m128_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x2B, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.M, LegacyOpKind.VX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovntpd_m128_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x2B, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.M, VexOpKind.VX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovntpd_m256_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x2B, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.M, VexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovntpd_m128_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x2B, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.M, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovntpd_m256_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x2B, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.M, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovntpd_m512_zmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x2B, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.M, EvexOpKind.VZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movntss_m32_xmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x2B, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.M, LegacyOpKind.VX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movntsd_m64_xmm)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x2B, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.M, LegacyOpKind.VX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cvttps2pi_mm_xmmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x2C, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cvttpd2pi_mm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x2C, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cvttss2si_r32_xmmm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x2C, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cvttss2si_r64_xmmm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x2C, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvttss2si_r32_xmmm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x2C, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG32, new[] { VexOpKind.Gd, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvttss2si_r64_xmmm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x2C, -1, VexVectorLength.LIG, OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W, new[] { VexOpKind.Gq, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvttss2si_r32_xmmm32_sae)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x2C, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar_4, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG32 | OpCodeFlags.SuppressAllExceptions, new[] { EvexOpKind.Gd, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvttss2si_r64_xmmm32_sae)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x2C, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar_4, OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.SuppressAllExceptions, new[] { EvexOpKind.Gq, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cvttsd2si_r32_xmmm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x2C, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cvttsd2si_r64_xmmm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x2C, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvttsd2si_r32_xmmm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x2C, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG32, new[] { VexOpKind.Gd, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvttsd2si_r64_xmmm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x2C, -1, VexVectorLength.LIG, OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W, new[] { VexOpKind.Gq, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvttsd2si_r32_xmmm64_sae)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x2C, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar_8, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG32 | OpCodeFlags.SuppressAllExceptions, new[] { EvexOpKind.Gd, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvttsd2si_r64_xmmm64_sae)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x2C, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar_8, OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.SuppressAllExceptions, new[] { EvexOpKind.Gq, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cvtps2pi_mm_xmmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x2D, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cvtpd2pi_mm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x2D, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cvtss2si_r32_xmmm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x2D, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cvtss2si_r64_xmmm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x2D, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvtss2si_r32_xmmm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x2D, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG32, new[] { VexOpKind.Gd, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvtss2si_r64_xmmm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x2D, -1, VexVectorLength.LIG, OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W, new[] { VexOpKind.Gq, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtss2si_r32_xmmm32_er)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x2D, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar_4, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG32 | OpCodeFlags.RoundingControl, new[] { EvexOpKind.Gd, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtss2si_r64_xmmm32_er)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x2D, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar_4, OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.RoundingControl, new[] { EvexOpKind.Gq, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cvtsd2si_r32_xmmm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x2D, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cvtsd2si_r64_xmmm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x2D, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvtsd2si_r32_xmmm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x2D, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG32, new[] { VexOpKind.Gd, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvtsd2si_r64_xmmm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x2D, -1, VexVectorLength.LIG, OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W, new[] { VexOpKind.Gq, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtsd2si_r32_xmmm64_er)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x2D, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar_8, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG32 | OpCodeFlags.RoundingControl, new[] { EvexOpKind.Gd, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtsd2si_r64_xmmm64_er)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x2D, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar_8, OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.RoundingControl, new[] { EvexOpKind.Gq, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ucomiss_xmm_xmmm32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x2E, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vucomiss_xmm_xmmm32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x2E, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vucomiss_xmm_xmmm32_sae)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x2E, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.SuppressAllExceptions, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ucomisd_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x2E, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vucomisd_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x2E, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vucomisd_xmm_xmmm64_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x2E, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.SuppressAllExceptions, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Comiss_xmm_xmmm32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x2F, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Comisd_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x2F, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcomiss_xmm_xmmm32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x2F, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcomisd_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x2F, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcomiss_xmm_xmmm32_sae)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x2F, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.SuppressAllExceptions, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcomisd_xmm_xmmm64_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x2F, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.SuppressAllExceptions, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Wrmsr)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x30, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rdtsc)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x31, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rdmsr)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x32, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rdpmc)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x33, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sysenter)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x34, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sysexitd)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x35, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sysexitq)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x35, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Getsec)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x37, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovo_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x40, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovo_r32_rm32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x40, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovo_r64_rm64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x40, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovno_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x41, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovno_r32_rm32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x41, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovno_r64_rm64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x41, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovb_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x42, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovb_r32_rm32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x42, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovb_r64_rm64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x42, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovae_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x43, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovae_r32_rm32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x43, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovae_r64_rm64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x43, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmove_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x44, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmove_r32_rm32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x44, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmove_r64_rm64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x44, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovne_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x45, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovne_r32_rm32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x45, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovne_r64_rm64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x45, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovbe_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x46, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovbe_r32_rm32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x46, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovbe_r64_rm64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x46, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmova_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x47, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmova_r32_rm32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x47, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmova_r64_rm64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x47, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovs_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x48, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovs_r32_rm32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x48, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovs_r64_rm64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x48, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovns_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x49, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovns_r32_rm32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x49, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovns_r64_rm64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x49, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovp_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x4A, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovp_r32_rm32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x4A, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovp_r64_rm64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x4A, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovnp_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x4B, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovnp_r32_rm32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x4B, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovnp_r64_rm64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x4B, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovl_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x4C, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovl_r32_rm32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x4C, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovl_r64_rm64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x4C, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovge_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x4D, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovge_r32_rm32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x4D, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovge_r64_rm64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x4D, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovle_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x4E, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovle_r32_rm32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x4E, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovle_r64_rm64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x4E, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovg_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x4F, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovg_r32_rm32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x4F, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmovg_r64_rm64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x4F, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kandw_k_k_k)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x41, -1, VexVectorLength.L1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VK, VexOpKind.HK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kandq_k_k_k)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x41, -1, VexVectorLength.L1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VK, VexOpKind.HK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kandb_k_k_k)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x41, -1, VexVectorLength.L1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VK, VexOpKind.HK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kandd_k_k_k)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x41, -1, VexVectorLength.L1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VK, VexOpKind.HK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kandnw_k_k_k)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x42, -1, VexVectorLength.L1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VK, VexOpKind.HK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kandnq_k_k_k)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x42, -1, VexVectorLength.L1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VK, VexOpKind.HK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kandnb_k_k_k)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x42, -1, VexVectorLength.L1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VK, VexOpKind.HK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kandnd_k_k_k)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x42, -1, VexVectorLength.L1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VK, VexOpKind.HK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Knotw_k_k)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x44, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Knotq_k_k)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x44, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Knotb_k_k)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x44, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Knotd_k_k)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x44, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Korw_k_k_k)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x45, -1, VexVectorLength.L1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VK, VexOpKind.HK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Korq_k_k_k)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x45, -1, VexVectorLength.L1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VK, VexOpKind.HK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Korb_k_k_k)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x45, -1, VexVectorLength.L1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VK, VexOpKind.HK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kord_k_k_k)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x45, -1, VexVectorLength.L1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VK, VexOpKind.HK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kxnorw_k_k_k)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x46, -1, VexVectorLength.L1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VK, VexOpKind.HK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kxnorq_k_k_k)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x46, -1, VexVectorLength.L1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VK, VexOpKind.HK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kxnorb_k_k_k)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x46, -1, VexVectorLength.L1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VK, VexOpKind.HK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kxnord_k_k_k)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x46, -1, VexVectorLength.L1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VK, VexOpKind.HK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kxorw_k_k_k)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x47, -1, VexVectorLength.L1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VK, VexOpKind.HK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kxorq_k_k_k)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x47, -1, VexVectorLength.L1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VK, VexOpKind.HK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kxorb_k_k_k)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x47, -1, VexVectorLength.L1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VK, VexOpKind.HK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kxord_k_k_k)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x47, -1, VexVectorLength.L1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VK, VexOpKind.HK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kaddw_k_k_k)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x4A, -1, VexVectorLength.L1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VK, VexOpKind.HK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kaddq_k_k_k)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x4A, -1, VexVectorLength.L1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VK, VexOpKind.HK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kaddb_k_k_k)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x4A, -1, VexVectorLength.L1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VK, VexOpKind.HK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kaddd_k_k_k)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x4A, -1, VexVectorLength.L1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VK, VexOpKind.HK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kunpckwd_k_k_k)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x4B, -1, VexVectorLength.L1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VK, VexOpKind.HK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kunpckdq_k_k_k)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x4B, -1, VexVectorLength.L1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VK, VexOpKind.HK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kunpckbw_k_k_k)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x4B, -1, VexVectorLength.L1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VK, VexOpKind.HK, VexOpKind.RK }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movmskps_r32_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x50, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.RX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movmskps_r64_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x50, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.RX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovmskps_r32_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x50, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { VexOpKind.Gd, VexOpKind.RX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovmskps_r64_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x50, -1, VexVectorLength.L128, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.Gq, VexOpKind.RX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovmskps_r32_ymm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x50, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { VexOpKind.Gd, VexOpKind.RY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovmskps_r64_ymm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x50, -1, VexVectorLength.L256, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.Gq, VexOpKind.RY }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movmskpd_r32_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x50, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.RX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movmskpd_r64_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x50, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.RX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovmskpd_r32_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x50, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { VexOpKind.Gd, VexOpKind.RX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovmskpd_r64_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x50, -1, VexVectorLength.L128, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.Gq, VexOpKind.RX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovmskpd_r32_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x50, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { VexOpKind.Gd, VexOpKind.RY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovmskpd_r64_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x50, -1, VexVectorLength.L256, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.Gq, VexOpKind.RY }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sqrtps_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x51, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vsqrtps_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x51, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vsqrtps_ymm_ymmm256)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x51, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vsqrtps_xmm_k1z_xmmm128b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x51, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vsqrtps_ymm_k1z_ymmm256b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x51, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vsqrtps_zmm_k1z_zmmm512b32_er)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x51, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sqrtpd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x51, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vsqrtpd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x51, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vsqrtpd_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x51, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vsqrtpd_xmm_k1z_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x51, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vsqrtpd_ymm_k1z_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x51, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vsqrtpd_zmm_k1z_zmmm512b64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x51, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sqrtss_xmm_xmmm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x51, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vsqrtss_xmm_xmm_xmmm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x51, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vsqrtss_xmm_k1z_xmm_xmmm32_er)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x51, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sqrtsd_xmm_xmmm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x51, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vsqrtsd_xmm_xmm_xmmm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x51, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vsqrtsd_xmm_k1z_xmm_xmmm64_er)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x51, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rsqrtps_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x52, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vrsqrtps_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x52, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vrsqrtps_ymm_ymmm256)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x52, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WY }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rsqrtss_xmm_xmmm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x52, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vrsqrtss_xmm_xmm_xmmm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x52, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rcpps_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x53, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vrcpps_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x53, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vrcpps_ymm_ymmm256)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x53, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WY }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rcpss_xmm_xmmm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x53, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vrcpss_xmm_xmm_xmmm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x53, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Andps_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x54, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vandps_xmm_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x54, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vandps_ymm_ymm_ymmm256)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x54, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vandps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x54, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vandps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x54, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vandps_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x54, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Andpd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x54, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vandpd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x54, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vandpd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x54, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vandpd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x54, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vandpd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x54, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vandpd_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x54, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Andnps_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x55, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vandnps_xmm_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x55, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vandnps_ymm_ymm_ymmm256)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x55, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vandnps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x55, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vandnps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x55, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vandnps_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x55, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Andnpd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x55, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vandnpd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x55, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vandnpd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x55, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vandnpd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x55, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vandnpd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x55, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vandnpd_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x55, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Orps_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x56, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vorps_xmm_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x56, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vorps_ymm_ymm_ymmm256)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x56, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vorps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x56, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vorps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x56, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vorps_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x56, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Orpd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x56, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vorpd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x56, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vorpd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x56, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vorpd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x56, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vorpd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x56, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vorpd_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x56, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xorps_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x57, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vxorps_xmm_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x57, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vxorps_ymm_ymm_ymmm256)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x57, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vxorps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x57, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vxorps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x57, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vxorps_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x57, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xorpd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x57, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vxorpd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x57, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vxorpd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x57, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vxorpd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x57, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vxorpd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x57, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vxorpd_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x57, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Addps_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x58, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vaddps_xmm_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x58, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vaddps_ymm_ymm_ymmm256)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x58, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vaddps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x58, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vaddps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x58, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vaddps_zmm_k1z_zmm_zmmm512b32_er)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x58, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Addpd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x58, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vaddpd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x58, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vaddpd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x58, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vaddpd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x58, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vaddpd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x58, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vaddpd_zmm_k1z_zmm_zmmm512b64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x58, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Addss_xmm_xmmm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x58, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vaddss_xmm_xmm_xmmm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x58, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vaddss_xmm_k1z_xmm_xmmm32_er)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x58, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Addsd_xmm_xmmm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x58, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vaddsd_xmm_xmm_xmmm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x58, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vaddsd_xmm_k1z_xmm_xmmm64_er)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x58, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mulps_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x59, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmulps_xmm_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x59, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmulps_ymm_ymm_ymmm256)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x59, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmulps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x59, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmulps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x59, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmulps_zmm_k1z_zmm_zmmm512b32_er)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x59, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mulpd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x59, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmulpd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x59, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmulpd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x59, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmulpd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x59, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmulpd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x59, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmulpd_zmm_k1z_zmm_zmmm512b64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x59, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mulss_xmm_xmmm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x59, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmulss_xmm_xmm_xmmm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x59, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmulss_xmm_k1z_xmm_xmmm32_er)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x59, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mulsd_xmm_xmmm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x59, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmulsd_xmm_xmm_xmmm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x59, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmulsd_xmm_k1z_xmm_xmmm64_er)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x59, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cvtps2pd_xmm_xmmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5A, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvtps2pd_xmm_xmmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5A, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvtps2pd_ymm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5A, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtps2pd_xmm_k1z_xmmm64b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5A, -1, EvexVectorLength.L128, TupleType.Half_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtps2pd_ymm_k1z_xmmm128b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5A, -1, EvexVectorLength.L256, TupleType.Half_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtps2pd_zmm_k1z_ymmm256b32_sae)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5A, -1, EvexVectorLength.L512, TupleType.Half_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WY }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cvtpd2ps_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5A, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvtpd2ps_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5A, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvtpd2ps_xmm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5A, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtpd2ps_xmm_k1z_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5A, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtpd2ps_xmm_k1z_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5A, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtpd2ps_ymm_k1z_zmmm512b64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5A, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cvtss2sd_xmm_xmmm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x5A, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvtss2sd_xmm_xmm_xmmm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x5A, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtss2sd_xmm_k1z_xmm_xmmm32_sae)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x5A, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cvtsd2ss_xmm_xmmm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x5A, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvtsd2ss_xmm_xmm_xmmm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x5A, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtsd2ss_xmm_k1z_xmm_xmmm64_er)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x5A, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cvtdq2ps_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5B, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvtdq2ps_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5B, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvtdq2ps_ymm_ymmm256)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5B, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtdq2ps_xmm_k1z_xmmm128b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5B, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtdq2ps_ymm_k1z_ymmm256b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5B, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtdq2ps_zmm_k1z_zmmm512b32_er)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5B, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtqq2ps_xmm_k1z_xmmm128b64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5B, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtqq2ps_xmm_k1z_ymmm256b64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5B, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtqq2ps_ymm_k1z_zmmm512b64_er)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5B, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cvtps2dq_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5B, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvtps2dq_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5B, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvtps2dq_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5B, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtps2dq_xmm_k1z_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5B, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtps2dq_ymm_k1z_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5B, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtps2dq_zmm_k1z_zmmm512b32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5B, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cvttps2dq_xmm_xmmm128)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x5B, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvttps2dq_xmm_xmmm128)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x5B, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvttps2dq_ymm_ymmm256)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x5B, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvttps2dq_xmm_k1z_xmmm128b32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x5B, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvttps2dq_ymm_k1z_ymmm256b32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x5B, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvttps2dq_zmm_k1z_zmmm512b32_sae)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x5B, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Subps_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5C, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vsubps_xmm_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5C, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vsubps_ymm_ymm_ymmm256)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5C, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vsubps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5C, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vsubps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5C, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vsubps_zmm_k1z_zmm_zmmm512b32_er)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5C, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Subpd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5C, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vsubpd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5C, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vsubpd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5C, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vsubpd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5C, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vsubpd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5C, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vsubpd_zmm_k1z_zmm_zmmm512b64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5C, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Subss_xmm_xmmm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x5C, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vsubss_xmm_xmm_xmmm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x5C, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vsubss_xmm_k1z_xmm_xmmm32_er)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x5C, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Subsd_xmm_xmmm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x5C, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vsubsd_xmm_xmm_xmmm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x5C, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vsubsd_xmm_k1z_xmm_xmmm64_er)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x5C, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Minps_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5D, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vminps_xmm_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5D, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vminps_ymm_ymm_ymmm256)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5D, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vminps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5D, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vminps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5D, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vminps_zmm_k1z_zmm_zmmm512b32_sae)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5D, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Minpd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5D, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vminpd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5D, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vminpd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5D, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vminpd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5D, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vminpd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5D, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vminpd_zmm_k1z_zmm_zmmm512b64_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5D, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Minss_xmm_xmmm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x5D, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vminss_xmm_xmm_xmmm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x5D, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vminss_xmm_k1z_xmm_xmmm32_sae)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x5D, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Minsd_xmm_xmmm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x5D, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vminsd_xmm_xmm_xmmm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x5D, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vminsd_xmm_k1z_xmm_xmmm64_sae)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x5D, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Divps_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5E, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vdivps_xmm_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5E, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vdivps_ymm_ymm_ymmm256)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5E, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vdivps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5E, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vdivps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5E, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vdivps_zmm_k1z_zmm_zmmm512b32_er)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5E, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Divpd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5E, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vdivpd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5E, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vdivpd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5E, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vdivpd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5E, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vdivpd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5E, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vdivpd_zmm_k1z_zmm_zmmm512b64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5E, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Divss_xmm_xmmm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x5E, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vdivss_xmm_xmm_xmmm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x5E, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vdivss_xmm_k1z_xmm_xmmm32_er)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x5E, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Divsd_xmm_xmmm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x5E, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vdivsd_xmm_xmm_xmmm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x5E, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vdivsd_xmm_k1z_xmm_xmmm64_er)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x5E, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Maxps_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5F, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmaxps_xmm_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5F, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmaxps_ymm_ymm_ymmm256)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5F, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmaxps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5F, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmaxps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5F, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmaxps_zmm_k1z_zmm_zmmm512b32_sae)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x5F, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Maxpd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5F, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmaxpd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5F, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmaxpd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5F, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmaxpd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5F, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmaxpd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5F, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmaxpd_zmm_k1z_zmm_zmmm512b64_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x5F, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Maxss_xmm_xmmm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x5F, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmaxss_xmm_xmm_xmmm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x5F, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmaxss_xmm_k1z_xmm_xmmm32_sae)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x5F, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Maxsd_xmm_xmmm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x5F, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmaxsd_xmm_xmm_xmmm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x5F, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmaxsd_xmm_k1z_xmm_xmmm64_sae)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x5F, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Punpcklbw_mm_mmm32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x60, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Punpcklbw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x60, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpunpcklbw_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x60, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpunpcklbw_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x60, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpunpcklbw_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x60, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpunpcklbw_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x60, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpunpcklbw_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x60, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Punpcklwd_mm_mmm32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x61, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Punpcklwd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x61, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpunpcklwd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x61, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpunpcklwd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x61, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpunpcklwd_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x61, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpunpcklwd_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x61, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpunpcklwd_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x61, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Punpckldq_mm_mmm32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x62, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Punpckldq_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x62, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpunpckldq_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x62, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpunpckldq_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x62, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpunpckldq_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x62, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpunpckldq_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x62, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpunpckldq_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x62, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Packsswb_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x63, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Packsswb_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x63, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpacksswb_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x63, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpacksswb_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x63, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpacksswb_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x63, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpacksswb_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x63, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpacksswb_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x63, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pcmpgtb_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x64, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pcmpgtb_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x64, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpcmpgtb_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x64, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpcmpgtb_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x64, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpgtb_k_k1_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x64, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpgtb_k_k1_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x64, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpgtb_k_k1_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x64, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pcmpgtw_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x65, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pcmpgtw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x65, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpcmpgtw_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x65, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpcmpgtw_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x65, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpgtw_k_k1_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x65, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpgtw_k_k1_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x65, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpgtw_k_k1_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x65, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pcmpgtd_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x66, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pcmpgtd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x66, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpcmpgtd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x66, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpcmpgtd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x66, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpgtd_k_k1_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x66, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpgtd_k_k1_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x66, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpgtd_k_k1_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x66, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Packuswb_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x67, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Packuswb_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x67, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpackuswb_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x67, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpackuswb_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x67, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpackuswb_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x67, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpackuswb_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x67, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpackuswb_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x67, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Punpckhbw_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x68, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Punpckhbw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x68, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpunpckhbw_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x68, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpunpckhbw_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x68, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpunpckhbw_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x68, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpunpckhbw_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x68, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpunpckhbw_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x68, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Punpckhwd_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x69, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Punpckhwd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x69, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpunpckhwd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x69, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpunpckhwd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x69, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpunpckhwd_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x69, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpunpckhwd_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x69, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpunpckhwd_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x69, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Punpckhdq_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x6A, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Punpckhdq_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6A, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpunpckhdq_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6A, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpunpckhdq_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6A, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpunpckhdq_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6A, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpunpckhdq_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6A, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpunpckhdq_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6A, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Packssdw_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x6B, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Packssdw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6B, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpackssdw_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6B, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpackssdw_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6B, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpackssdw_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6B, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpackssdw_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6B, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpackssdw_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6B, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Punpcklqdq_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6C, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpunpcklqdq_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6C, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpunpcklqdq_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6C, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpunpcklqdq_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6C, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpunpcklqdq_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6C, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpunpcklqdq_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6C, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Punpckhqdq_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6D, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpunpckhqdq_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6D, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpunpckhqdq_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6D, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpunpckhqdq_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6D, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpunpckhqdq_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6D, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpunpckhqdq_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6D, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movd_mm_rm32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x6E, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movq_mm_rm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x6E, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movd_xmm_rm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6E, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movq_xmm_rm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6E, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.Eq }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovd_xmm_rm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6E, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { VexOpKind.VX, VexOpKind.Ed }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovq_xmm_rm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6E, -1, VexVectorLength.L128, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.Eq }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovd_xmm_rm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6E, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { EvexOpKind.VX, EvexOpKind.Ed }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovq_xmm_rm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6E, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.VX, EvexOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movq_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x6F, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movdqa_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6F, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovdqa_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6F, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovdqa_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6F, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqa32_xmm_k1z_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6F, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqa32_ymm_k1z_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6F, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqa32_zmm_k1z_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6F, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqa64_xmm_k1z_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6F, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqa64_ymm_k1z_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6F, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqa64_zmm_k1z_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x6F, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movdqu_xmm_xmmm128)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x6F, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovdqu_xmm_xmmm128)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x6F, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovdqu_ymm_ymmm256)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x6F, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqu32_xmm_k1z_xmmm128)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x6F, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqu32_ymm_k1z_ymmm256)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x6F, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqu32_zmm_k1z_zmmm512)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x6F, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqu64_xmm_k1z_xmmm128)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x6F, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqu64_ymm_k1z_ymmm256)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x6F, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqu64_zmm_k1z_zmmm512)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x6F, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqu8_xmm_k1z_xmmm128)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x6F, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqu8_ymm_k1z_ymmm256)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x6F, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqu8_zmm_k1z_zmmm512)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x6F, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqu16_xmm_k1z_xmmm128)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x6F, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqu16_ymm_k1z_ymmm256)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x6F, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqu16_zmm_k1z_zmmm512)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x6F, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pshufw_mm_mmm64_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x70, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pshufd_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x70, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpshufd_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x70, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpshufd_ymm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x70, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WY, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshufd_xmm_k1z_xmmm128b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x70, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshufd_ymm_k1z_ymmm256b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x70, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshufd_zmm_k1z_zmmm512b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x70, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pshufhw_xmm_xmmm128_imm8)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x70, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpshufhw_xmm_xmmm128_imm8)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x70, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpshufhw_ymm_ymmm256_imm8)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x70, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WY, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshufhw_xmm_k1z_xmmm128_imm8)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x70, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshufhw_ymm_k1z_ymmm256_imm8)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x70, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshufhw_zmm_k1z_zmmm512_imm8)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x70, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pshuflw_xmm_xmmm128_imm8)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x70, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpshuflw_xmm_xmmm128_imm8)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x70, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpshuflw_ymm_ymmm256_imm8)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x70, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WY, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshuflw_xmm_k1z_xmmm128_imm8)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x70, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshuflw_ymm_k1z_ymmm256_imm8)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x70, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshuflw_zmm_k1z_zmmm512_imm8)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x70, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psrlw_mm_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x71, 2, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.N, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psrlw_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x71, 2, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.RX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsrlw_xmm_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x71, 2, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.HX, VexOpKind.RX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsrlw_ymm_ymm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x71, 2, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.HY, VexOpKind.RY, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrlw_xmm_k1z_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x71, 2, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrlw_ymm_k1z_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x71, 2, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrlw_zmm_k1z_zmmm512_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x71, 2, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psraw_mm_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x71, 4, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.N, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psraw_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x71, 4, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.RX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsraw_xmm_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x71, 4, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.HX, VexOpKind.RX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsraw_ymm_ymm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x71, 4, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.HY, VexOpKind.RY, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsraw_xmm_k1z_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x71, 4, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsraw_ymm_k1z_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x71, 4, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsraw_zmm_k1z_zmmm512_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x71, 4, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psllw_mm_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x71, 6, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.N, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psllw_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x71, 6, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.RX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsllw_xmm_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x71, 6, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.HX, VexOpKind.RX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsllw_ymm_ymm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x71, 6, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.HY, VexOpKind.RY, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsllw_xmm_k1z_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x71, 6, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsllw_ymm_k1z_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x71, 6, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsllw_zmm_k1z_zmmm512_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x71, 6, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vprord_xmm_k1z_xmmm128b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x72, 0, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vprord_ymm_k1z_ymmm256b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x72, 0, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vprord_zmm_k1z_zmmm512b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x72, 0, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vprorq_xmm_k1z_xmmm128b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x72, 0, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vprorq_ymm_k1z_ymmm256b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x72, 0, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vprorq_zmm_k1z_zmmm512b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x72, 0, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vprold_xmm_k1z_xmmm128b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x72, 1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vprold_ymm_k1z_ymmm256b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x72, 1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vprold_zmm_k1z_zmmm512b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x72, 1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vprolq_xmm_k1z_xmmm128b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x72, 1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vprolq_ymm_k1z_ymmm256b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x72, 1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vprolq_zmm_k1z_zmmm512b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x72, 1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psrld_mm_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x72, 2, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.N, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psrld_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x72, 2, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.RX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsrld_xmm_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x72, 2, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.HX, VexOpKind.RX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsrld_ymm_ymm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x72, 2, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.HY, VexOpKind.RY, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrld_xmm_k1z_xmmm128b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x72, 2, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrld_ymm_k1z_ymmm256b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x72, 2, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrld_zmm_k1z_zmmm512b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x72, 2, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psrad_mm_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x72, 4, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.N, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psrad_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x72, 4, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.RX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsrad_xmm_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x72, 4, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.HX, VexOpKind.RX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsrad_ymm_ymm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x72, 4, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.HY, VexOpKind.RY, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrad_xmm_k1z_xmmm128b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x72, 4, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrad_ymm_k1z_ymmm256b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x72, 4, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrad_zmm_k1z_zmmm512b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x72, 4, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsraq_xmm_k1z_xmmm128b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x72, 4, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsraq_ymm_k1z_ymmm256b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x72, 4, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsraq_zmm_k1z_zmmm512b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x72, 4, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pslld_mm_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x72, 6, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.N, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pslld_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x72, 6, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.RX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpslld_xmm_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x72, 6, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.HX, VexOpKind.RX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpslld_ymm_ymm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x72, 6, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.HY, VexOpKind.RY, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpslld_xmm_k1z_xmmm128b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x72, 6, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpslld_ymm_k1z_ymmm256b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x72, 6, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpslld_zmm_k1z_zmmm512b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x72, 6, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psrlq_mm_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x73, 2, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.N, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psrlq_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x73, 2, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.RX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsrlq_xmm_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x73, 2, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.HX, VexOpKind.RX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsrlq_ymm_ymm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x73, 2, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.HY, VexOpKind.RY, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrlq_xmm_k1z_xmmm128b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x73, 2, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrlq_ymm_k1z_ymmm256b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x73, 2, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrlq_zmm_k1z_zmmm512b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x73, 2, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psrldq_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x73, 3, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.RX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsrldq_xmm_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x73, 3, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.HX, VexOpKind.RX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsrldq_ymm_ymm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x73, 3, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.HY, VexOpKind.RY, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrldq_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x73, 3, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrldq_ymm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x73, 3, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrldq_zmm_zmmm512_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x73, 3, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psllq_mm_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x73, 6, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.N, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psllq_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x73, 6, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.RX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsllq_xmm_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x73, 6, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.HX, VexOpKind.RX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsllq_ymm_ymm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x73, 6, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.HY, VexOpKind.RY, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsllq_xmm_k1z_xmmm128b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x73, 6, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsllq_ymm_k1z_ymmm256b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x73, 6, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsllq_zmm_k1z_zmmm512b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x73, 6, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pslldq_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x73, 7, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.RX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpslldq_xmm_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x73, 7, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.HX, VexOpKind.RX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpslldq_ymm_ymm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x73, 7, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.HY, VexOpKind.RY, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpslldq_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x73, 7, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpslldq_ymm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x73, 7, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpslldq_zmm_zmmm512_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x73, 7, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pcmpeqb_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x74, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pcmpeqb_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x74, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpcmpeqb_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x74, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpcmpeqb_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x74, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpeqb_k_k1_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x74, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpeqb_k_k1_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x74, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpeqb_k_k1_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x74, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pcmpeqw_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x75, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pcmpeqw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x75, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpcmpeqw_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x75, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpcmpeqw_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x75, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpeqw_k_k1_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x75, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpeqw_k_k1_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x75, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpeqw_k_k1_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x75, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pcmpeqd_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x76, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pcmpeqd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x76, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpcmpeqd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x76, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpcmpeqd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x76, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpeqd_k_k1_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x76, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpeqd_k_k1_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x76, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpeqd_k_k1_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x76, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Emms)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x77, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vzeroupper)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x77, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, Array.Empty<VexOpKind>()),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vzeroall)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x77, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, Array.Empty<VexOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Vmread_rm32_r32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x78, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Vmread_rm64_r64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x78, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Gq }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvttps2udq_xmm_k1z_xmmm128b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x78, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvttps2udq_ymm_k1z_ymmm256b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x78, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvttps2udq_zmm_k1z_zmmm512b32_sae)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x78, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvttpd2udq_xmm_k1z_xmmm128b64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x78, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvttpd2udq_xmm_k1z_ymmm256b64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x78, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvttpd2udq_ymm_k1z_zmmm512b64_sae)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x78, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Extrq_xmm_imm8_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x78, 0, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.RX, LegacyOpKind.Ib, LegacyOpKind.Ib11 }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvttps2uqq_xmm_k1z_xmmm64b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x78, -1, EvexVectorLength.L128, TupleType.Half_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvttps2uqq_ymm_k1z_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x78, -1, EvexVectorLength.L256, TupleType.Half_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvttps2uqq_zmm_k1z_ymmm256b32_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x78, -1, EvexVectorLength.L512, TupleType.Half_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvttpd2uqq_xmm_k1z_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x78, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvttpd2uqq_ymm_k1z_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x78, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvttpd2uqq_zmm_k1z_zmmm512b64_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x78, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvttss2usi_r32_xmmm32_sae)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x78, -1, EvexVectorLength.L128, TupleType.Tuple1_Fixed_4, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG32 | OpCodeFlags.SuppressAllExceptions, new[] { EvexOpKind.Gd, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvttss2usi_r64_xmmm32_sae)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x78, -1, EvexVectorLength.L128, TupleType.Tuple1_Fixed_4, OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.SuppressAllExceptions, new[] { EvexOpKind.Gq, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Insertq_xmm_xmm_imm8_imm8)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x78, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.RX, LegacyOpKind.Ib, LegacyOpKind.Ib11 }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvttsd2usi_r32_xmmm64_sae)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x78, -1, EvexVectorLength.L128, TupleType.Tuple1_Fixed_8, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG32 | OpCodeFlags.SuppressAllExceptions, new[] { EvexOpKind.Gd, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvttsd2usi_r64_xmmm64_sae)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x78, -1, EvexVectorLength.L128, TupleType.Tuple1_Fixed_8, OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.SuppressAllExceptions, new[] { EvexOpKind.Gq, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Vmwrite_r32_rm32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x79, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Vmwrite_r64_rm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x79, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtps2udq_xmm_k1z_xmmm128b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x79, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtps2udq_ymm_k1z_ymmm256b32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x79, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtps2udq_zmm_k1z_zmmm512b32_er)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x79, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtpd2udq_xmm_k1z_xmmm128b64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x79, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtpd2udq_xmm_k1z_ymmm256b64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x79, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtpd2udq_ymm_k1z_zmmm512b64_er)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x79, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Extrq_xmm_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x79, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.RX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtps2uqq_xmm_k1z_xmmm64b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x79, -1, EvexVectorLength.L128, TupleType.Half_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtps2uqq_ymm_k1z_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x79, -1, EvexVectorLength.L256, TupleType.Half_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtps2uqq_zmm_k1z_ymmm256b32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x79, -1, EvexVectorLength.L512, TupleType.Half_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtpd2uqq_xmm_k1z_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x79, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtpd2uqq_ymm_k1z_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x79, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtpd2uqq_zmm_k1z_zmmm512b64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x79, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtss2usi_r32_xmmm32_er)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x79, -1, EvexVectorLength.L128, TupleType.Tuple1_Fixed_4, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG32 | OpCodeFlags.RoundingControl, new[] { EvexOpKind.Gd, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtss2usi_r64_xmmm32_er)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x79, -1, EvexVectorLength.L128, TupleType.Tuple1_Fixed_4, OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.RoundingControl, new[] { EvexOpKind.Gq, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Insertq_xmm_xmm)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x79, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.RX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtsd2usi_r32_xmmm64_er)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x79, -1, EvexVectorLength.L128, TupleType.Tuple1_Fixed_8, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG32 | OpCodeFlags.RoundingControl, new[] { EvexOpKind.Gd, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtsd2usi_r64_xmmm64_er)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x79, -1, EvexVectorLength.L128, TupleType.Tuple1_Fixed_8, OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.RoundingControl, new[] { EvexOpKind.Gq, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvttps2qq_xmm_k1z_xmmm64b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x7A, -1, EvexVectorLength.L128, TupleType.Half_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvttps2qq_ymm_k1z_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x7A, -1, EvexVectorLength.L256, TupleType.Half_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvttps2qq_zmm_k1z_ymmm256b32_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x7A, -1, EvexVectorLength.L512, TupleType.Half_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvttpd2qq_xmm_k1z_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x7A, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvttpd2qq_ymm_k1z_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x7A, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvttpd2qq_zmm_k1z_zmmm512b64_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x7A, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtudq2pd_xmm_k1z_xmmm64b32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x7A, -1, EvexVectorLength.L128, TupleType.Half_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtudq2pd_ymm_k1z_xmmm128b32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x7A, -1, EvexVectorLength.L256, TupleType.Half_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtudq2pd_zmm_k1z_ymmm256b32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x7A, -1, EvexVectorLength.L512, TupleType.Half_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtuqq2pd_xmm_k1z_xmmm128b64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x7A, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtuqq2pd_ymm_k1z_ymmm256b64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x7A, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtuqq2pd_zmm_k1z_zmmm512b64_er)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x7A, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtudq2ps_xmm_k1z_xmmm128b32)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x7A, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtudq2ps_ymm_k1z_ymmm256b32)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x7A, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtudq2ps_zmm_k1z_zmmm512b32_er)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x7A, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtuqq2ps_xmm_k1z_xmmm128b64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x7A, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtuqq2ps_xmm_k1z_ymmm256b64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x7A, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtuqq2ps_ymm_k1z_zmmm512b64_er)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x7A, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtps2qq_xmm_k1z_xmmm64b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x7B, -1, EvexVectorLength.L128, TupleType.Half_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtps2qq_ymm_k1z_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x7B, -1, EvexVectorLength.L256, TupleType.Half_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtps2qq_zmm_k1z_ymmm256b32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x7B, -1, EvexVectorLength.L512, TupleType.Half_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtpd2qq_xmm_k1z_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x7B, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtpd2qq_ymm_k1z_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x7B, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtpd2qq_zmm_k1z_zmmm512b64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x7B, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtusi2ss_xmm_xmm_rm32_er)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x7B, -1, EvexVectorLength.L128, TupleType.Tuple1_Fixed_4, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG32 | OpCodeFlags.RoundingControl, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.Ed }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtusi2ss_xmm_xmm_rm64_er)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x7B, -1, EvexVectorLength.L128, TupleType.Tuple1_Fixed_8, OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.RoundingControl, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.Eq }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtusi2sd_xmm_xmm_rm32_er)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x7B, -1, EvexVectorLength.L128, TupleType.Tuple1_Fixed_4, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG32 | OpCodeFlags.RoundingControl, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.Ed }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtusi2sd_xmm_xmm_rm64_er)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x7B, -1, EvexVectorLength.L128, TupleType.Tuple1_Fixed_8, OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.RoundingControl, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Haddpd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x7C, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vhaddpd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x7C, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vhaddpd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x7C, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Haddps_xmm_xmmm128)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x7C, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vhaddps_xmm_xmm_xmmm128)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x7C, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vhaddps_ymm_ymm_ymmm256)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x7C, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Hsubpd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x7D, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vhsubpd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x7D, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vhsubpd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x7D, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Hsubps_xmm_xmmm128)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x7D, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vhsubps_xmm_xmm_xmmm128)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x7D, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vhsubps_ymm_ymm_ymmm256)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x7D, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movd_rm32_mm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x7E, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.P }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movq_rm64_mm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x7E, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.P }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movd_rm32_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x7E, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.VX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movq_rm64_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x7E, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.VX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovd_rm32_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x7E, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { VexOpKind.Ed, VexOpKind.VX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovq_rm64_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x7E, -1, VexVectorLength.L128, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.Eq, VexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovd_rm32_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x7E, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { EvexOpKind.Ed, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovq_rm64_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x7E, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.Eq, EvexOpKind.VX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movq_xmm_xmmm64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x7E, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovq_xmm_xmmm64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x7E, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovq_xmm_xmmm64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x7E, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movq_mmm64_mm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x7F, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Q, LegacyOpKind.P }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movdqa_xmmm128_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x7F, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.WX, LegacyOpKind.VX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovdqa_xmmm128_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x7F, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.WX, VexOpKind.VX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovdqa_ymmm256_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x7F, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.WY, VexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqa32_xmmm128_k1z_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x7F, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqa32_ymmm256_k1z_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x7F, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WY, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqa32_zmmm512_k1z_zmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x7F, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WZ, EvexOpKind.VZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqa64_xmmm128_k1z_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x7F, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqa64_ymmm256_k1z_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x7F, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WY, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqa64_zmmm512_k1z_zmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x7F, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WZ, EvexOpKind.VZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movdqu_xmmm128_xmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x7F, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.WX, LegacyOpKind.VX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovdqu_xmmm128_xmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x7F, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.WX, VexOpKind.VX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovdqu_ymmm256_ymm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x7F, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.WY, VexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqu32_xmmm128_k1z_xmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x7F, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqu32_ymmm256_k1z_ymm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x7F, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WY, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqu32_zmmm512_k1z_zmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x7F, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WZ, EvexOpKind.VZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqu64_xmmm128_k1z_xmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x7F, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqu64_ymmm256_k1z_ymm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x7F, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WY, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqu64_zmmm512_k1z_zmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0x7F, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WZ, EvexOpKind.VZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqu8_xmmm128_k1z_xmm)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x7F, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqu8_ymmm256_k1z_ymm)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x7F, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WY, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqu8_zmmm512_k1z_zmm)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x7F, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WZ, EvexOpKind.VZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqu16_xmmm128_k1z_xmm)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x7F, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqu16_ymmm256_k1z_ymm)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x7F, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WY, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovdqu16_zmmm512_k1z_zmm)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x7F, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WZ, EvexOpKind.VZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jo_rel16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x80, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.Jw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jo_rel32_32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x80, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.dJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jo_rel32_64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x80, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.qJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jno_rel16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x81, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.Jw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jno_rel32_32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x81, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.dJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jno_rel32_64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x81, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.qJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jb_rel16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x82, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.Jw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jb_rel32_32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x82, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.dJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jb_rel32_64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x82, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.qJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jae_rel16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x83, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.Jw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jae_rel32_32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x83, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.dJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jae_rel32_64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x83, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.qJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Je_rel16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x84, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.Jw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Je_rel32_32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x84, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.dJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Je_rel32_64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x84, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.qJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jne_rel16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x85, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.Jw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jne_rel32_32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x85, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.dJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jne_rel32_64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x85, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.qJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jbe_rel16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x86, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.Jw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jbe_rel32_32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x86, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.dJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jbe_rel32_64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x86, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.qJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ja_rel16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x87, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.Jw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ja_rel32_32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x87, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.dJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ja_rel32_64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x87, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.qJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Js_rel16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x88, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.Jw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Js_rel32_32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x88, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.dJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Js_rel32_64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x88, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.qJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jns_rel16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x89, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.Jw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jns_rel32_32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x89, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.dJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jns_rel32_64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x89, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.qJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jp_rel16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x8A, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.Jw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jp_rel32_32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x8A, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.dJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jp_rel32_64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x8A, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.qJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jnp_rel16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x8B, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.Jw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jnp_rel32_32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x8B, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.dJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jnp_rel32_64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x8B, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.qJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jl_rel16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x8C, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.Jw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jl_rel32_32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x8C, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.dJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jl_rel32_64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x8C, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.qJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jge_rel16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x8D, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.Jw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jge_rel32_32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x8D, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.dJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jge_rel32_64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x8D, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.qJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jle_rel16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x8E, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.Jw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jle_rel32_32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x8E, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.dJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jle_rel32_64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x8E, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.qJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jg_rel16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x8F, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.Jw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jg_rel32_32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x8F, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.dJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jg_rel32_64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x8F, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.BndPrefix | OpCodeFlags.HintTakenPrefix, new[] { LegacyOpKind.qJd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Seto_rm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x90, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Setno_rm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x91, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Setb_rm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x92, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Setae_rm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x93, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sete_rm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x94, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Setne_rm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x95, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Setbe_rm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x96, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Seta_rm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x97, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sets_rm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x98, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Setns_rm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x99, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Setp_rm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x9A, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Setnp_rm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x9B, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Setl_rm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x9C, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Setge_rm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x9D, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Setle_rm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x9E, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Setg_rm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0x9F, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Eb }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kmovw_k_km16)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x90, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VK, VexOpKind.WK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kmovq_k_km64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x90, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VK, VexOpKind.WK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kmovb_k_km8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x90, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VK, VexOpKind.WK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kmovd_k_km32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x90, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VK, VexOpKind.WK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kmovw_m16_k)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x91, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.MK, VexOpKind.VK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kmovq_m64_k)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x91, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.MK, VexOpKind.VK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kmovb_m8_k)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x91, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.MK, VexOpKind.VK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kmovd_m32_k)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x91, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.MK, VexOpKind.VK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kmovw_k_r32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x92, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VK, VexOpKind.Rd }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kmovb_k_r32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x92, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VK, VexOpKind.Rd }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kmovd_k_r32)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x92, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VK, VexOpKind.Rd }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kmovq_k_r64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x92, -1, VexVectorLength.L0, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VK, VexOpKind.Rq }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kmovw_r32_k)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x93, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.Gd, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kmovb_r32_k)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x93, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.Gd, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kmovd_r32_k)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x93, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.Gd, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kmovq_r64_k)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0x93, -1, VexVectorLength.L0, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.Gq, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kortestw_k_k)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x98, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kortestq_k_k)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x98, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kortestb_k_k)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x98, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kortestd_k_k)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x98, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Ktestw_k_k)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x99, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Ktestq_k_k)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0x99, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Ktestb_k_k)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x99, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VK, VexOpKind.RK }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Ktestd_k_k)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0x99, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VK, VexOpKind.RK }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pushw_FS)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA0, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.FS }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pushd_FS)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA0, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.FS }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pushq_FS)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.FS }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Popw_FS)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA1, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.FS }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Popd_FS)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA1, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.FS }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Popq_FS)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA1, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.FS }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cpuid)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA2, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bt_rm16_r16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA3, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bt_rm32_r32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA3, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bt_rm64_r64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA3, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Gq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shld_rm16_r16_imm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA4, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shld_rm32_r32_imm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA4, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shld_rm64_r64_imm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA4, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Gq, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shld_rm16_r16_CL)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA5, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shld_rm32_r32_CL)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA5, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shld_rm64_r64_CL)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA5, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Gq, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Montmul_16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA6C0, -1, OperandSize.None, AddressSize.Size16, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.RepPrefix, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Montmul_32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA6C0, -1, OperandSize.None, AddressSize.Size32, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Montmul_64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA6C0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xsha1_16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA6C8, -1, OperandSize.None, AddressSize.Size16, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.RepPrefix, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xsha1_32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA6C8, -1, OperandSize.None, AddressSize.Size32, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xsha1_64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA6C8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xsha256_16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA6D0, -1, OperandSize.None, AddressSize.Size16, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.RepPrefix, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xsha256_32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA6D0, -1, OperandSize.None, AddressSize.Size32, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xsha256_64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA6D0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xbts_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA6, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xbts_r32_rm32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA6, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xstore_16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA7C0, -1, OperandSize.None, AddressSize.Size16, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.RepPrefix, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xstore_32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA7C0, -1, OperandSize.None, AddressSize.Size32, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xstore_64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA7C0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.XcryptEcb_16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA7C8, -1, OperandSize.None, AddressSize.Size16, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.RepPrefix, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.XcryptEcb_32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA7C8, -1, OperandSize.None, AddressSize.Size32, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.XcryptEcb_64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA7C8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.XcryptCbc_16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA7D0, -1, OperandSize.None, AddressSize.Size16, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.RepPrefix, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.XcryptCbc_32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA7D0, -1, OperandSize.None, AddressSize.Size32, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.XcryptCbc_64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA7D0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.XcryptCtr_16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA7D8, -1, OperandSize.None, AddressSize.Size16, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.RepPrefix, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.XcryptCtr_32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA7D8, -1, OperandSize.None, AddressSize.Size32, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.XcryptCtr_64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA7D8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.XcryptCfb_16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA7E0, -1, OperandSize.None, AddressSize.Size16, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.RepPrefix, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.XcryptCfb_32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA7E0, -1, OperandSize.None, AddressSize.Size32, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.XcryptCfb_64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA7E0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.XcryptOfb_16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA7E8, -1, OperandSize.None, AddressSize.Size16, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.RepPrefix, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.XcryptOfb_32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA7E8, -1, OperandSize.None, AddressSize.Size32, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.XcryptOfb_64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA7E8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.RepPrefix, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ibts_rm16_r16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA7, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ibts_rm32_r32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA7, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmpxchg486_rm8_r8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA6, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Eb, LegacyOpKind.Gb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmpxchg486_rm16_r16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA7, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmpxchg486_rm32_r32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA7, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pushw_GS)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA8, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.GS }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pushd_GS)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA8, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.GS }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pushq_GS)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.GS }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Popw_GS)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA9, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.GS }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Popd_GS)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA9, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.GS }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Popq_GS)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xA9, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.GS }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rsm)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xAA, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bts_rm16_r16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xAB, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bts_rm32_r32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xAB, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bts_rm64_r64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xAB, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq, LegacyOpKind.Gq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shrd_rm16_r16_imm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xAC, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shrd_rm32_r32_imm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xAC, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shrd_rm64_r64_imm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xAC, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Gq, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shrd_rm16_r16_CL)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xAD, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shrd_rm32_r32_CL)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xAD, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shrd_rm64_r64_CL)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xAD, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Gq, LegacyOpKind.CL }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fxsave_m512byte)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAE, 0, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fxsave64_m512byte)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAE, 0, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rdfsbase_r32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xAE, 0, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Rd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rdfsbase_r64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xAE, 0, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Rq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fxrstor_m512byte)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAE, 1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Fxrstor64_m512byte)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAE, 1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rdgsbase_r32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xAE, 1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Rd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rdgsbase_r64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xAE, 1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Rq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ldmxcsr_m32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAE, 2, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Md }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Wrfsbase_r32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xAE, 2, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Rd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Wrfsbase_r64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xAE, 2, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Rq }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vldmxcsr_m32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAE, 2, VexVectorLength.LZ, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.Md }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Stmxcsr_m32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAE, 3, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Md }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Wrgsbase_r32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xAE, 3, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Rd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Wrgsbase_r64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xAE, 3, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Rq }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vstmxcsr_m32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAE, 3, VexVectorLength.LZ, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.Md }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xsave_mem)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAE, 4, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xsave64_mem)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAE, 4, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ptwrite_rm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xAE, 4, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ptwrite_rm64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xAE, 4, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xrstor_mem)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAE, 5, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xrstor64_mem)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAE, 5, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Incsspd_r32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xAE, 5, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Rd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Incsspq_r64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xAE, 5, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Rq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xsaveopt_mem)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAE, 6, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xsaveopt64_mem)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAE, 6, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Clwb_m8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xAE, 6, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Tpause_r32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xAE, 6, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Rd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Tpause_r64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xAE, 6, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Rq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Clrssbsy_m64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xAE, 6, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Umonitor_r16)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xAE, 6, OperandSize.None, AddressSize.Size16, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Rw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Umonitor_r32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xAE, 6, OperandSize.None, AddressSize.Size32, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Rd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Umonitor_r64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xAE, 6, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Rq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Umwait_r32)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0xAE, 6, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Rd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Umwait_r64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0xAE, 6, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Rq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Clflush_m8)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAE, 7, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Clflushopt_m8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xAE, 7, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lfence)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAEE8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lfence_E9)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAEE9, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lfence_EA)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAEEA, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lfence_EB)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAEEB, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lfence_EC)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAEEC, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lfence_ED)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAEED, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lfence_EE)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAEEE, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lfence_EF)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAEEF, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mfence)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAEF0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mfence_F1)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAEF1, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mfence_F2)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAEF2, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mfence_F3)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAEF3, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mfence_F4)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAEF4, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mfence_F5)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAEF5, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mfence_F6)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAEF6, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mfence_F7)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAEF7, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sfence)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAEF8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sfence_F9)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAEF9, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sfence_FA)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAEFA, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sfence_FB)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAEFB, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sfence_FC)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAEFC, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sfence_FD)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAEFD, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sfence_FE)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAEFE, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sfence_FF)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xAEFF, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pcommit)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xAEF8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, Array.Empty<LegacyOpKind>()),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Imul_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xAF, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Imul_r32_rm32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xAF, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Imul_r64_rm64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xAF, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmpxchg_rm8_r8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xB0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eb, LegacyOpKind.Gb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmpxchg_rm16_r16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xB1, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmpxchg_rm32_r32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xB1, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmpxchg_rm64_r64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xB1, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq, LegacyOpKind.Gq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lss_r16_m1616)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xB2, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Mp }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lss_r32_m1632)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xB2, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Mp }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lss_r64_m1664)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xB2, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Mp }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Btr_rm16_r16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xB3, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Btr_rm32_r32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xB3, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Btr_rm64_r64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xB3, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq, LegacyOpKind.Gq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lfs_r16_m1616)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xB4, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Mp }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lfs_r32_m1632)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xB4, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Mp }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lfs_r64_m1664)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xB4, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Mp }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lgs_r16_m1616)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xB5, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Mp }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lgs_r32_m1632)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xB5, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Mp }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lgs_r64_m1664)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xB5, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Mp }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movzx_r16_rm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xB6, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movzx_r32_rm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xB6, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movzx_r64_rm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xB6, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movzx_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xB7, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movzx_r32_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xB7, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movzx_r64_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xB7, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jmpe_disp16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xB8, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Jdisp16 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Jmpe_disp32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xB8, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Jdisp32 }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Popcnt_r16_rm16)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xB8, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Popcnt_r32_rm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xB8, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Popcnt_r64_rm64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xB8, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ud1_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xB9, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ud1_r32_rm32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xB9, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ud1_r64_rm64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xB9, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bt_rm16_imm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xBA, 4, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ew, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bt_rm32_imm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xBA, 4, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bt_rm64_imm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xBA, 4, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bts_rm16_imm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xBA, 5, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bts_rm32_imm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xBA, 5, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bts_rm64_imm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xBA, 5, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Btr_rm16_imm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xBA, 6, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Btr_rm32_imm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xBA, 6, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Btr_rm64_imm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xBA, 6, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Btc_rm16_imm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xBA, 7, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Btc_rm32_imm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xBA, 7, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Btc_rm64_imm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xBA, 7, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Btc_rm16_r16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xBB, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Btc_rm32_r32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xBB, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Btc_rm64_r64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xBB, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq, LegacyOpKind.Gq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bsf_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xBC, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bsf_r32_rm32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xBC, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bsf_r64_rm64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xBC, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Tzcnt_r16_rm16)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xBC, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Tzcnt_r32_rm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xBC, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Tzcnt_r64_rm64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xBC, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bsr_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xBD, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bsr_r32_rm32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xBD, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bsr_r64_rm64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xBD, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lzcnt_r16_rm16)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xBD, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lzcnt_r32_rm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xBD, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lzcnt_r64_rm64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xBD, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movsx_r16_rm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xBE, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movsx_r32_rm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xBE, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movsx_r64_rm8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xBE, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movsx_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xBF, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movsx_r32_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xBF, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movsx_r64_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xBF, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xadd_rm8_r8)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xC0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eb, LegacyOpKind.Gb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xadd_rm16_r16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xC1, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ew, LegacyOpKind.Gw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xadd_rm32_r32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xC1, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Ed, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xadd_rm64_r64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xC1, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Eq, LegacyOpKind.Gq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmpps_xmm_xmmm128_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xC2, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcmpps_xmm_xmm_xmmm128_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xC2, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcmpps_ymm_ymm_ymmm256_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xC2, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcmpps_k_k1_xmm_xmmm128b32_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xC2, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcmpps_k_k1_ymm_ymmm256b32_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xC2, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcmpps_k_k1_zmm_zmmm512b32_imm8_sae)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xC2, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmppd_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xC2, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcmppd_xmm_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xC2, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcmppd_ymm_ymm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xC2, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcmppd_k_k1_xmm_xmmm128b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xC2, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcmppd_k_k1_ymm_ymmm256b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xC2, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcmppd_k_k1_zmm_zmmm512b64_imm8_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xC2, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmpss_xmm_xmmm32_imm8)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xC2, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcmpss_xmm_xmm_xmmm32_imm8)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xC2, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcmpss_k_k1_xmm_xmmm32_imm8_sae)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xC2, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmpsd_xmm_xmmm64_imm8)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0xC2, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcmpsd_xmm_xmm_xmmm64_imm8)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0xC2, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcmpsd_k_k1_xmm_xmmm64_imm8_sae)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0xC2, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movnti_m32_r32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xC3, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Md, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movnti_m64_r64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xC3, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Mq, LegacyOpKind.Gq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pinsrw_mm_r32m16_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xC4, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.RdMw, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pinsrw_mm_r64m16_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xC4, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.RqMw, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pinsrw_xmm_r32m16_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xC4, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.RdMw, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pinsrw_xmm_r64m16_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xC4, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.RqMw, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpinsrw_xmm_xmm_r32m16_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xC4, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.RdMw, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpinsrw_xmm_xmm_r64m16_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xC4, -1, VexVectorLength.L128, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.RqMw, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpinsrw_xmm_xmm_r32m16_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xC4, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar_2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.RdMw, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpinsrw_xmm_xmm_r64m16_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xC4, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar_2, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.RqMw, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pextrw_r32_mm_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xC5, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.N, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pextrw_r64_mm_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xC5, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.N, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pextrw_r32_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xC5, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.RX, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pextrw_r64_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xC5, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.RX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpextrw_r32_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xC5, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { VexOpKind.Gd, VexOpKind.RX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpextrw_r64_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xC5, -1, VexVectorLength.L128, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.Gq, VexOpKind.RX, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpextrw_r32_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xC5, -1, EvexVectorLength.L128, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { EvexOpKind.Gd, EvexOpKind.RX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpextrw_r64_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xC5, -1, EvexVectorLength.L128, TupleType.None, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.Gq, EvexOpKind.RX, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shufps_xmm_xmmm128_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xC6, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vshufps_xmm_xmm_xmmm128_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xC6, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vshufps_ymm_ymm_ymmm256_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xC6, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vshufps_xmm_k1z_xmm_xmmm128b32_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xC6, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vshufps_ymm_k1z_ymm_ymmm256b32_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xC6, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vshufps_zmm_k1z_zmm_zmmm512b32_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xC6, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Shufpd_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xC6, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vshufpd_xmm_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xC6, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vshufpd_ymm_ymm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xC6, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vshufpd_xmm_k1z_xmm_xmmm128b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xC6, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vshufpd_ymm_k1z_ymm_ymmm256b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xC6, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vshufpd_zmm_k1z_zmm_zmmm512b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xC6, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmpxchg8b_m64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xC7, 1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix | OpCodeFlags.XacquirePrefix | OpCodeFlags.XreleasePrefix, new[] { LegacyOpKind.Mq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cmpxchg16b_m128)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xC7, 1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64 | OpCodeFlags.LockPrefix, new[] { LegacyOpKind.Mo }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xrstors_mem)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xC7, 3, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xrstors64_mem)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xC7, 3, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xsavec_mem)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xC7, 4, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xsavec64_mem)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xC7, 4, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xsaves_mem)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xC7, 5, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Xsaves64_mem)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xC7, 5, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Vmptrld_m64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xC7, 6, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Vmclear_m64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xC7, 6, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Vmxon_m64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xC7, 6, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rdrand_r16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xC7, 6, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Rw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rdrand_r32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xC7, 6, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Rd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rdrand_r64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xC7, 6, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Rq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Vmptrst_m64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xC7, 7, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rdseed_r16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xC7, 7, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Rw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rdseed_r32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xC7, 7, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Rd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rdseed_r64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xC7, 7, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Rq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rdpid_r32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xC7, 7, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Rd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Rdpid_r64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xC7, 7, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Rq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bswap_r16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xC8, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.r16_rw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bswap_r32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xC8, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.r32_rd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Bswap_r64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xC8, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.r64_ro }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Addsubpd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vaddsubpd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD0, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vaddsubpd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD0, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Addsubps_xmm_xmmm128)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0xD0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vaddsubps_xmm_xmm_xmmm128)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0xD0, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vaddsubps_ymm_ymm_ymmm256)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0xD0, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psrlw_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xD1, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psrlw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD1, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsrlw_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD1, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsrlw_ymm_ymm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD1, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrlw_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD1, -1, EvexVectorLength.L128, TupleType.Mem128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrlw_ymm_k1z_ymm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD1, -1, EvexVectorLength.L256, TupleType.Mem128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrlw_zmm_k1z_zmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD1, -1, EvexVectorLength.L512, TupleType.Mem128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psrld_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xD2, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psrld_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD2, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsrld_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD2, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsrld_ymm_ymm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD2, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrld_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD2, -1, EvexVectorLength.L128, TupleType.Mem128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrld_ymm_k1z_ymm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD2, -1, EvexVectorLength.L256, TupleType.Mem128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrld_zmm_k1z_zmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD2, -1, EvexVectorLength.L512, TupleType.Mem128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psrlq_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xD3, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psrlq_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD3, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsrlq_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD3, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsrlq_ymm_ymm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD3, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrlq_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD3, -1, EvexVectorLength.L128, TupleType.Mem128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrlq_ymm_k1z_ymm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD3, -1, EvexVectorLength.L256, TupleType.Mem128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrlq_zmm_k1z_zmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD3, -1, EvexVectorLength.L512, TupleType.Mem128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Paddq_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xD4, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Paddq_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD4, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpaddq_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD4, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpaddq_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD4, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpaddq_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD4, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpaddq_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD4, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpaddq_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD4, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmullw_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xD5, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmullw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD5, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmullw_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD5, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmullw_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD5, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmullw_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD5, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmullw_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD5, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmullw_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD5, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movq_xmmm64_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD6, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.WX, LegacyOpKind.VX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovq_xmmm64_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD6, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.WX, VexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovq_xmmm64_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD6, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movq2dq_xmm_mm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xD6, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.N }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movdq2q_mm_xmm)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0xD6, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.RX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmovmskb_r32_mm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xD7, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.N }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmovmskb_r64_mm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xD7, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.N }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmovmskb_r32_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD7, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.RX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmovmskb_r64_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD7, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.RX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmovmskb_r32_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD7, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { VexOpKind.Gd, VexOpKind.RX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmovmskb_r64_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD7, -1, VexVectorLength.L128, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.Gq, VexOpKind.RX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmovmskb_r32_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD7, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { VexOpKind.Gd, VexOpKind.RY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmovmskb_r64_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD7, -1, VexVectorLength.L256, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.Gq, VexOpKind.RY }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psubusb_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xD8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psubusb_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsubusb_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD8, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsubusb_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD8, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsubusb_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD8, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsubusb_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD8, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsubusb_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD8, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psubusw_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xD9, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psubusw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD9, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsubusw_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD9, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsubusw_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD9, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsubusw_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD9, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsubusw_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD9, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsubusw_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xD9, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pminub_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xDA, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pminub_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDA, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpminub_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDA, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpminub_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDA, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpminub_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDA, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpminub_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDA, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpminub_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDA, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pand_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xDB, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pand_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDB, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpand_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDB, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpand_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDB, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpandd_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDB, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpandd_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDB, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpandd_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDB, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpandq_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDB, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpandq_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDB, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpandq_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDB, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Paddusb_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xDC, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Paddusb_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDC, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpaddusb_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDC, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpaddusb_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDC, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpaddusb_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDC, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpaddusb_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDC, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpaddusb_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDC, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Paddusw_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xDD, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Paddusw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDD, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpaddusw_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDD, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpaddusw_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDD, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpaddusw_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDD, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpaddusw_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDD, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpaddusw_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDD, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmaxub_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xDE, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmaxub_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDE, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmaxub_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDE, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmaxub_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDE, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmaxub_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDE, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmaxub_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDE, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmaxub_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDE, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pandn_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xDF, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pandn_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDF, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpandn_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDF, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpandn_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDF, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpandnd_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDF, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpandnd_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDF, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpandnd_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDF, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpandnq_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDF, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpandnq_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDF, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpandnq_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xDF, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pavgb_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xE0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pavgb_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpavgb_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE0, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpavgb_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE0, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpavgb_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE0, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpavgb_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE0, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpavgb_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE0, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psraw_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xE1, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psraw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE1, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsraw_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE1, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsraw_ymm_ymm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE1, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsraw_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE1, -1, EvexVectorLength.L128, TupleType.Mem128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsraw_ymm_k1z_ymm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE1, -1, EvexVectorLength.L256, TupleType.Mem128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsraw_zmm_k1z_zmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE1, -1, EvexVectorLength.L512, TupleType.Mem128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psrad_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xE2, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psrad_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE2, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsrad_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE2, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsrad_ymm_ymm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE2, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrad_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE2, -1, EvexVectorLength.L128, TupleType.Mem128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrad_ymm_k1z_ymm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE2, -1, EvexVectorLength.L256, TupleType.Mem128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrad_zmm_k1z_zmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE2, -1, EvexVectorLength.L512, TupleType.Mem128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsraq_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE2, -1, EvexVectorLength.L128, TupleType.Mem128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsraq_ymm_k1z_ymm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE2, -1, EvexVectorLength.L256, TupleType.Mem128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsraq_zmm_k1z_zmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE2, -1, EvexVectorLength.L512, TupleType.Mem128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pavgw_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xE3, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pavgw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE3, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpavgw_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE3, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpavgw_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE3, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpavgw_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE3, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpavgw_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE3, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpavgw_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE3, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmulhuw_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xE4, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmulhuw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE4, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmulhuw_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE4, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmulhuw_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE4, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmulhuw_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE4, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmulhuw_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE4, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmulhuw_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE4, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmulhw_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xE5, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmulhw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE5, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmulhw_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE5, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmulhw_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE5, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmulhw_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE5, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmulhw_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE5, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmulhw_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE5, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cvttpd2dq_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE6, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvttpd2dq_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE6, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvttpd2dq_xmm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE6, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvttpd2dq_xmm_k1z_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE6, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvttpd2dq_xmm_k1z_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE6, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvttpd2dq_ymm_k1z_zmmm512b64_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE6, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cvtdq2pd_xmm_xmmm64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xE6, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvtdq2pd_xmm_xmmm64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xE6, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvtdq2pd_ymm_xmmm128)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xE6, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtdq2pd_xmm_k1z_xmmm64b32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xE6, -1, EvexVectorLength.L128, TupleType.Half_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtdq2pd_ymm_k1z_xmmm128b32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xE6, -1, EvexVectorLength.L256, TupleType.Half_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtdq2pd_zmm_k1z_ymmm256b32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xE6, -1, EvexVectorLength.L512, TupleType.Half_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtqq2pd_xmm_k1z_xmmm128b64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xE6, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtqq2pd_ymm_k1z_ymmm256b64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xE6, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtqq2pd_zmm_k1z_zmmm512b64_er)], MandatoryPrefix.PF3, OpCodeTableKind.T0F, 0xE6, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Cvtpd2dq_xmm_xmmm128)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0xE6, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvtpd2dq_xmm_xmmm128)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0xE6, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvtpd2dq_xmm_ymmm256)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0xE6, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtpd2dq_xmm_k1z_xmmm128b64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0xE6, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtpd2dq_xmm_k1z_ymmm256b64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0xE6, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtpd2dq_ymm_k1z_zmmm512b64_er)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0xE6, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movntq_m64_mm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xE7, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.M, LegacyOpKind.P }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movntdq_m128_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE7, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.M, LegacyOpKind.VX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovntdq_m128_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE7, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.M, VexOpKind.VX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovntdq_m256_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE7, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.M, VexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovntdq_m128_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE7, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { EvexOpKind.M, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovntdq_m256_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE7, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { EvexOpKind.M, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovntdq_m512_zmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE7, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { EvexOpKind.M, EvexOpKind.VZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psubsb_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xE8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psubsb_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsubsb_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE8, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsubsb_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE8, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsubsb_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE8, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsubsb_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE8, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsubsb_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE8, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psubsw_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xE9, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psubsw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE9, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsubsw_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE9, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsubsw_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE9, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsubsw_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE9, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsubsw_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE9, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsubsw_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xE9, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pminsw_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xEA, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pminsw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEA, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpminsw_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEA, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpminsw_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEA, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpminsw_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEA, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpminsw_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEA, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpminsw_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEA, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Por_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xEB, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Por_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEB, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpor_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEB, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpor_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEB, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpord_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEB, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpord_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEB, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpord_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEB, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vporq_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEB, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vporq_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEB, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vporq_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEB, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Paddsb_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xEC, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Paddsb_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEC, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpaddsb_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEC, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpaddsb_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEC, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpaddsb_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEC, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpaddsb_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEC, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpaddsb_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEC, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Paddsw_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xED, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Paddsw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xED, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpaddsw_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xED, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpaddsw_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xED, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpaddsw_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xED, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpaddsw_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xED, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpaddsw_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xED, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmaxsw_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xEE, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmaxsw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEE, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmaxsw_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEE, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmaxsw_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEE, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmaxsw_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEE, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmaxsw_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEE, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmaxsw_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEE, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pxor_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xEF, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pxor_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEF, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpxor_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEF, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpxor_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEF, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpxord_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEF, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpxord_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEF, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpxord_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEF, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpxorq_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEF, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpxorq_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEF, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpxorq_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xEF, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Lddqu_xmm_m128)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0xF0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.M }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vlddqu_xmm_m128)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0xF0, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.M }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vlddqu_ymm_m256)], MandatoryPrefix.PF2, OpCodeTableKind.T0F, 0xF0, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psllw_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xF1, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psllw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF1, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsllw_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF1, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsllw_ymm_ymm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF1, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsllw_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF1, -1, EvexVectorLength.L128, TupleType.Mem128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsllw_ymm_k1z_ymm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF1, -1, EvexVectorLength.L256, TupleType.Mem128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsllw_zmm_k1z_zmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF1, -1, EvexVectorLength.L512, TupleType.Mem128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pslld_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xF2, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pslld_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF2, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpslld_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF2, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpslld_ymm_ymm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF2, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpslld_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF2, -1, EvexVectorLength.L128, TupleType.Mem128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpslld_ymm_k1z_ymm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF2, -1, EvexVectorLength.L256, TupleType.Mem128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpslld_zmm_k1z_zmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF2, -1, EvexVectorLength.L512, TupleType.Mem128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psllq_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xF3, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psllq_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF3, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsllq_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF3, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsllq_ymm_ymm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF3, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsllq_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF3, -1, EvexVectorLength.L128, TupleType.Mem128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsllq_ymm_k1z_ymm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF3, -1, EvexVectorLength.L256, TupleType.Mem128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsllq_zmm_k1z_zmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF3, -1, EvexVectorLength.L512, TupleType.Mem128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmuludq_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xF4, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmuludq_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF4, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmuludq_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF4, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmuludq_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF4, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmuludq_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF4, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmuludq_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF4, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmuludq_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF4, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmaddwd_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xF5, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmaddwd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF5, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmaddwd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF5, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmaddwd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF5, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmaddwd_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF5, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmaddwd_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF5, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmaddwd_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF5, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psadbw_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xF6, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psadbw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF6, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsadbw_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF6, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsadbw_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF6, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsadbw_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF6, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsadbw_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF6, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsadbw_zmm_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF6, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Maskmovq_rDI_mm_mm)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xF7, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.rDI, LegacyOpKind.P, LegacyOpKind.N }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Maskmovdqu_rDI_xmm_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF7, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.rDI, LegacyOpKind.VX, LegacyOpKind.RX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmaskmovdqu_rDI_xmm_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF7, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.rDI, VexOpKind.VX, VexOpKind.RX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psubb_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xF8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psubb_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsubb_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF8, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsubb_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF8, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsubb_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF8, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsubb_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF8, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsubb_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF8, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psubw_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xF9, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psubw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF9, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsubw_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF9, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsubw_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF9, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsubw_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF9, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsubw_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF9, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsubw_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xF9, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psubd_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xFA, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psubd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xFA, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsubd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xFA, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsubd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xFA, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsubd_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xFA, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsubd_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xFA, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsubd_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xFA, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psubq_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xFB, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psubq_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xFB, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsubq_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xFB, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsubq_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xFB, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsubq_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xFB, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsubq_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xFB, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsubq_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xFB, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Paddb_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xFC, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Paddb_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xFC, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpaddb_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xFC, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpaddb_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xFC, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpaddb_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xFC, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpaddb_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xFC, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpaddb_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xFC, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Paddw_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xFD, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Paddw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xFD, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpaddw_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xFD, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpaddw_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xFD, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpaddw_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xFD, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpaddw_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xFD, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpaddw_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xFD, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Paddd_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F, 0xFE, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Paddd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xFE, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpaddd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xFE, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpaddd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xFE, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpaddd_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xFE, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpaddd_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xFE, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpaddd_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F, 0xFE, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ud0_r16_rm16)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xFF, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ud0_r32_rm32)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xFF, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ud0_r64_rm64)], MandatoryPrefix.None, OpCodeTableKind.T0F, 0xFF, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pshufb_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0x00, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pshufb_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x00, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpshufb_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x00, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpshufb_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x00, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshufb_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x00, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshufb_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x00, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshufb_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x00, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Phaddw_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0x01, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Phaddw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x01, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vphaddw_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x01, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vphaddw_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x01, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Phaddd_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0x02, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Phaddd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x02, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vphaddd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x02, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vphaddd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x02, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Phaddsw_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0x03, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Phaddsw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x03, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vphaddsw_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x03, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vphaddsw_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x03, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmaddubsw_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0x04, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmaddubsw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x04, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmaddubsw_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x04, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmaddubsw_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x04, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmaddubsw_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x04, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmaddubsw_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x04, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmaddubsw_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x04, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Phsubw_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0x05, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Phsubw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x05, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vphsubw_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x05, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vphsubw_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x05, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Phsubd_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0x06, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Phsubd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x06, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vphsubd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x06, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vphsubd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x06, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Phsubsw_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0x07, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Phsubsw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x07, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vphsubsw_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x07, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vphsubsw_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x07, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psignb_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0x08, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psignb_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x08, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsignb_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x08, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsignb_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x08, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psignw_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0x09, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psignw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x09, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsignw_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x09, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsignw_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x09, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psignd_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0x0A, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Psignd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x0A, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsignd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x0A, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsignd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x0A, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmulhrsw_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0x0B, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmulhrsw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x0B, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmulhrsw_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x0B, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmulhrsw_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x0B, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmulhrsw_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x0B, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmulhrsw_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x0B, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmulhrsw_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x0B, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpermilps_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x0C, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpermilps_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x0C, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermilps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x0C, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermilps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x0C, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermilps_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x0C, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpermilpd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x0D, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpermilpd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x0D, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermilpd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x0D, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermilpd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x0D, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermilpd_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x0D, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vtestps_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x0E, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vtestps_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x0E, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.WY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vtestpd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x0F, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vtestpd_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x0F, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.WY }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pblendvb_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x10, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrlvw_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x10, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrlvw_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x10, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrlvw_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x10, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovuswb_xmmm64_k1z_xmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x10, -1, EvexVectorLength.L128, TupleType.Half_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovuswb_xmmm128_k1z_ymm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x10, -1, EvexVectorLength.L256, TupleType.Half_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovuswb_ymmm256_k1z_zmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x10, -1, EvexVectorLength.L512, TupleType.Half_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WY, EvexOpKind.VZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsravw_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x11, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsravw_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x11, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsravw_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x11, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovusdb_xmmm32_k1z_xmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x11, -1, EvexVectorLength.L128, TupleType.Quarter_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovusdb_xmmm64_k1z_ymm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x11, -1, EvexVectorLength.L256, TupleType.Quarter_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovusdb_xmmm128_k1z_zmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x11, -1, EvexVectorLength.L512, TupleType.Quarter_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsllvw_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x12, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsllvw_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x12, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsllvw_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x12, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovusqb_xmmm16_k1z_xmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x12, -1, EvexVectorLength.L128, TupleType.Eighth_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovusqb_xmmm32_k1z_ymm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x12, -1, EvexVectorLength.L256, TupleType.Eighth_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovusqb_xmmm64_k1z_zmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x12, -1, EvexVectorLength.L512, TupleType.Eighth_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvtph2ps_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x13, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvtph2ps_ymm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x13, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtph2ps_xmm_k1z_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x13, -1, EvexVectorLength.L128, TupleType.Half_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtph2ps_ymm_k1z_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x13, -1, EvexVectorLength.L256, TupleType.Half_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtph2ps_zmm_k1z_ymmm256_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x13, -1, EvexVectorLength.L512, TupleType.Half_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovusdw_xmmm64_k1z_xmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x13, -1, EvexVectorLength.L128, TupleType.Half_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovusdw_xmmm128_k1z_ymm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x13, -1, EvexVectorLength.L256, TupleType.Half_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovusdw_ymmm256_k1z_zmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x13, -1, EvexVectorLength.L512, TupleType.Half_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WY, EvexOpKind.VZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Blendvps_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x14, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vprorvd_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x14, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vprorvd_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x14, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vprorvd_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x14, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vprorvq_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x14, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vprorvq_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x14, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vprorvq_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x14, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovusqw_xmmm32_k1z_xmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x14, -1, EvexVectorLength.L128, TupleType.Quarter_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovusqw_xmmm64_k1z_ymm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x14, -1, EvexVectorLength.L256, TupleType.Quarter_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovusqw_xmmm128_k1z_zmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x14, -1, EvexVectorLength.L512, TupleType.Quarter_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Blendvpd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x15, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vprolvd_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x15, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vprolvd_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x15, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vprolvd_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x15, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vprolvq_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x15, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vprolvq_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x15, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vprolvq_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x15, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovusqd_xmmm64_k1z_xmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x15, -1, EvexVectorLength.L128, TupleType.Half_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovusqd_xmmm128_k1z_ymm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x15, -1, EvexVectorLength.L256, TupleType.Half_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovusqd_ymmm256_k1z_zmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x15, -1, EvexVectorLength.L512, TupleType.Half_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WY, EvexOpKind.VZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpermps_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x16, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x16, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermps_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x16, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermpd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x16, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermpd_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x16, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Ptest_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x17, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vptest_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x17, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vptest_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x17, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vbroadcastss_xmm_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x18, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vbroadcastss_ymm_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x18, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vbroadcastss_xmm_k1z_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x18, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vbroadcastss_ymm_k1z_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x18, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vbroadcastss_zmm_k1z_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x18, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vbroadcastsd_ymm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x19, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vbroadcastf32x2_ymm_k1z_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x19, -1, EvexVectorLength.L256, TupleType.Tuple2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vbroadcastf32x2_zmm_k1z_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x19, -1, EvexVectorLength.L512, TupleType.Tuple2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vbroadcastsd_ymm_k1z_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x19, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vbroadcastsd_zmm_k1z_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x19, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vbroadcastf128_ymm_m128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x1A, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.M }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vbroadcastf32x4_ymm_k1z_m128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x1A, -1, EvexVectorLength.L256, TupleType.Tuple4, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.M }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vbroadcastf32x4_zmm_k1z_m128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x1A, -1, EvexVectorLength.L512, TupleType.Tuple4, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.M }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vbroadcastf64x2_ymm_k1z_m128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x1A, -1, EvexVectorLength.L256, TupleType.Tuple2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.M }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vbroadcastf64x2_zmm_k1z_m128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x1A, -1, EvexVectorLength.L512, TupleType.Tuple2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.M }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vbroadcastf32x8_zmm_k1z_m256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x1B, -1, EvexVectorLength.L512, TupleType.Tuple8, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.M }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vbroadcastf64x4_zmm_k1z_m256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x1B, -1, EvexVectorLength.L512, TupleType.Tuple4, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pabsb_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0x1C, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pabsb_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x1C, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpabsb_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x1C, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpabsb_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x1C, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpabsb_xmm_k1z_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x1C, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpabsb_ymm_k1z_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x1C, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpabsb_zmm_k1z_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x1C, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pabsw_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0x1D, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pabsw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x1D, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpabsw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x1D, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpabsw_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x1D, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpabsw_xmm_k1z_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x1D, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpabsw_ymm_k1z_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x1D, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpabsw_zmm_k1z_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x1D, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pabsd_mm_mmm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0x1E, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pabsd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x1E, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpabsd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x1E, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpabsd_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x1E, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpabsd_xmm_k1z_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x1E, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpabsd_ymm_k1z_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x1E, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpabsd_zmm_k1z_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x1E, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpabsq_xmm_k1z_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x1F, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpabsq_ymm_k1z_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x1F, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpabsq_zmm_k1z_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x1F, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmovsxbw_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x20, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmovsxbw_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x20, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmovsxbw_ymm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x20, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovsxbw_xmm_k1z_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x20, -1, EvexVectorLength.L128, TupleType.Half_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovsxbw_ymm_k1z_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x20, -1, EvexVectorLength.L256, TupleType.Half_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovsxbw_zmm_k1z_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x20, -1, EvexVectorLength.L512, TupleType.Half_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovswb_xmmm64_k1z_xmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x20, -1, EvexVectorLength.L128, TupleType.Half_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovswb_xmmm128_k1z_ymm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x20, -1, EvexVectorLength.L256, TupleType.Half_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovswb_ymmm256_k1z_zmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x20, -1, EvexVectorLength.L512, TupleType.Half_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WY, EvexOpKind.VZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmovsxbd_xmm_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x21, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmovsxbd_xmm_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x21, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmovsxbd_ymm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x21, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovsxbd_xmm_k1z_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x21, -1, EvexVectorLength.L128, TupleType.Quarter_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovsxbd_ymm_k1z_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x21, -1, EvexVectorLength.L256, TupleType.Quarter_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovsxbd_zmm_k1z_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x21, -1, EvexVectorLength.L512, TupleType.Quarter_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovsdb_xmmm32_k1z_xmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x21, -1, EvexVectorLength.L128, TupleType.Quarter_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovsdb_xmmm64_k1z_ymm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x21, -1, EvexVectorLength.L256, TupleType.Quarter_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovsdb_xmmm128_k1z_zmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x21, -1, EvexVectorLength.L512, TupleType.Quarter_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmovsxbq_xmm_xmmm16)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x22, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmovsxbq_xmm_xmmm16)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x22, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmovsxbq_ymm_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x22, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovsxbq_xmm_k1z_xmmm16)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x22, -1, EvexVectorLength.L128, TupleType.Eighth_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovsxbq_ymm_k1z_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x22, -1, EvexVectorLength.L256, TupleType.Eighth_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovsxbq_zmm_k1z_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x22, -1, EvexVectorLength.L512, TupleType.Eighth_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovsqb_xmmm16_k1z_xmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x22, -1, EvexVectorLength.L128, TupleType.Eighth_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovsqb_xmmm32_k1z_ymm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x22, -1, EvexVectorLength.L256, TupleType.Eighth_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovsqb_xmmm64_k1z_zmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x22, -1, EvexVectorLength.L512, TupleType.Eighth_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmovsxwd_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x23, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmovsxwd_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x23, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmovsxwd_ymm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x23, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovsxwd_xmm_k1z_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x23, -1, EvexVectorLength.L128, TupleType.Half_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovsxwd_ymm_k1z_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x23, -1, EvexVectorLength.L256, TupleType.Half_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovsxwd_zmm_k1z_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x23, -1, EvexVectorLength.L512, TupleType.Half_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovsdw_xmmm64_k1z_xmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x23, -1, EvexVectorLength.L128, TupleType.Half_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovsdw_xmmm128_k1z_ymm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x23, -1, EvexVectorLength.L256, TupleType.Half_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovsdw_ymmm256_k1z_zmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x23, -1, EvexVectorLength.L512, TupleType.Half_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WY, EvexOpKind.VZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmovsxwq_xmm_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x24, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmovsxwq_xmm_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x24, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmovsxwq_ymm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x24, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovsxwq_xmm_k1z_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x24, -1, EvexVectorLength.L128, TupleType.Quarter_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovsxwq_ymm_k1z_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x24, -1, EvexVectorLength.L256, TupleType.Quarter_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovsxwq_zmm_k1z_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x24, -1, EvexVectorLength.L512, TupleType.Quarter_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovsqw_xmmm32_k1z_xmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x24, -1, EvexVectorLength.L128, TupleType.Quarter_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovsqw_xmmm64_k1z_ymm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x24, -1, EvexVectorLength.L256, TupleType.Quarter_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovsqw_xmmm128_k1z_zmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x24, -1, EvexVectorLength.L512, TupleType.Quarter_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmovsxdq_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x25, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmovsxdq_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x25, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmovsxdq_ymm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x25, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovsxdq_xmm_k1z_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x25, -1, EvexVectorLength.L128, TupleType.Half_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovsxdq_ymm_k1z_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x25, -1, EvexVectorLength.L256, TupleType.Half_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovsxdq_zmm_k1z_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x25, -1, EvexVectorLength.L512, TupleType.Half_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovsqd_xmmm64_k1z_xmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x25, -1, EvexVectorLength.L128, TupleType.Half_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovsqd_xmmm128_k1z_ymm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x25, -1, EvexVectorLength.L256, TupleType.Half_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovsqd_ymmm256_k1z_zmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x25, -1, EvexVectorLength.L512, TupleType.Half_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WY, EvexOpKind.VZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vptestmb_k_k1_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x26, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vptestmb_k_k1_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x26, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vptestmb_k_k1_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x26, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vptestmw_k_k1_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x26, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vptestmw_k_k1_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x26, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vptestmw_k_k1_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x26, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vptestnmb_k_k1_xmm_xmmm128)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x26, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vptestnmb_k_k1_ymm_ymmm256)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x26, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vptestnmb_k_k1_zmm_zmmm512)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x26, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vptestnmw_k_k1_xmm_xmmm128)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x26, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vptestnmw_k_k1_ymm_ymmm256)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x26, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vptestnmw_k_k1_zmm_zmmm512)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x26, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vptestmd_k_k1_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x27, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vptestmd_k_k1_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x27, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vptestmd_k_k1_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x27, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vptestmq_k_k1_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x27, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vptestmq_k_k1_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x27, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vptestmq_k_k1_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x27, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vptestnmd_k_k1_xmm_xmmm128b32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x27, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vptestnmd_k_k1_ymm_ymmm256b32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x27, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vptestnmd_k_k1_zmm_zmmm512b32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x27, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vptestnmq_k_k1_xmm_xmmm128b64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x27, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vptestnmq_k_k1_ymm_ymmm256b64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x27, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vptestnmq_k_k1_zmm_zmmm512b64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x27, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmuldq_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x28, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmuldq_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x28, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmuldq_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x28, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmuldq_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x28, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmuldq_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x28, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmuldq_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x28, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovm2b_xmm_k)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x28, -1, EvexVectorLength.L128, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { EvexOpKind.VX, EvexOpKind.RK }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovm2b_ymm_k)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x28, -1, EvexVectorLength.L256, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { EvexOpKind.VY, EvexOpKind.RK }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovm2b_zmm_k)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x28, -1, EvexVectorLength.L512, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { EvexOpKind.VZ, EvexOpKind.RK }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovm2w_xmm_k)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x28, -1, EvexVectorLength.L128, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.VX, EvexOpKind.RK }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovm2w_ymm_k)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x28, -1, EvexVectorLength.L256, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.VY, EvexOpKind.RK }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovm2w_zmm_k)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x28, -1, EvexVectorLength.L512, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.VZ, EvexOpKind.RK }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pcmpeqq_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x29, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpcmpeqq_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x29, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpcmpeqq_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x29, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpeqq_k_k1_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x29, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpeqq_k_k1_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x29, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpeqq_k_k1_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x29, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovb2m_k_xmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x29, -1, EvexVectorLength.L128, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { EvexOpKind.VK, EvexOpKind.RX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovb2m_k_ymm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x29, -1, EvexVectorLength.L256, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { EvexOpKind.VK, EvexOpKind.RY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovb2m_k_zmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x29, -1, EvexVectorLength.L512, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { EvexOpKind.VK, EvexOpKind.RZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovw2m_k_xmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x29, -1, EvexVectorLength.L128, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.VK, EvexOpKind.RX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovw2m_k_ymm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x29, -1, EvexVectorLength.L256, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.VK, EvexOpKind.RY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovw2m_k_zmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x29, -1, EvexVectorLength.L512, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.VK, EvexOpKind.RZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movntdqa_xmm_m128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x2A, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.M }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovntdqa_xmm_m128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x2A, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.M }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmovntdqa_ymm_m256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x2A, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.M }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovntdqa_xmm_m128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x2A, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { EvexOpKind.VX, EvexOpKind.M }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovntdqa_ymm_m256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x2A, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { EvexOpKind.VY, EvexOpKind.M }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vmovntdqa_zmm_m512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x2A, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { EvexOpKind.VZ, EvexOpKind.M }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpbroadcastmb2q_xmm_k)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x2A, -1, EvexVectorLength.L128, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.VX, EvexOpKind.RK }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpbroadcastmb2q_ymm_k)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x2A, -1, EvexVectorLength.L256, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.VY, EvexOpKind.RK }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpbroadcastmb2q_zmm_k)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x2A, -1, EvexVectorLength.L512, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.VZ, EvexOpKind.RK }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Packusdw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x2B, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpackusdw_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x2B, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpackusdw_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x2B, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpackusdw_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x2B, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpackusdw_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x2B, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpackusdw_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x2B, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmaskmovps_xmm_xmm_m128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x2C, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.M }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmaskmovps_ymm_ymm_m256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x2C, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.M }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vscalefps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x2C, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vscalefps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x2C, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vscalefps_zmm_k1z_zmm_zmmm512b32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x2C, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vscalefpd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x2C, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vscalefpd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x2C, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vscalefpd_zmm_k1z_zmm_zmmm512b64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x2C, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmaskmovpd_xmm_xmm_m128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x2D, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.M }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmaskmovpd_ymm_ymm_m256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x2D, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.M }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vscalefss_xmm_k1z_xmm_xmmm32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x2D, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vscalefsd_xmm_k1z_xmm_xmmm64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x2D, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmaskmovps_m128_xmm_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x2E, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.M, VexOpKind.HX, VexOpKind.VX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmaskmovps_m256_ymm_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x2E, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.M, VexOpKind.HY, VexOpKind.VY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmaskmovpd_m128_xmm_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x2F, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.M, VexOpKind.HX, VexOpKind.VX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmaskmovpd_m256_ymm_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x2F, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.M, VexOpKind.HY, VexOpKind.VY }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmovzxbw_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x30, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmovzxbw_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x30, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmovzxbw_ymm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x30, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovzxbw_xmm_k1z_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x30, -1, EvexVectorLength.L128, TupleType.Half_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovzxbw_ymm_k1z_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x30, -1, EvexVectorLength.L256, TupleType.Half_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovzxbw_zmm_k1z_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x30, -1, EvexVectorLength.L512, TupleType.Half_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovwb_xmmm64_k1z_xmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x30, -1, EvexVectorLength.L128, TupleType.Half_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovwb_xmmm128_k1z_ymm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x30, -1, EvexVectorLength.L256, TupleType.Half_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovwb_ymmm256_k1z_zmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x30, -1, EvexVectorLength.L512, TupleType.Half_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WY, EvexOpKind.VZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmovzxbd_xmm_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x31, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmovzxbd_xmm_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x31, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmovzxbd_ymm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x31, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovzxbd_xmm_k1z_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x31, -1, EvexVectorLength.L128, TupleType.Quarter_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovzxbd_ymm_k1z_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x31, -1, EvexVectorLength.L256, TupleType.Quarter_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovzxbd_zmm_k1z_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x31, -1, EvexVectorLength.L512, TupleType.Quarter_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovdb_xmmm32_k1z_xmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x31, -1, EvexVectorLength.L128, TupleType.Quarter_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovdb_xmmm64_k1z_ymm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x31, -1, EvexVectorLength.L256, TupleType.Quarter_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovdb_xmmm128_k1z_zmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x31, -1, EvexVectorLength.L512, TupleType.Quarter_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmovzxbq_xmm_xmmm16)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x32, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmovzxbq_xmm_xmmm16)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x32, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmovzxbq_ymm_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x32, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovzxbq_xmm_k1z_xmmm16)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x32, -1, EvexVectorLength.L128, TupleType.Eighth_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovzxbq_ymm_k1z_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x32, -1, EvexVectorLength.L256, TupleType.Eighth_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovzxbq_zmm_k1z_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x32, -1, EvexVectorLength.L512, TupleType.Eighth_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovqb_xmmm16_k1z_xmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x32, -1, EvexVectorLength.L128, TupleType.Eighth_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovqb_xmmm32_k1z_ymm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x32, -1, EvexVectorLength.L256, TupleType.Eighth_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovqb_xmmm64_k1z_zmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x32, -1, EvexVectorLength.L512, TupleType.Eighth_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmovzxwd_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x33, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmovzxwd_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x33, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmovzxwd_ymm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x33, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovzxwd_xmm_k1z_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x33, -1, EvexVectorLength.L128, TupleType.Half_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovzxwd_ymm_k1z_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x33, -1, EvexVectorLength.L256, TupleType.Half_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovzxwd_zmm_k1z_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x33, -1, EvexVectorLength.L512, TupleType.Half_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovdw_xmmm64_k1z_xmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x33, -1, EvexVectorLength.L128, TupleType.Half_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovdw_xmmm128_k1z_ymm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x33, -1, EvexVectorLength.L256, TupleType.Half_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovdw_ymmm256_k1z_zmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x33, -1, EvexVectorLength.L512, TupleType.Half_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WY, EvexOpKind.VZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmovzxwq_xmm_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x34, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmovzxwq_xmm_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x34, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmovzxwq_ymm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x34, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovzxwq_xmm_k1z_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x34, -1, EvexVectorLength.L128, TupleType.Quarter_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovzxwq_ymm_k1z_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x34, -1, EvexVectorLength.L256, TupleType.Quarter_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovzxwq_zmm_k1z_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x34, -1, EvexVectorLength.L512, TupleType.Quarter_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovqw_xmmm32_k1z_xmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x34, -1, EvexVectorLength.L128, TupleType.Quarter_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovqw_xmmm64_k1z_ymm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x34, -1, EvexVectorLength.L256, TupleType.Quarter_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovqw_xmmm128_k1z_zmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x34, -1, EvexVectorLength.L512, TupleType.Quarter_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmovzxdq_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x35, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmovzxdq_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x35, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmovzxdq_ymm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x35, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovzxdq_xmm_k1z_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x35, -1, EvexVectorLength.L128, TupleType.Half_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovzxdq_ymm_k1z_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x35, -1, EvexVectorLength.L256, TupleType.Half_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovzxdq_zmm_k1z_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x35, -1, EvexVectorLength.L512, TupleType.Half_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovqd_xmmm64_k1z_xmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x35, -1, EvexVectorLength.L128, TupleType.Half_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovqd_xmmm128_k1z_ymm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x35, -1, EvexVectorLength.L256, TupleType.Half_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovqd_ymmm256_k1z_zmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x35, -1, EvexVectorLength.L512, TupleType.Half_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WY, EvexOpKind.VZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpermd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x36, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermd_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x36, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermd_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x36, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermq_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x36, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermq_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x36, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pcmpgtq_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x37, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpcmpgtq_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x37, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpcmpgtq_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x37, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpgtq_k_k1_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x37, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpgtq_k_k1_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x37, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpgtq_k_k1_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x37, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pminsb_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x38, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpminsb_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x38, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpminsb_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x38, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpminsb_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x38, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpminsb_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x38, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpminsb_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x38, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovm2d_xmm_k)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x38, -1, EvexVectorLength.L128, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { EvexOpKind.VX, EvexOpKind.RK }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovm2d_ymm_k)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x38, -1, EvexVectorLength.L256, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { EvexOpKind.VY, EvexOpKind.RK }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovm2d_zmm_k)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x38, -1, EvexVectorLength.L512, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { EvexOpKind.VZ, EvexOpKind.RK }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovm2q_xmm_k)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x38, -1, EvexVectorLength.L128, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.VX, EvexOpKind.RK }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovm2q_ymm_k)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x38, -1, EvexVectorLength.L256, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.VY, EvexOpKind.RK }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovm2q_zmm_k)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x38, -1, EvexVectorLength.L512, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.VZ, EvexOpKind.RK }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pminsd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x39, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpminsd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x39, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpminsd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x39, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpminsd_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x39, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpminsd_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x39, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpminsd_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x39, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpminsq_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x39, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpminsq_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x39, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpminsq_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x39, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovd2m_k_xmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x39, -1, EvexVectorLength.L128, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { EvexOpKind.VK, EvexOpKind.RX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovd2m_k_ymm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x39, -1, EvexVectorLength.L256, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { EvexOpKind.VK, EvexOpKind.RY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovd2m_k_zmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x39, -1, EvexVectorLength.L512, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { EvexOpKind.VK, EvexOpKind.RZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovq2m_k_xmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x39, -1, EvexVectorLength.L128, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.VK, EvexOpKind.RX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovq2m_k_ymm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x39, -1, EvexVectorLength.L256, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.VK, EvexOpKind.RY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmovq2m_k_zmm)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x39, -1, EvexVectorLength.L512, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.VK, EvexOpKind.RZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pminuw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3A, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpminuw_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3A, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpminuw_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3A, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpminuw_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3A, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpminuw_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3A, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpminuw_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3A, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpbroadcastmw2d_xmm_k)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x3A, -1, EvexVectorLength.L128, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { EvexOpKind.VX, EvexOpKind.RK }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpbroadcastmw2d_ymm_k)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x3A, -1, EvexVectorLength.L256, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { EvexOpKind.VY, EvexOpKind.RK }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpbroadcastmw2d_zmm_k)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x3A, -1, EvexVectorLength.L512, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { EvexOpKind.VZ, EvexOpKind.RK }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pminud_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3B, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpminud_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3B, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpminud_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3B, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpminud_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3B, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpminud_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3B, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpminud_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3B, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpminuq_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3B, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpminuq_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3B, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpminuq_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3B, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmaxsb_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3C, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmaxsb_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3C, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmaxsb_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3C, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmaxsb_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3C, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmaxsb_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3C, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmaxsb_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3C, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmaxsd_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3D, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmaxsd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3D, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmaxsd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3D, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmaxsd_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3D, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmaxsd_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3D, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmaxsd_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3D, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmaxsq_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3D, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmaxsq_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3D, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmaxsq_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3D, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmaxuw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3E, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmaxuw_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3E, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmaxuw_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3E, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmaxuw_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3E, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmaxuw_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3E, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmaxuw_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3E, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmaxud_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3F, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmaxud_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3F, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmaxud_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3F, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmaxud_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3F, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmaxud_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3F, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmaxud_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3F, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmaxuq_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3F, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmaxuq_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3F, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmaxuq_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x3F, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pmulld_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x40, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmulld_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x40, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmulld_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x40, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmulld_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x40, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmulld_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x40, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmulld_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x40, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmullq_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x40, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmullq_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x40, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmullq_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x40, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Phminposuw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x41, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vphminposuw_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x41, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgetexpps_xmm_k1z_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x42, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgetexpps_ymm_k1z_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x42, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgetexpps_zmm_k1z_zmmm512b32_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x42, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgetexppd_xmm_k1z_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x42, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgetexppd_ymm_k1z_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x42, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgetexppd_zmm_k1z_zmmm512b64_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x42, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgetexpss_xmm_k1z_xmm_xmmm32_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x43, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgetexpsd_xmm_k1z_xmm_xmmm64_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x43, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vplzcntd_xmm_k1z_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x44, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vplzcntd_ymm_k1z_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x44, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vplzcntd_zmm_k1z_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x44, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vplzcntq_xmm_k1z_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x44, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vplzcntq_ymm_k1z_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x44, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vplzcntq_zmm_k1z_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x44, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsrlvd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x45, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsrlvd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x45, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsrlvq_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x45, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsrlvq_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x45, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrlvd_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x45, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrlvd_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x45, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrlvd_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x45, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrlvq_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x45, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrlvq_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x45, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsrlvq_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x45, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsravd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x46, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsravd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x46, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsravd_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x46, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsravd_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x46, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsravd_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x46, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsravq_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x46, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsravq_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x46, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsravq_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x46, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsllvd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x47, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsllvd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x47, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsllvq_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x47, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpsllvq_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x47, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsllvd_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x47, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsllvd_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x47, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsllvd_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x47, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsllvq_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x47, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsllvq_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x47, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpsllvq_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x47, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrcp14ps_xmm_k1z_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x4C, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrcp14ps_ymm_k1z_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x4C, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrcp14ps_zmm_k1z_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x4C, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrcp14pd_xmm_k1z_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x4C, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrcp14pd_ymm_k1z_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x4C, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrcp14pd_zmm_k1z_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x4C, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrcp14ss_xmm_k1z_xmm_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x4D, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrcp14sd_xmm_k1z_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x4D, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrsqrt14ps_xmm_k1z_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x4E, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrsqrt14ps_ymm_k1z_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x4E, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrsqrt14ps_zmm_k1z_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x4E, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrsqrt14pd_xmm_k1z_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x4E, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrsqrt14pd_ymm_k1z_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x4E, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrsqrt14pd_zmm_k1z_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x4E, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrsqrt14ss_xmm_k1z_xmm_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x4F, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrsqrt14sd_xmm_k1z_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x4F, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpdpbusd_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x50, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpdpbusd_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x50, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpdpbusd_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x50, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpdpbusds_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x51, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpdpbusds_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x51, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpdpbusds_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x51, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpdpwssd_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x52, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpdpwssd_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x52, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpdpwssd_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x52, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vdpbf16ps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x52, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vdpbf16ps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x52, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vdpbf16ps_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x52, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vp4dpwssd_zmm_k1z_zmmp3_m128)], MandatoryPrefix.PF2, OpCodeTableKind.T0F38, 0x52, -1, EvexVectorLength.L512, TupleType.Tuple1_4X, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZP3, EvexOpKind.M }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpdpwssds_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x53, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpdpwssds_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x53, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpdpwssds_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x53, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vp4dpwssds_zmm_k1z_zmmp3_m128)], MandatoryPrefix.PF2, OpCodeTableKind.T0F38, 0x53, -1, EvexVectorLength.L512, TupleType.Tuple1_4X, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZP3, EvexOpKind.M }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpopcntb_xmm_k1z_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x54, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpopcntb_ymm_k1z_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x54, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpopcntb_zmm_k1z_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x54, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpopcntw_xmm_k1z_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x54, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpopcntw_ymm_k1z_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x54, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpopcntw_zmm_k1z_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x54, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpopcntd_xmm_k1z_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x55, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpopcntd_ymm_k1z_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x55, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpopcntd_zmm_k1z_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x55, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpopcntq_xmm_k1z_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x55, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpopcntq_ymm_k1z_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x55, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpopcntq_zmm_k1z_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x55, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpbroadcastd_xmm_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x58, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpbroadcastd_ymm_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x58, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpbroadcastd_xmm_k1z_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x58, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpbroadcastd_ymm_k1z_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x58, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpbroadcastd_zmm_k1z_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x58, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpbroadcastq_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x59, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpbroadcastq_ymm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x59, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vbroadcasti32x2_xmm_k1z_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x59, -1, EvexVectorLength.L128, TupleType.Tuple2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vbroadcasti32x2_ymm_k1z_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x59, -1, EvexVectorLength.L256, TupleType.Tuple2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vbroadcasti32x2_zmm_k1z_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x59, -1, EvexVectorLength.L512, TupleType.Tuple2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpbroadcastq_xmm_k1z_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x59, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpbroadcastq_ymm_k1z_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x59, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpbroadcastq_zmm_k1z_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x59, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vbroadcasti128_ymm_m128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x5A, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.M }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vbroadcasti32x4_ymm_k1z_m128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x5A, -1, EvexVectorLength.L256, TupleType.Tuple4, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.M }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vbroadcasti32x4_zmm_k1z_m128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x5A, -1, EvexVectorLength.L512, TupleType.Tuple4, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.M }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vbroadcasti64x2_ymm_k1z_m128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x5A, -1, EvexVectorLength.L256, TupleType.Tuple2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.M }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vbroadcasti64x2_zmm_k1z_m128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x5A, -1, EvexVectorLength.L512, TupleType.Tuple2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.M }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vbroadcasti32x8_zmm_k1z_m256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x5B, -1, EvexVectorLength.L512, TupleType.Tuple8, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.M }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vbroadcasti64x4_zmm_k1z_m256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x5B, -1, EvexVectorLength.L512, TupleType.Tuple4, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.M }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpexpandb_xmm_k1z_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x62, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar_1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpexpandb_ymm_k1z_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x62, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar_1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpexpandb_zmm_k1z_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x62, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar_1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpexpandw_xmm_k1z_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x62, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar_2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpexpandw_ymm_k1z_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x62, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar_2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpexpandw_zmm_k1z_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x62, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar_2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcompressb_xmmm128_k1z_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x63, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar_1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcompressb_ymmm256_k1z_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x63, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar_1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WY, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcompressb_zmmm512_k1z_zmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x63, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar_1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WZ, EvexOpKind.VZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcompressw_xmmm128_k1z_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x63, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar_2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcompressw_ymmm256_k1z_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x63, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar_2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WY, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcompressw_zmmm512_k1z_zmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x63, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar_2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WZ, EvexOpKind.VZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpblendmd_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x64, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpblendmd_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x64, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpblendmd_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x64, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpblendmq_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x64, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpblendmq_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x64, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpblendmq_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x64, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vblendmps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x65, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vblendmps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x65, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vblendmps_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x65, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vblendmpd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x65, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vblendmpd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x65, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vblendmpd_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x65, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpblendmb_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x66, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpblendmb_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x66, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpblendmb_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x66, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpblendmw_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x66, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpblendmw_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x66, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpblendmw_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x66, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vp2intersectd_kp1_xmm_xmmm128b32)], MandatoryPrefix.PF2, OpCodeTableKind.T0F38, 0x68, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast, new[] { EvexOpKind.VKP1, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vp2intersectd_kp1_ymm_ymmm256b32)], MandatoryPrefix.PF2, OpCodeTableKind.T0F38, 0x68, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast, new[] { EvexOpKind.VKP1, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vp2intersectd_kp1_zmm_zmmm512b32)], MandatoryPrefix.PF2, OpCodeTableKind.T0F38, 0x68, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast, new[] { EvexOpKind.VKP1, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vp2intersectq_kp1_xmm_xmmm128b64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F38, 0x68, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast, new[] { EvexOpKind.VKP1, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vp2intersectq_kp1_ymm_ymmm256b64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F38, 0x68, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast, new[] { EvexOpKind.VKP1, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vp2intersectq_kp1_zmm_zmmm512b64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F38, 0x68, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast, new[] { EvexOpKind.VKP1, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshldvw_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x70, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshldvw_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x70, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshldvw_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x70, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshldvd_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x71, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshldvd_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x71, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshldvd_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x71, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshldvq_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x71, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshldvq_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x71, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshldvq_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x71, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshrdvw_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x72, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshrdvw_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x72, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshrdvw_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x72, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtneps2bf16_xmm_k1z_xmmm128b32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x72, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtneps2bf16_xmm_k1z_ymmm256b32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x72, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtneps2bf16_ymm_k1z_zmmm512b32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0x72, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtne2ps2bf16_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.PF2, OpCodeTableKind.T0F38, 0x72, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtne2ps2bf16_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.PF2, OpCodeTableKind.T0F38, 0x72, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtne2ps2bf16_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.PF2, OpCodeTableKind.T0F38, 0x72, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshrdvd_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x73, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshrdvd_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x73, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshrdvd_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x73, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshrdvq_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x73, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshrdvq_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x73, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshrdvq_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x73, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermi2b_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x75, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermi2b_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x75, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermi2b_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x75, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermi2w_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x75, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermi2w_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x75, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermi2w_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x75, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermi2d_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x76, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermi2d_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x76, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermi2d_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x76, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermi2q_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x76, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermi2q_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x76, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermi2q_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x76, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermi2ps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x77, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermi2ps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x77, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermi2ps_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x77, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermi2pd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x77, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermi2pd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x77, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermi2pd_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x77, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpbroadcastb_xmm_xmmm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x78, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpbroadcastb_ymm_xmmm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x78, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpbroadcastb_xmm_k1z_xmmm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x78, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar_1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpbroadcastb_ymm_k1z_xmmm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x78, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar_1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpbroadcastb_zmm_k1z_xmmm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x78, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar_1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpbroadcastw_xmm_xmmm16)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x79, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpbroadcastw_ymm_xmmm16)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x79, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpbroadcastw_xmm_k1z_xmmm16)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x79, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar_2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpbroadcastw_ymm_k1z_xmmm16)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x79, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar_2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpbroadcastw_zmm_k1z_xmmm16)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x79, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar_2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpbroadcastb_xmm_k1z_r32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x7A, -1, EvexVectorLength.L128, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.Rd }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpbroadcastb_ymm_k1z_r32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x7A, -1, EvexVectorLength.L256, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.Rd }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpbroadcastb_zmm_k1z_r32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x7A, -1, EvexVectorLength.L512, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.Rd }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpbroadcastw_xmm_k1z_r32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x7B, -1, EvexVectorLength.L128, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.Rd }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpbroadcastw_ymm_k1z_r32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x7B, -1, EvexVectorLength.L256, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.Rd }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpbroadcastw_zmm_k1z_r32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x7B, -1, EvexVectorLength.L512, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.Rd }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpbroadcastd_xmm_k1z_r32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x7C, -1, EvexVectorLength.L128, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.Rd }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpbroadcastd_ymm_k1z_r32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x7C, -1, EvexVectorLength.L256, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.Rd }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpbroadcastd_zmm_k1z_r32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x7C, -1, EvexVectorLength.L512, TupleType.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.Rd }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpbroadcastq_xmm_k1z_r64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x7C, -1, EvexVectorLength.L128, TupleType.None, OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.Rq }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpbroadcastq_ymm_k1z_r64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x7C, -1, EvexVectorLength.L256, TupleType.None, OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.Rq }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpbroadcastq_zmm_k1z_r64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x7C, -1, EvexVectorLength.L512, TupleType.None, OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.Rq }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermt2b_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x7D, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermt2b_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x7D, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermt2b_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x7D, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermt2w_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x7D, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermt2w_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x7D, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermt2w_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x7D, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermt2d_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x7E, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermt2d_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x7E, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermt2d_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x7E, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermt2q_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x7E, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermt2q_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x7E, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermt2q_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x7E, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermt2ps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x7F, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermt2ps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x7F, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermt2ps_zmm_k1z_zmm_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x7F, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermt2pd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x7F, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermt2pd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x7F, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermt2pd_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x7F, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Invept_r32_m128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x80, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Gd, LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Invept_r64_m128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x80, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Invvpid_r32_m128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x81, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Gd, LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Invvpid_r64_m128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x81, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Invpcid_r32_m128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x82, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Gd, LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Invpcid_r64_m128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x82, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.M }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmultishiftqb_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x83, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmultishiftqb_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x83, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmultishiftqb_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x83, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vexpandps_xmm_k1z_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x88, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vexpandps_ymm_k1z_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x88, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vexpandps_zmm_k1z_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x88, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vexpandpd_xmm_k1z_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x88, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vexpandpd_ymm_k1z_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x88, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vexpandpd_zmm_k1z_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x88, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpexpandd_xmm_k1z_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x89, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpexpandd_ymm_k1z_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x89, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpexpandd_zmm_k1z_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x89, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpexpandq_xmm_k1z_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x89, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpexpandq_ymm_k1z_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x89, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpexpandq_zmm_k1z_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x89, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcompressps_xmmm128_k1z_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x8A, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcompressps_ymmm256_k1z_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x8A, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WY, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcompressps_zmmm512_k1z_zmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x8A, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WZ, EvexOpKind.VZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcompresspd_xmmm128_k1z_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x8A, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcompresspd_ymmm256_k1z_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x8A, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WY, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcompresspd_zmmm512_k1z_zmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x8A, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WZ, EvexOpKind.VZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcompressd_xmmm128_k1z_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x8B, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcompressd_ymmm256_k1z_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x8B, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WY, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcompressd_zmmm512_k1z_zmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x8B, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WZ, EvexOpKind.VZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcompressq_xmmm128_k1z_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x8B, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcompressq_ymmm256_k1z_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x8B, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WY, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcompressq_zmmm512_k1z_zmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x8B, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WZ, EvexOpKind.VZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmaskmovd_xmm_xmm_m128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x8C, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.M }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmaskmovd_ymm_ymm_m256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x8C, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.M }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmaskmovq_xmm_xmm_m128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x8C, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.M }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmaskmovq_ymm_ymm_m256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x8C, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.M }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermb_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x8D, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermb_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x8D, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermb_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x8D, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermw_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x8D, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermw_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x8D, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermw_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x8D, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmaskmovd_m128_xmm_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x8E, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.M, VexOpKind.HX, VexOpKind.VX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmaskmovd_m256_ymm_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x8E, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.M, VexOpKind.HY, VexOpKind.VY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmaskmovq_m128_xmm_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x8E, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.M, VexOpKind.HX, VexOpKind.VX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpmaskmovq_m256_ymm_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x8E, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.M, VexOpKind.HY, VexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshufbitqmb_k_k1_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x8F, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshufbitqmb_k_k1_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x8F, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshufbitqmb_k_k1_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x8F, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HZ, EvexOpKind.WZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpgatherdd_xmm_vm32x_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x90, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.VM32X, VexOpKind.HX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpgatherdd_ymm_vm32y_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x90, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.VM32Y, VexOpKind.HY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpgatherdq_xmm_vm32x_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x90, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.VM32X, VexOpKind.HX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpgatherdq_ymm_vm32x_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x90, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.VM32X, VexOpKind.HY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpgatherdd_xmm_k1_vm32x)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x90, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VX, EvexOpKind.VM32X }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpgatherdd_ymm_k1_vm32y)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x90, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VY, EvexOpKind.VM32Y }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpgatherdd_zmm_k1_vm32z)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x90, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VZ, EvexOpKind.VM32Z }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpgatherdq_xmm_k1_vm32x)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x90, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VX, EvexOpKind.VM32X }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpgatherdq_ymm_k1_vm32x)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x90, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VY, EvexOpKind.VM32X }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpgatherdq_zmm_k1_vm32y)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x90, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VZ, EvexOpKind.VM32Y }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpgatherqd_xmm_vm64x_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x91, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.VM64X, VexOpKind.HX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpgatherqd_xmm_vm64y_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x91, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.VM64Y, VexOpKind.HX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpgatherqq_xmm_vm64x_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x91, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.VM64X, VexOpKind.HX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpgatherqq_ymm_vm64y_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x91, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.VM64Y, VexOpKind.HY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpgatherqd_xmm_k1_vm64x)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x91, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VX, EvexOpKind.VM64X }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpgatherqd_xmm_k1_vm64y)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x91, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VX, EvexOpKind.VM64Y }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpgatherqd_ymm_k1_vm64z)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x91, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VY, EvexOpKind.VM64Z }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpgatherqq_xmm_k1_vm64x)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x91, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VX, EvexOpKind.VM64X }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpgatherqq_ymm_k1_vm64y)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x91, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VY, EvexOpKind.VM64Y }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpgatherqq_zmm_k1_vm64z)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x91, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VZ, EvexOpKind.VM64Z }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vgatherdps_xmm_vm32x_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x92, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.VM32X, VexOpKind.HX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vgatherdps_ymm_vm32y_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x92, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.VM32Y, VexOpKind.HY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vgatherdpd_xmm_vm32x_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x92, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.VM32X, VexOpKind.HX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vgatherdpd_ymm_vm32x_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x92, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.VM32X, VexOpKind.HY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgatherdps_xmm_k1_vm32x)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x92, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VX, EvexOpKind.VM32X }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgatherdps_ymm_k1_vm32y)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x92, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VY, EvexOpKind.VM32Y }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgatherdps_zmm_k1_vm32z)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x92, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VZ, EvexOpKind.VM32Z }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgatherdpd_xmm_k1_vm32x)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x92, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VX, EvexOpKind.VM32X }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgatherdpd_ymm_k1_vm32x)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x92, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VY, EvexOpKind.VM32X }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgatherdpd_zmm_k1_vm32y)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x92, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VZ, EvexOpKind.VM32Y }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vgatherqps_xmm_vm64x_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x93, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.VM64X, VexOpKind.HX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vgatherqps_xmm_vm64y_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x93, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.VM64Y, VexOpKind.HX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vgatherqpd_xmm_vm64x_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x93, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.VM64X, VexOpKind.HX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vgatherqpd_ymm_vm64y_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x93, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.VM64Y, VexOpKind.HY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgatherqps_xmm_k1_vm64x)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x93, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VX, EvexOpKind.VM64X }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgatherqps_xmm_k1_vm64y)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x93, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VX, EvexOpKind.VM64Y }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgatherqps_ymm_k1_vm64z)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x93, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VY, EvexOpKind.VM64Z }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgatherqpd_xmm_k1_vm64x)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x93, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VX, EvexOpKind.VM64X }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgatherqpd_ymm_k1_vm64y)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x93, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VY, EvexOpKind.VM64Y }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgatherqpd_zmm_k1_vm64z)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x93, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VZ, EvexOpKind.VM64Z }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmaddsub132ps_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x96, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmaddsub132ps_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x96, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmaddsub132pd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x96, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmaddsub132pd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x96, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmaddsub132ps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x96, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmaddsub132ps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x96, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmaddsub132ps_zmm_k1z_zmm_zmmm512b32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x96, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmaddsub132pd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x96, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmaddsub132pd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x96, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmaddsub132pd_zmm_k1z_zmm_zmmm512b64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x96, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsubadd132ps_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x97, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsubadd132ps_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x97, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsubadd132pd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x97, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsubadd132pd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x97, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsubadd132ps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x97, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsubadd132ps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x97, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsubadd132ps_zmm_k1z_zmm_zmmm512b32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x97, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsubadd132pd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x97, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsubadd132pd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x97, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsubadd132pd_zmm_k1z_zmm_zmmm512b64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x97, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmadd132ps_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x98, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmadd132ps_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x98, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmadd132pd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x98, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmadd132pd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x98, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmadd132ps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x98, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmadd132ps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x98, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmadd132ps_zmm_k1z_zmm_zmmm512b32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x98, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmadd132pd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x98, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmadd132pd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x98, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmadd132pd_zmm_k1z_zmm_zmmm512b64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x98, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmadd132ss_xmm_xmm_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x99, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmadd132sd_xmm_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x99, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmadd132ss_xmm_k1z_xmm_xmmm32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x99, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmadd132sd_xmm_k1z_xmm_xmmm64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x99, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsub132ps_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9A, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsub132ps_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9A, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsub132pd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9A, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsub132pd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9A, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsub132ps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9A, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsub132ps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9A, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsub132ps_zmm_k1z_zmm_zmmm512b32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9A, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsub132pd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9A, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsub132pd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9A, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsub132pd_zmm_k1z_zmm_zmmm512b64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9A, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_V4fmaddps_zmm_k1z_zmmp3_m128)], MandatoryPrefix.PF2, OpCodeTableKind.T0F38, 0x9A, -1, EvexVectorLength.L512, TupleType.Tuple1_4X, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZP3, EvexOpKind.M }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsub132ss_xmm_xmm_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9B, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsub132sd_xmm_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9B, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsub132ss_xmm_k1z_xmm_xmmm32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9B, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsub132sd_xmm_k1z_xmm_xmmm64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9B, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_V4fmaddss_xmm_k1z_xmmp3_m128)], MandatoryPrefix.PF2, OpCodeTableKind.T0F38, 0x9B, -1, EvexVectorLength.L128, TupleType.Tuple1_4X, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HXP3, EvexOpKind.M }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmadd132ps_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9C, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmadd132ps_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9C, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmadd132pd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9C, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmadd132pd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9C, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmadd132ps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9C, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmadd132ps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9C, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmadd132ps_zmm_k1z_zmm_zmmm512b32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9C, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmadd132pd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9C, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmadd132pd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9C, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmadd132pd_zmm_k1z_zmm_zmmm512b64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9C, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmadd132ss_xmm_xmm_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9D, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmadd132sd_xmm_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9D, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmadd132ss_xmm_k1z_xmm_xmmm32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9D, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmadd132sd_xmm_k1z_xmm_xmmm64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9D, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmsub132ps_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9E, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmsub132ps_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9E, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmsub132pd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9E, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmsub132pd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9E, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmsub132ps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9E, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmsub132ps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9E, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmsub132ps_zmm_k1z_zmm_zmmm512b32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9E, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmsub132pd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9E, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmsub132pd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9E, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmsub132pd_zmm_k1z_zmm_zmmm512b64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9E, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmsub132ss_xmm_xmm_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9F, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmsub132sd_xmm_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9F, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmsub132ss_xmm_k1z_xmm_xmmm32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9F, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmsub132sd_xmm_k1z_xmm_xmmm64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0x9F, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpscatterdd_vm32x_k1_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA0, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM32X, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpscatterdd_vm32y_k1_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA0, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM32Y, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpscatterdd_vm32z_k1_zmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA0, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM32Z, EvexOpKind.VZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpscatterdq_vm32x_k1_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA0, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM32X, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpscatterdq_vm32x_k1_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA0, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM32X, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpscatterdq_vm32y_k1_zmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA0, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM32Y, EvexOpKind.VZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpscatterqd_vm64x_k1_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA1, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM64X, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpscatterqd_vm64y_k1_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA1, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM64Y, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpscatterqd_vm64z_k1_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA1, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM64Z, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpscatterqq_vm64x_k1_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA1, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM64X, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpscatterqq_vm64y_k1_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA1, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM64Y, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpscatterqq_vm64z_k1_zmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA1, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM64Z, EvexOpKind.VZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vscatterdps_vm32x_k1_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA2, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM32X, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vscatterdps_vm32y_k1_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA2, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM32Y, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vscatterdps_vm32z_k1_zmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA2, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM32Z, EvexOpKind.VZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vscatterdpd_vm32x_k1_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA2, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM32X, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vscatterdpd_vm32x_k1_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA2, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM32X, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vscatterdpd_vm32y_k1_zmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA2, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM32Y, EvexOpKind.VZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vscatterqps_vm64x_k1_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA3, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM64X, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vscatterqps_vm64y_k1_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA3, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM64Y, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vscatterqps_vm64z_k1_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA3, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM64Z, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vscatterqpd_vm64x_k1_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA3, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM64X, EvexOpKind.VX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vscatterqpd_vm64y_k1_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA3, -1, EvexVectorLength.L256, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM64Y, EvexOpKind.VY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vscatterqpd_vm64z_k1_zmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA3, -1, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM64Z, EvexOpKind.VZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmaddsub213ps_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA6, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmaddsub213ps_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA6, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmaddsub213pd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA6, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmaddsub213pd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA6, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmaddsub213ps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA6, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmaddsub213ps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA6, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmaddsub213ps_zmm_k1z_zmm_zmmm512b32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA6, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmaddsub213pd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA6, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmaddsub213pd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA6, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmaddsub213pd_zmm_k1z_zmm_zmmm512b64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA6, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsubadd213ps_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA7, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsubadd213ps_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA7, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsubadd213pd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA7, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsubadd213pd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA7, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsubadd213ps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA7, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsubadd213ps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA7, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsubadd213ps_zmm_k1z_zmm_zmmm512b32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA7, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsubadd213pd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA7, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsubadd213pd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA7, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsubadd213pd_zmm_k1z_zmm_zmmm512b64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA7, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmadd213ps_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA8, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmadd213ps_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA8, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmadd213pd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA8, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmadd213pd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA8, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmadd213ps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA8, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmadd213ps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA8, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmadd213ps_zmm_k1z_zmm_zmmm512b32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA8, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmadd213pd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA8, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmadd213pd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA8, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmadd213pd_zmm_k1z_zmm_zmmm512b64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA8, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmadd213ss_xmm_xmm_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA9, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmadd213sd_xmm_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA9, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmadd213ss_xmm_k1z_xmm_xmmm32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA9, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmadd213sd_xmm_k1z_xmm_xmmm64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xA9, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsub213ps_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAA, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsub213ps_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAA, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsub213pd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAA, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsub213pd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAA, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsub213ps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAA, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsub213ps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAA, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsub213ps_zmm_k1z_zmm_zmmm512b32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAA, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsub213pd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAA, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsub213pd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAA, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsub213pd_zmm_k1z_zmm_zmmm512b64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAA, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_V4fnmaddps_zmm_k1z_zmmp3_m128)], MandatoryPrefix.PF2, OpCodeTableKind.T0F38, 0xAA, -1, EvexVectorLength.L512, TupleType.Tuple1_4X, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZP3, EvexOpKind.M }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsub213ss_xmm_xmm_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAB, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsub213sd_xmm_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAB, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsub213ss_xmm_k1z_xmm_xmmm32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAB, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsub213sd_xmm_k1z_xmm_xmmm64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAB, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_V4fnmaddss_xmm_k1z_xmmp3_m128)], MandatoryPrefix.PF2, OpCodeTableKind.T0F38, 0xAB, -1, EvexVectorLength.L128, TupleType.Tuple1_4X, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HXP3, EvexOpKind.M }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmadd213ps_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAC, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmadd213ps_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAC, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmadd213pd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAC, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmadd213pd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAC, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmadd213ps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAC, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmadd213ps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAC, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmadd213ps_zmm_k1z_zmm_zmmm512b32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAC, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmadd213pd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAC, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmadd213pd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAC, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmadd213pd_zmm_k1z_zmm_zmmm512b64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAC, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmadd213ss_xmm_xmm_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAD, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmadd213sd_xmm_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAD, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmadd213ss_xmm_k1z_xmm_xmmm32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAD, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmadd213sd_xmm_k1z_xmm_xmmm64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAD, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmsub213ps_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAE, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmsub213ps_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAE, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmsub213pd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAE, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmsub213pd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAE, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmsub213ps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAE, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmsub213ps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAE, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmsub213ps_zmm_k1z_zmm_zmmm512b32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAE, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmsub213pd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAE, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmsub213pd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAE, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmsub213pd_zmm_k1z_zmm_zmmm512b64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAE, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmsub213ss_xmm_xmm_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAF, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmsub213sd_xmm_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAF, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmsub213ss_xmm_k1z_xmm_xmmm32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAF, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmsub213sd_xmm_k1z_xmm_xmmm64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xAF, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmadd52luq_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB4, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmadd52luq_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB4, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmadd52luq_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB4, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmadd52huq_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB5, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmadd52huq_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB5, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpmadd52huq_zmm_k1z_zmm_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB5, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmaddsub231ps_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB6, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmaddsub231ps_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB6, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmaddsub231pd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB6, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmaddsub231pd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB6, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmaddsub231ps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB6, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmaddsub231ps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB6, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmaddsub231ps_zmm_k1z_zmm_zmmm512b32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB6, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmaddsub231pd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB6, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmaddsub231pd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB6, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmaddsub231pd_zmm_k1z_zmm_zmmm512b64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB6, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsubadd231ps_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB7, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsubadd231ps_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB7, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsubadd231pd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB7, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsubadd231pd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB7, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsubadd231ps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB7, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsubadd231ps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB7, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsubadd231ps_zmm_k1z_zmm_zmmm512b32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB7, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsubadd231pd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB7, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsubadd231pd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB7, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsubadd231pd_zmm_k1z_zmm_zmmm512b64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB7, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmadd231ps_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB8, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmadd231ps_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB8, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmadd231pd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB8, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmadd231pd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB8, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmadd231ps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB8, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmadd231ps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB8, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmadd231ps_zmm_k1z_zmm_zmmm512b32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB8, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmadd231pd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB8, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmadd231pd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB8, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmadd231pd_zmm_k1z_zmm_zmmm512b64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB8, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmadd231ss_xmm_xmm_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB9, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmadd231sd_xmm_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB9, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmadd231ss_xmm_k1z_xmm_xmmm32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB9, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmadd231sd_xmm_k1z_xmm_xmmm64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xB9, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsub231ps_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBA, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsub231ps_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBA, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsub231pd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBA, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsub231pd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBA, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsub231ps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBA, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsub231ps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBA, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsub231ps_zmm_k1z_zmm_zmmm512b32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBA, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsub231pd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBA, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsub231pd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBA, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsub231pd_zmm_k1z_zmm_zmmm512b64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBA, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsub231ss_xmm_xmm_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBB, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsub231sd_xmm_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBB, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsub231ss_xmm_k1z_xmm_xmmm32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBB, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfmsub231sd_xmm_k1z_xmm_xmmm64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBB, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmadd231ps_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBC, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmadd231ps_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBC, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmadd231pd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBC, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmadd231pd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBC, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmadd231ps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBC, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmadd231ps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBC, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmadd231ps_zmm_k1z_zmm_zmmm512b32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBC, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmadd231pd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBC, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmadd231pd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBC, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmadd231pd_zmm_k1z_zmm_zmmm512b64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBC, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmadd231ss_xmm_xmm_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBD, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmadd231sd_xmm_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBD, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmadd231ss_xmm_k1z_xmm_xmmm32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBD, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmadd231sd_xmm_k1z_xmm_xmmm64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBD, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmsub231ps_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBE, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmsub231ps_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBE, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmsub231pd_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBE, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmsub231pd_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBE, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmsub231ps_xmm_k1z_xmm_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBE, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmsub231ps_ymm_k1z_ymm_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBE, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmsub231ps_zmm_k1z_zmm_zmmm512b32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBE, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmsub231pd_xmm_k1z_xmm_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBE, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmsub231pd_ymm_k1z_ymm_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBE, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmsub231pd_zmm_k1z_zmm_zmmm512b64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBE, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmsub231ss_xmm_xmm_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBF, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmsub231sd_xmm_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBF, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmsub231ss_xmm_k1z_xmm_xmmm32_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBF, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfnmsub231sd_xmm_k1z_xmm_xmmm64_er)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xBF, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.RoundingControl | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpconflictd_xmm_k1z_xmmm128b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xC4, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpconflictd_ymm_k1z_ymmm256b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xC4, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpconflictd_zmm_k1z_zmmm512b32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xC4, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpconflictq_xmm_k1z_xmmm128b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xC4, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpconflictq_ymm_k1z_ymmm256b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xC4, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpconflictq_zmm_k1z_zmmm512b64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xC4, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgatherpf0dps_vm32z_k1)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xC6, 1, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM32Z }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgatherpf0dpd_vm32y_k1)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xC6, 1, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM32Y }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgatherpf1dps_vm32z_k1)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xC6, 2, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM32Z }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgatherpf1dpd_vm32y_k1)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xC6, 2, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM32Y }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vscatterpf0dps_vm32z_k1)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xC6, 5, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM32Z }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vscatterpf0dpd_vm32y_k1)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xC6, 5, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM32Y }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vscatterpf1dps_vm32z_k1)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xC6, 6, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM32Z }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vscatterpf1dpd_vm32y_k1)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xC6, 6, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM32Y }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgatherpf0qps_vm64z_k1)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xC7, 1, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM64Z }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgatherpf0qpd_vm64z_k1)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xC7, 1, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM64Z }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgatherpf1qps_vm64z_k1)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xC7, 2, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM64Z }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgatherpf1qpd_vm64z_k1)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xC7, 2, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM64Z }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vscatterpf0qps_vm64z_k1)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xC7, 5, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM64Z }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vscatterpf0qpd_vm64z_k1)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xC7, 5, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM64Z }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vscatterpf1qps_vm64z_k1)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xC7, 6, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM64Z }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vscatterpf1qpd_vm64z_k1)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xC7, 6, EvexVectorLength.L512, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.NonZeroOpMaskRegister, new[] { EvexOpKind.VM64Z }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sha1nexte_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0xC8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vexp2ps_zmm_k1z_zmmm512b32_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xC8, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vexp2pd_zmm_k1z_zmmm512b64_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xC8, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sha1msg1_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0xC9, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sha1msg2_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0xCA, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrcp28ps_zmm_k1z_zmmm512b32_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xCA, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrcp28pd_zmm_k1z_zmmm512b64_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xCA, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sha256rnds2_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0xCB, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrcp28ss_xmm_k1z_xmm_xmmm32_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xCB, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrcp28sd_xmm_k1z_xmm_xmmm64_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xCB, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sha256msg1_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0xCC, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrsqrt28ps_zmm_k1z_zmmm512b32_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xCC, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrsqrt28pd_zmm_k1z_zmmm512b64_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xCC, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sha256msg2_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0xCD, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrsqrt28ss_xmm_k1z_xmm_xmmm32_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xCD, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrsqrt28sd_xmm_k1z_xmm_xmmm64_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xCD, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Gf2p8mulb_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xCF, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vgf2p8mulb_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xCF, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vgf2p8mulb_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xCF, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgf2p8mulb_xmm_k1z_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xCF, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgf2p8mulb_ymm_k1z_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xCF, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgf2p8mulb_zmm_k1z_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xCF, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Aesimc_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xDB, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vaesimc_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xDB, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Aesenc_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xDC, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vaesenc_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xDC, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vaesenc_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xDC, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vaesenc_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xDC, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vaesenc_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xDC, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vaesenc_zmm_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xDC, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Aesenclast_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xDD, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vaesenclast_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xDD, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vaesenclast_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xDD, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vaesenclast_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xDD, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vaesenclast_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xDD, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vaesenclast_zmm_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xDD, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Aesdec_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xDE, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vaesdec_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xDE, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vaesdec_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xDE, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vaesdec_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xDE, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vaesdec_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xDE, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vaesdec_zmm_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xDE, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Aesdeclast_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xDF, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vaesdeclast_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xDF, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vaesdeclast_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xDF, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vaesdeclast_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xDF, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vaesdeclast_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xDF, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vaesdeclast_zmm_zmm_zmmm512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xDF, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movbe_r16_m16)], MandatoryPrefix.None, OpCodeTableKind.T0F38, 0xF0, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gw, LegacyOpKind.Mw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movbe_r32_m32)], MandatoryPrefix.None, OpCodeTableKind.T0F38, 0xF0, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Md }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movbe_r64_m64)], MandatoryPrefix.None, OpCodeTableKind.T0F38, 0xF0, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Mq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Crc32_r32_rm8)], MandatoryPrefix.PF2, OpCodeTableKind.T0F38, 0xF0, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Crc32_r64_rm8)], MandatoryPrefix.PF2, OpCodeTableKind.T0F38, 0xF0, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eb }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movbe_m16_r16)], MandatoryPrefix.None, OpCodeTableKind.T0F38, 0xF1, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Mw, LegacyOpKind.Gw }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movbe_m32_r32)], MandatoryPrefix.None, OpCodeTableKind.T0F38, 0xF1, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Md, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movbe_m64_r64)], MandatoryPrefix.None, OpCodeTableKind.T0F38, 0xF1, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Mq, LegacyOpKind.Gq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Crc32_r32_rm16)], MandatoryPrefix.PF2, OpCodeTableKind.T0F38, 0xF1, -1, OperandSize.Size16, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ew }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Crc32_r32_rm32)], MandatoryPrefix.PF2, OpCodeTableKind.T0F38, 0xF1, -1, OperandSize.Size32, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Crc32_r64_rm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F38, 0xF1, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Andn_r32_r32_rm32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0xF2, -1, VexVectorLength.LZ, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { VexOpKind.Gd, VexOpKind.Hd, VexOpKind.Ed }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Andn_r64_r64_rm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0xF2, -1, VexVectorLength.LZ, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.Gq, VexOpKind.Hq, VexOpKind.Eq }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Blsr_r32_rm32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0xF3, 1, VexVectorLength.LZ, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { VexOpKind.Hd, VexOpKind.Ed }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Blsr_r64_rm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0xF3, 1, VexVectorLength.LZ, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.Hq, VexOpKind.Eq }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Blsmsk_r32_rm32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0xF3, 2, VexVectorLength.LZ, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { VexOpKind.Hd, VexOpKind.Ed }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Blsmsk_r64_rm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0xF3, 2, VexVectorLength.LZ, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.Hq, VexOpKind.Eq }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Blsi_r32_rm32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0xF3, 3, VexVectorLength.LZ, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { VexOpKind.Hd, VexOpKind.Ed }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Blsi_r64_rm64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0xF3, 3, VexVectorLength.LZ, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.Hq, VexOpKind.Eq }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Bzhi_r32_rm32_r32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0xF5, -1, VexVectorLength.LZ, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { VexOpKind.Gd, VexOpKind.Ed, VexOpKind.Hd }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Bzhi_r64_rm64_r64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0xF5, -1, VexVectorLength.LZ, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.Gq, VexOpKind.Eq, VexOpKind.Hq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Wrussd_m32_r32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xF5, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Md, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Wrussq_m64_r64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xF5, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Mq, LegacyOpKind.Gq }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Pext_r32_r32_rm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0xF5, -1, VexVectorLength.LZ, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { VexOpKind.Gd, VexOpKind.Hd, VexOpKind.Ed }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Pext_r64_r64_rm64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0xF5, -1, VexVectorLength.LZ, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.Gq, VexOpKind.Hq, VexOpKind.Eq }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Pdep_r32_r32_rm32)], MandatoryPrefix.PF2, OpCodeTableKind.T0F38, 0xF5, -1, VexVectorLength.LZ, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { VexOpKind.Gd, VexOpKind.Hd, VexOpKind.Ed }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Pdep_r64_r64_rm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F38, 0xF5, -1, VexVectorLength.LZ, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.Gq, VexOpKind.Hq, VexOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Wrssd_m32_r32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0xF6, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Md, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Wrssq_m64_r64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0xF6, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Mq, LegacyOpKind.Gq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Adcx_r32_rm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xF6, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Adcx_r64_rm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xF6, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Adox_r32_rm32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0xF6, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd, LegacyOpKind.Ed }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Adox_r64_rm64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0xF6, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq, LegacyOpKind.Eq }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Mulx_r32_r32_rm32)], MandatoryPrefix.PF2, OpCodeTableKind.T0F38, 0xF6, -1, VexVectorLength.LZ, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { VexOpKind.Gd, VexOpKind.Hd, VexOpKind.Ed }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Mulx_r64_r64_rm64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F38, 0xF6, -1, VexVectorLength.LZ, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.Gq, VexOpKind.Hq, VexOpKind.Eq }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Bextr_r32_rm32_r32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0xF7, -1, VexVectorLength.LZ, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { VexOpKind.Gd, VexOpKind.Ed, VexOpKind.Hd }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Bextr_r64_rm64_r64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0xF7, -1, VexVectorLength.LZ, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.Gq, VexOpKind.Eq, VexOpKind.Hq }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Shlx_r32_rm32_r32)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xF7, -1, VexVectorLength.LZ, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { VexOpKind.Gd, VexOpKind.Ed, VexOpKind.Hd }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Shlx_r64_rm64_r64)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xF7, -1, VexVectorLength.LZ, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.Gq, VexOpKind.Eq, VexOpKind.Hq }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Sarx_r32_rm32_r32)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0xF7, -1, VexVectorLength.LZ, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { VexOpKind.Gd, VexOpKind.Ed, VexOpKind.Hd }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Sarx_r64_rm64_r64)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0xF7, -1, VexVectorLength.LZ, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.Gq, VexOpKind.Eq, VexOpKind.Hq }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Shrx_r32_rm32_r32)], MandatoryPrefix.PF2, OpCodeTableKind.T0F38, 0xF7, -1, VexVectorLength.LZ, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { VexOpKind.Gd, VexOpKind.Ed, VexOpKind.Hd }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Shrx_r64_rm64_r64)], MandatoryPrefix.PF2, OpCodeTableKind.T0F38, 0xF7, -1, VexVectorLength.LZ, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.Gq, VexOpKind.Eq, VexOpKind.Hq }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movdir64b_r16_m512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xF8, -1, OperandSize.None, AddressSize.Size16, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Gw_mem, LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movdir64b_r32_m512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xF8, -1, OperandSize.None, AddressSize.Size32, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd_mem, LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movdir64b_r64_m512)], MandatoryPrefix.P66, OpCodeTableKind.T0F38, 0xF8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq_mem, LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Enqcmds_r16_m512)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0xF8, -1, OperandSize.None, AddressSize.Size16, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Gw_mem, LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Enqcmds_r32_m512)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0xF8, -1, OperandSize.None, AddressSize.Size32, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd_mem, LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Enqcmds_r64_m512)], MandatoryPrefix.PF3, OpCodeTableKind.T0F38, 0xF8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq_mem, LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Enqcmd_r16_m512)], MandatoryPrefix.PF2, OpCodeTableKind.T0F38, 0xF8, -1, OperandSize.None, AddressSize.Size16, OpCodeFlags.Mode16 | OpCodeFlags.Mode32, new[] { LegacyOpKind.Gw_mem, LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Enqcmd_r32_m512)], MandatoryPrefix.PF2, OpCodeTableKind.T0F38, 0xF8, -1, OperandSize.None, AddressSize.Size32, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Gd_mem, LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Enqcmd_r64_m512)], MandatoryPrefix.PF2, OpCodeTableKind.T0F38, 0xF8, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Gq_mem, LegacyOpKind.M }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movdiri_m32_r32)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0xF9, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Md, LegacyOpKind.Gd }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Movdiri_m64_r64)], MandatoryPrefix.PNP, OpCodeTableKind.T0F38, 0xF9, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Mq, LegacyOpKind.Gq }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpermq_ymm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x00, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.WY, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermq_ymm_k1z_ymmm256b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x00, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermq_zmm_k1z_zmmm512b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x00, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpermpd_ymm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x01, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.WY, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermpd_ymm_k1z_ymmm256b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x01, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermpd_zmm_k1z_zmmm512b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x01, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpblendd_xmm_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x02, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpblendd_ymm_ymm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x02, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Valignd_xmm_k1z_xmm_xmmm128b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x03, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Valignd_ymm_k1z_ymm_ymmm256b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x03, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Valignd_zmm_k1z_zmm_zmmm512b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x03, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Valignq_xmm_k1z_xmm_xmmm128b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x03, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Valignq_ymm_k1z_ymm_ymmm256b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x03, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Valignq_zmm_k1z_zmm_zmmm512b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x03, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpermilps_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x04, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.WX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpermilps_ymm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x04, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.WY, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermilps_xmm_k1z_xmmm128b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x04, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermilps_ymm_k1z_ymmm256b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x04, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermilps_zmm_k1z_zmmm512b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x04, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpermilpd_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x05, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.WX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpermilpd_ymm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x05, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.WY, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermilpd_xmm_k1z_xmmm128b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x05, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermilpd_ymm_k1z_ymmm256b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x05, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpermilpd_zmm_k1z_zmmm512b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x05, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vperm2f128_ymm_ymm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x06, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY, VexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Roundps_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x08, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vroundps_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x08, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vroundps_ymm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x08, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WY, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrndscaleps_xmm_k1z_xmmm128b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x08, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrndscaleps_ymm_k1z_ymmm256b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x08, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrndscaleps_zmm_k1z_zmmm512b32_imm8_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x08, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Roundpd_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x09, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vroundpd_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x09, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vroundpd_ymm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x09, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.WY, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrndscalepd_xmm_k1z_xmmm128b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x09, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrndscalepd_ymm_k1z_ymmm256b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x09, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrndscalepd_zmm_k1z_zmmm512b64_imm8_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x09, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Roundss_xmm_xmmm32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x0A, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vroundss_xmm_xmm_xmmm32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x0A, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrndscaless_xmm_k1z_xmm_xmmm32_imm8_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x0A, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Roundsd_xmm_xmmm64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x0B, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vroundsd_xmm_xmm_xmmm64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x0B, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrndscalesd_xmm_k1z_xmm_xmmm64_imm8_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x0B, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Blendps_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x0C, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vblendps_xmm_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x0C, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vblendps_ymm_ymm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x0C, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY, VexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Blendpd_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x0D, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vblendpd_xmm_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x0D, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vblendpd_ymm_ymm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x0D, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY, VexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pblendw_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x0E, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpblendw_xmm_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x0E, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpblendw_ymm_ymm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x0E, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY, VexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Palignr_mm_mmm64_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.T0F3A, 0x0F, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.P, LegacyOpKind.Q, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Palignr_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x0F, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpalignr_xmm_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x0F, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpalignr_ymm_ymm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x0F, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpalignr_xmm_k1z_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x0F, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpalignr_ymm_k1z_ymm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x0F, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpalignr_zmm_k1z_zmm_zmmm512_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x0F, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pextrb_r32m8_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x14, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.RdMb, LegacyOpKind.VX, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pextrb_r64m8_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x14, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.RqMb, LegacyOpKind.VX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpextrb_r32m8_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x14, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { VexOpKind.RdMb, VexOpKind.VX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpextrb_r64m8_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x14, -1, VexVectorLength.L128, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.RqMb, VexOpKind.VX, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpextrb_r32m8_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x14, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar_1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { EvexOpKind.RdMb, EvexOpKind.VX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpextrb_r64m8_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x14, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar_1, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.RqMb, EvexOpKind.VX, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pextrw_r32m16_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x15, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.RdMw, LegacyOpKind.VX, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pextrw_r64m16_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x15, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.RqMw, LegacyOpKind.VX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpextrw_r32m16_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x15, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { VexOpKind.RdMw, VexOpKind.VX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpextrw_r64m16_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x15, -1, VexVectorLength.L128, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.RqMw, VexOpKind.VX, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpextrw_r32m16_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x15, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar_2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { EvexOpKind.RdMw, EvexOpKind.VX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpextrw_r64m16_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x15, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar_2, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.RqMw, EvexOpKind.VX, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pextrd_rm32_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x16, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.VX, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pextrq_rm64_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x16, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.VX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpextrd_rm32_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x16, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { VexOpKind.Ed, VexOpKind.VX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpextrq_rm64_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x16, -1, VexVectorLength.L128, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.Eq, VexOpKind.VX, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpextrd_rm32_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x16, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar_4, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { EvexOpKind.Ed, EvexOpKind.VX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpextrq_rm64_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x16, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar_8, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.Eq, EvexOpKind.VX, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Extractps_rm32_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x17, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.Ed, LegacyOpKind.VX, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Extractps_r64m32_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x17, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.Eq, LegacyOpKind.VX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vextractps_rm32_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x17, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { VexOpKind.Ed, VexOpKind.VX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vextractps_r64m32_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x17, -1, VexVectorLength.L128, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.Eq, VexOpKind.VX, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vextractps_rm32_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x17, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar_4, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { EvexOpKind.Ed, EvexOpKind.VX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vextractps_r64m32_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x17, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar_4, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.Eq, EvexOpKind.VX, EvexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vinsertf128_ymm_ymm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x18, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WX, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vinsertf32x4_ymm_k1z_ymm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x18, -1, EvexVectorLength.L256, TupleType.Tuple4, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vinsertf32x4_zmm_k1z_zmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x18, -1, EvexVectorLength.L512, TupleType.Tuple4, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vinsertf64x2_ymm_k1z_ymm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x18, -1, EvexVectorLength.L256, TupleType.Tuple2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vinsertf64x2_zmm_k1z_zmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x18, -1, EvexVectorLength.L512, TupleType.Tuple2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WX, EvexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vextractf128_xmmm128_ymm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x19, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.WX, VexOpKind.VY, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vextractf32x4_xmmm128_k1z_ymm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x19, -1, EvexVectorLength.L256, TupleType.Tuple4, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vextractf32x4_xmmm128_k1z_zmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x19, -1, EvexVectorLength.L512, TupleType.Tuple4, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vextractf64x2_xmmm128_k1z_ymm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x19, -1, EvexVectorLength.L256, TupleType.Tuple2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vextractf64x2_xmmm128_k1z_zmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x19, -1, EvexVectorLength.L512, TupleType.Tuple2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vinsertf32x8_zmm_k1z_zmm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x1A, -1, EvexVectorLength.L512, TupleType.Tuple8, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vinsertf64x4_zmm_k1z_zmm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x1A, -1, EvexVectorLength.L512, TupleType.Tuple4, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vextractf32x8_ymmm256_k1z_zmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x1B, -1, EvexVectorLength.L512, TupleType.Tuple8, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WY, EvexOpKind.VZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vextractf64x4_ymmm256_k1z_zmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x1B, -1, EvexVectorLength.L512, TupleType.Tuple4, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WY, EvexOpKind.VZ, EvexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvtps2ph_xmmm64_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x1D, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.WX, VexOpKind.VX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vcvtps2ph_xmmm128_ymm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x1D, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.WX, VexOpKind.VY, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtps2ph_xmmm64_k1z_xmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x1D, -1, EvexVectorLength.L128, TupleType.Half_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtps2ph_xmmm128_k1z_ymm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x1D, -1, EvexVectorLength.L256, TupleType.Half_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vcvtps2ph_ymmm256_k1z_zmm_imm8_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x1D, -1, EvexVectorLength.L512, TupleType.Half_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WY, EvexOpKind.VZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpud_k_k1_xmm_xmmm128b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x1E, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpud_k_k1_ymm_ymmm256b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x1E, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpud_k_k1_zmm_zmmm512b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x1E, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpuq_k_k1_xmm_xmmm128b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x1E, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpuq_k_k1_ymm_ymmm256b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x1E, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpuq_k_k1_zmm_zmmm512b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x1E, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpd_k_k1_xmm_xmmm128b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x1F, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpd_k_k1_ymm_ymmm256b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x1F, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpd_k_k1_zmm_zmmm512b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x1F, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpq_k_k1_xmm_xmmm128b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x1F, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpq_k_k1_ymm_ymmm256b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x1F, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpq_k_k1_zmm_zmmm512b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x1F, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pinsrb_xmm_r32m8_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x20, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.RdMb, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pinsrb_xmm_r64m8_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x20, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.RqMb, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpinsrb_xmm_xmm_r32m8_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x20, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.RdMb, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpinsrb_xmm_xmm_r64m8_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x20, -1, VexVectorLength.L128, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.RqMb, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpinsrb_xmm_xmm_r32m8_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x20, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar_1, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.RdMb, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpinsrb_xmm_xmm_r64m8_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x20, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar_1, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.RqMb, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Insertps_xmm_xmmm32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x21, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vinsertps_xmm_xmm_xmmm32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x21, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vinsertps_xmm_xmm_xmmm32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x21, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pinsrd_xmm_rm32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x22, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.Ed, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pinsrq_xmm_rm64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x22, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.Eq, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpinsrd_xmm_xmm_rm32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x22, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.Ed, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpinsrq_xmm_xmm_rm64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x22, -1, VexVectorLength.L128, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.Eq, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpinsrd_xmm_xmm_rm32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x22, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar_4, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.Ed, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpinsrq_xmm_xmm_rm64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x22, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar_8, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.Eq, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vshuff32x4_ymm_k1z_ymm_ymmm256b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x23, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vshuff32x4_zmm_k1z_zmm_zmmm512b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x23, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vshuff64x2_ymm_k1z_ymm_ymmm256b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x23, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vshuff64x2_zmm_k1z_zmm_zmmm512b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x23, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpternlogd_xmm_k1z_xmm_xmmm128b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x25, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpternlogd_ymm_k1z_ymm_ymmm256b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x25, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpternlogd_zmm_k1z_zmm_zmmm512b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x25, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpternlogq_xmm_k1z_xmm_xmmm128b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x25, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpternlogq_ymm_k1z_ymm_ymmm256b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x25, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpternlogq_zmm_k1z_zmm_zmmm512b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x25, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgetmantps_xmm_k1z_xmmm128b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x26, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgetmantps_ymm_k1z_ymmm256b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x26, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgetmantps_zmm_k1z_zmmm512b32_imm8_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x26, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgetmantpd_xmm_k1z_xmmm128b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x26, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgetmantpd_ymm_k1z_ymmm256b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x26, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgetmantpd_zmm_k1z_zmmm512b64_imm8_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x26, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgetmantss_xmm_k1z_xmm_xmmm32_imm8_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x27, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgetmantsd_xmm_k1z_xmm_xmmm64_imm8_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x27, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kshiftrb_k_k_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x30, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VK, VexOpKind.RK, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kshiftrw_k_k_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x30, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VK, VexOpKind.RK, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kshiftrd_k_k_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x31, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VK, VexOpKind.RK, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kshiftrq_k_k_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x31, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VK, VexOpKind.RK, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kshiftlb_k_k_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x32, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VK, VexOpKind.RK, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kshiftlw_k_k_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x32, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VK, VexOpKind.RK, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kshiftld_k_k_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x33, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VK, VexOpKind.RK, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Kshiftlq_k_k_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x33, -1, VexVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VK, VexOpKind.RK, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vinserti128_ymm_ymm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x38, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WX, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vinserti32x4_ymm_k1z_ymm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x38, -1, EvexVectorLength.L256, TupleType.Tuple4, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vinserti32x4_zmm_k1z_zmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x38, -1, EvexVectorLength.L512, TupleType.Tuple4, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vinserti64x2_ymm_k1z_ymm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x38, -1, EvexVectorLength.L256, TupleType.Tuple2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vinserti64x2_zmm_k1z_zmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x38, -1, EvexVectorLength.L512, TupleType.Tuple2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WX, EvexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vextracti128_xmmm128_ymm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x39, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.WX, VexOpKind.VY, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vextracti32x4_xmmm128_k1z_ymm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x39, -1, EvexVectorLength.L256, TupleType.Tuple4, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vextracti32x4_xmmm128_k1z_zmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x39, -1, EvexVectorLength.L512, TupleType.Tuple4, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vextracti64x2_xmmm128_k1z_ymm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x39, -1, EvexVectorLength.L256, TupleType.Tuple2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vextracti64x2_xmmm128_k1z_zmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x39, -1, EvexVectorLength.L512, TupleType.Tuple2, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WX, EvexOpKind.VZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vinserti32x8_zmm_k1z_zmm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x3A, -1, EvexVectorLength.L512, TupleType.Tuple8, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vinserti64x4_zmm_k1z_zmm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x3A, -1, EvexVectorLength.L512, TupleType.Tuple4, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vextracti32x8_ymmm256_k1z_zmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x3B, -1, EvexVectorLength.L512, TupleType.Tuple8, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WY, EvexOpKind.VZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vextracti64x4_ymmm256_k1z_zmm_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x3B, -1, EvexVectorLength.L512, TupleType.Tuple4, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.WY, EvexOpKind.VZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpub_k_k1_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x3E, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpub_k_k1_ymm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x3E, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpub_k_k1_zmm_zmmm512_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x3E, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpuw_k_k1_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x3E, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpuw_k_k1_ymm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x3E, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpuw_k_k1_zmm_zmmm512_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x3E, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpb_k_k1_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x3F, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpb_k_k1_ymm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x3F, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpb_k_k1_zmm_zmmm512_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x3F, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpw_k_k1_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x3F, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpw_k_k1_ymm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x3F, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpcmpw_k_k1_zmm_zmmm512_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x3F, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Dpps_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x40, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vdpps_xmm_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x40, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vdpps_ymm_ymm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x40, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY, VexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Dppd_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x41, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vdppd_xmm_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x41, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Mpsadbw_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x42, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmpsadbw_xmm_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x42, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vmpsadbw_ymm_ymm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x42, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vdbpsadbw_xmm_k1z_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x42, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vdbpsadbw_ymm_k1z_ymm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x42, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vdbpsadbw_zmm_k1z_zmm_zmmm512_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x42, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vshufi32x4_ymm_k1z_ymm_ymmm256b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x43, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vshufi32x4_zmm_k1z_zmm_zmmm512b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x43, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vshufi64x2_ymm_k1z_ymm_ymmm256b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x43, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vshufi64x2_zmm_k1z_zmm_zmmm512b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x43, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pclmulqdq_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x44, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpclmulqdq_xmm_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x44, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpclmulqdq_ymm_ymm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x44, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpclmulqdq_xmm_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x44, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpclmulqdq_ymm_ymm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x44, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpclmulqdq_zmm_zmm_zmmm512_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x44, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vperm2i128_ymm_ymm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x46, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpermil2ps_xmm_xmm_xmmm128_xmm_imm2)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x48, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Is5X, VexOpKind.I2 }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpermil2ps_ymm_ymm_ymmm256_ymm_imm2)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x48, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY, VexOpKind.Is5Y, VexOpKind.I2 }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpermil2ps_xmm_xmm_xmm_xmmm128_imm2)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x48, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.Is5X, VexOpKind.WX, VexOpKind.I2 }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpermil2ps_ymm_ymm_ymm_ymmm256_imm2)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x48, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.Is5Y, VexOpKind.WY, VexOpKind.I2 }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpermil2pd_xmm_xmm_xmmm128_xmm_imm2)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x49, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Is5X, VexOpKind.I2 }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpermil2pd_ymm_ymm_ymmm256_ymm_imm2)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x49, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY, VexOpKind.Is5Y, VexOpKind.I2 }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpermil2pd_xmm_xmm_xmm_xmmm128_imm2)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x49, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.Is5X, VexOpKind.WX, VexOpKind.I2 }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpermil2pd_ymm_ymm_ymm_ymmm256_imm2)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x49, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.Is5Y, VexOpKind.WY, VexOpKind.I2 }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vblendvps_xmm_xmm_xmmm128_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x4A, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Is4X }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vblendvps_ymm_ymm_ymmm256_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x4A, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY, VexOpKind.Is4Y }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vblendvpd_xmm_xmm_xmmm128_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x4B, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Is4X }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vblendvpd_ymm_ymm_ymmm256_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x4B, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY, VexOpKind.Is4Y }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpblendvb_xmm_xmm_xmmm128_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x4C, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Is4X }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpblendvb_ymm_ymm_ymmm256_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x4C, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY, VexOpKind.Is4Y }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrangeps_xmm_k1z_xmm_xmmm128b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x50, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrangeps_ymm_k1z_ymm_ymmm256b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x50, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrangeps_zmm_k1z_zmm_zmmm512b32_imm8_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x50, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrangepd_xmm_k1z_xmm_xmmm128b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x50, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrangepd_ymm_k1z_ymm_ymmm256b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x50, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrangepd_zmm_k1z_zmm_zmmm512b64_imm8_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x50, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrangess_xmm_k1z_xmm_xmmm32_imm8_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x51, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vrangesd_xmm_k1z_xmm_xmmm64_imm8_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x51, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfixupimmps_xmm_k1z_xmm_xmmm128b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x54, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfixupimmps_ymm_k1z_ymm_ymmm256b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x54, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfixupimmps_zmm_k1z_zmm_zmmm512b32_imm8_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x54, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfixupimmpd_xmm_k1z_xmm_xmmm128b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x54, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfixupimmpd_ymm_k1z_ymm_ymmm256b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x54, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfixupimmpd_zmm_k1z_zmm_zmmm512b64_imm8_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x54, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfixupimmss_xmm_k1z_xmm_xmmm32_imm8_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x55, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfixupimmsd_xmm_k1z_xmm_xmmm64_imm8_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x55, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vreduceps_xmm_k1z_xmmm128b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x56, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vreduceps_ymm_k1z_ymmm256b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x56, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vreduceps_zmm_k1z_zmmm512b32_imm8_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x56, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vreducepd_xmm_k1z_xmmm128b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x56, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vreducepd_ymm_k1z_ymmm256b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x56, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vreducepd_zmm_k1z_zmmm512b64_imm8_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x56, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vreducess_xmm_k1z_xmm_xmmm32_imm8_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x57, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vreducesd_xmm_k1z_xmm_xmmm64_imm8_sae)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x57, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.SuppressAllExceptions | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmaddsubps_xmm_xmm_xmmm128_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x5C, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Is4X }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmaddsubps_ymm_ymm_ymmm256_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x5C, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY, VexOpKind.Is4Y }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmaddsubps_xmm_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x5C, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.Is4X, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmaddsubps_ymm_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x5C, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.Is4Y, VexOpKind.WY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmaddsubpd_xmm_xmm_xmmm128_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x5D, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Is4X }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmaddsubpd_ymm_ymm_ymmm256_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x5D, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY, VexOpKind.Is4Y }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmaddsubpd_xmm_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x5D, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.Is4X, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmaddsubpd_ymm_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x5D, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.Is4Y, VexOpKind.WY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsubaddps_xmm_xmm_xmmm128_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x5E, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Is4X }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsubaddps_ymm_ymm_ymmm256_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x5E, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY, VexOpKind.Is4Y }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsubaddps_xmm_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x5E, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.Is4X, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsubaddps_ymm_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x5E, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.Is4Y, VexOpKind.WY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsubaddpd_xmm_xmm_xmmm128_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x5F, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Is4X }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsubaddpd_ymm_ymm_ymmm256_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x5F, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY, VexOpKind.Is4Y }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsubaddpd_xmm_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x5F, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.Is4X, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsubaddpd_ymm_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x5F, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.Is4Y, VexOpKind.WY }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pcmpestrm_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x60, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pcmpestrm64_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x60, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpcmpestrm_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x60, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { VexOpKind.VX, VexOpKind.WX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpcmpestrm64_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x60, -1, VexVectorLength.L128, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.WX, VexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pcmpestri_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x61, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pcmpestri64_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x61, -1, OperandSize.Size64, AddressSize.None, OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpcmpestri_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x61, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { VexOpKind.VX, VexOpKind.WX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpcmpestri64_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x61, -1, VexVectorLength.L128, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.WX, VexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pcmpistrm_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x62, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpcmpistrm_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x62, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX, VexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Pcmpistri_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x63, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vpcmpistri_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x63, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfpclassps_k_k1_xmmm128b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x66, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfpclassps_k_k1_ymmm256b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x66, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfpclassps_k_k1_zmmm512b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x66, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.WZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfpclasspd_k_k1_xmmm128b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x66, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfpclasspd_k_k1_ymmm256b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x66, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfpclasspd_k_k1_zmmm512b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x66, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.WZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfpclassss_k_k1_xmmm32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x67, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vfpclasssd_k_k1_xmmm64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x67, -1, EvexVectorLength.L128, TupleType.Tuple1_Scalar, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W | OpCodeFlags.OpMaskRegister, new[] { EvexOpKind.VK, EvexOpKind.WX, EvexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmaddps_xmm_xmm_xmmm128_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x68, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Is4X }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmaddps_ymm_ymm_ymmm256_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x68, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY, VexOpKind.Is4Y }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmaddps_xmm_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x68, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.Is4X, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmaddps_ymm_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x68, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.Is4Y, VexOpKind.WY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmaddpd_xmm_xmm_xmmm128_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x69, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Is4X }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmaddpd_ymm_ymm_ymmm256_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x69, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY, VexOpKind.Is4Y }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmaddpd_xmm_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x69, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.Is4X, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmaddpd_ymm_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x69, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.Is4Y, VexOpKind.WY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmaddss_xmm_xmm_xmmm32_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x6A, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Is4X }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmaddss_xmm_xmm_xmm_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x6A, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.Is4X, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmaddsd_xmm_xmm_xmmm64_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x6B, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Is4X }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmaddsd_xmm_xmm_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x6B, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.Is4X, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsubps_xmm_xmm_xmmm128_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x6C, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Is4X }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsubps_ymm_ymm_ymmm256_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x6C, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY, VexOpKind.Is4Y }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsubps_xmm_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x6C, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.Is4X, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsubps_ymm_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x6C, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.Is4Y, VexOpKind.WY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsubpd_xmm_xmm_xmmm128_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x6D, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Is4X }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsubpd_ymm_ymm_ymmm256_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x6D, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY, VexOpKind.Is4Y }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsubpd_xmm_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x6D, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.Is4X, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsubpd_ymm_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x6D, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.Is4Y, VexOpKind.WY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsubss_xmm_xmm_xmmm32_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x6E, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Is4X }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsubss_xmm_xmm_xmm_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x6E, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.Is4X, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsubsd_xmm_xmm_xmmm64_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x6F, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Is4X }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfmsubsd_xmm_xmm_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x6F, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.Is4X, VexOpKind.WX }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshldw_xmm_k1z_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x70, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshldw_ymm_k1z_ymm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x70, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshldw_zmm_k1z_zmm_zmmm512_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x70, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshldd_xmm_k1z_xmm_xmmm128b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x71, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshldd_ymm_k1z_ymm_ymmm256b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x71, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshldd_zmm_k1z_zmm_zmmm512b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x71, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshldq_xmm_k1z_xmm_xmmm128b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x71, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshldq_ymm_k1z_ymm_ymmm256b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x71, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshldq_zmm_k1z_zmm_zmmm512b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x71, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshrdw_xmm_k1z_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x72, -1, EvexVectorLength.L128, TupleType.Full_Mem_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshrdw_ymm_k1z_ymm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x72, -1, EvexVectorLength.L256, TupleType.Full_Mem_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshrdw_zmm_k1z_zmm_zmmm512_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x72, -1, EvexVectorLength.L512, TupleType.Full_Mem_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshrdd_xmm_k1z_xmm_xmmm128b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x73, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshrdd_ymm_k1z_ymm_ymmm256b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x73, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshrdd_zmm_k1z_zmm_zmmm512b32_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x73, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshrdq_xmm_k1z_xmm_xmmm128b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x73, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshrdq_ymm_k1z_ymm_ymmm256b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x73, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vpshrdq_zmm_k1z_zmm_zmmm512b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x73, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmaddps_xmm_xmm_xmmm128_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x78, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Is4X }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmaddps_ymm_ymm_ymmm256_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x78, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY, VexOpKind.Is4Y }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmaddps_xmm_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x78, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.Is4X, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmaddps_ymm_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x78, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.Is4Y, VexOpKind.WY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmaddpd_xmm_xmm_xmmm128_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x79, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Is4X }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmaddpd_ymm_ymm_ymmm256_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x79, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY, VexOpKind.Is4Y }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmaddpd_xmm_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x79, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.Is4X, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmaddpd_ymm_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x79, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.Is4Y, VexOpKind.WY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmaddss_xmm_xmm_xmmm32_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x7A, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Is4X }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmaddss_xmm_xmm_xmm_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x7A, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.Is4X, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmaddsd_xmm_xmm_xmmm64_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x7B, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Is4X }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmaddsd_xmm_xmm_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x7B, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.Is4X, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmsubps_xmm_xmm_xmmm128_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x7C, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Is4X }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmsubps_ymm_ymm_ymmm256_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x7C, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY, VexOpKind.Is4Y }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmsubps_xmm_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x7C, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.Is4X, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmsubps_ymm_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x7C, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.Is4Y, VexOpKind.WY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmsubpd_xmm_xmm_xmmm128_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x7D, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Is4X }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmsubpd_ymm_ymm_ymmm256_ymm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x7D, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY, VexOpKind.Is4Y }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmsubpd_xmm_xmm_xmm_xmmm128)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x7D, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.Is4X, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmsubpd_ymm_ymm_ymm_ymmm256)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x7D, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.Is4Y, VexOpKind.WY }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmsubss_xmm_xmm_xmmm32_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x7E, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Is4X }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmsubss_xmm_xmm_xmm_xmmm32)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x7E, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.Is4X, VexOpKind.WX }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmsubsd_xmm_xmm_xmmm64_xmm)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x7F, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Is4X }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vfnmsubsd_xmm_xmm_xmm_xmmm64)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0x7F, -1, VexVectorLength.LIG, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.LIG | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.Is4X, VexOpKind.WX }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Sha1rnds4_xmm_xmmm128_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.T0F3A, 0xCC, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX, LegacyOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Gf2p8affineqb_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0xCE, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vgf2p8affineqb_xmm_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0xCE, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vgf2p8affineqb_ymm_ymm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0xCE, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgf2p8affineqb_xmm_k1z_xmm_xmmm128b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0xCE, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgf2p8affineqb_ymm_k1z_ymm_ymmm256b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0xCE, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgf2p8affineqb_zmm_k1z_zmm_zmmm512b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0xCE, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Gf2p8affineinvqb_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0xCF, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vgf2p8affineinvqb_xmm_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0xCF, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VX, VexOpKind.HX, VexOpKind.WX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vgf2p8affineinvqb_ymm_ymm_ymmm256_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0xCF, -1, VexVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.VY, VexOpKind.HY, VexOpKind.WY, VexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgf2p8affineinvqb_xmm_k1z_xmm_xmmm128b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0xCF, -1, EvexVectorLength.L128, TupleType.Full_128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VX, EvexOpKind.HX, EvexOpKind.WX, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgf2p8affineinvqb_ymm_k1z_ymm_ymmm256b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0xCF, -1, EvexVectorLength.L256, TupleType.Full_256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VY, EvexOpKind.HY, EvexOpKind.WY, EvexOpKind.Ib }),
				new EvexOpCodeInfo(CodeEnum.Instance[nameof(Code.EVEX_Vgf2p8affineinvqb_zmm_k1z_zmm_zmmm512b64_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0xCF, -1, EvexVectorLength.L512, TupleType.Full_512, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W | OpCodeFlags.Broadcast | OpCodeFlags.OpMaskRegister | OpCodeFlags.ZeroingMasking, new[] { EvexOpKind.VZ, EvexOpKind.HZ, EvexOpKind.WZ, EvexOpKind.Ib }),
				new LegacyOpCodeInfo(CodeEnum.Instance[nameof(Code.Aeskeygenassist_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0xDF, -1, OperandSize.None, AddressSize.None, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { LegacyOpKind.VX, LegacyOpKind.WX, LegacyOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Vaeskeygenassist_xmm_xmmm128_imm8)], MandatoryPrefix.P66, OpCodeTableKind.T0F3A, 0xDF, -1, VexVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG, new[] { VexOpKind.VX, VexOpKind.WX, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Rorx_r32_rm32_imm8)], MandatoryPrefix.PF2, OpCodeTableKind.T0F3A, 0xF0, -1, VexVectorLength.LZ, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { VexOpKind.Gd, VexOpKind.Ed, VexOpKind.Ib }),
				new VexOpCodeInfo(CodeEnum.Instance[nameof(Code.VEX_Rorx_r64_rm64_imm8)], MandatoryPrefix.PF2, OpCodeTableKind.T0F3A, 0xF0, -1, VexVectorLength.LZ, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { VexOpKind.Gq, VexOpKind.Eq, VexOpKind.Ib }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpmacssww_xmm_xmm_xmmm128_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.XOP8, 0x85, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX, XopOpKind.Is4X }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpmacsswd_xmm_xmm_xmmm128_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.XOP8, 0x86, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX, XopOpKind.Is4X }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpmacssdql_xmm_xmm_xmmm128_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.XOP8, 0x87, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX, XopOpKind.Is4X }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpmacssdd_xmm_xmm_xmmm128_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.XOP8, 0x8E, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX, XopOpKind.Is4X }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpmacssdqh_xmm_xmm_xmmm128_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.XOP8, 0x8F, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX, XopOpKind.Is4X }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpmacsww_xmm_xmm_xmmm128_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.XOP8, 0x95, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX, XopOpKind.Is4X }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpmacswd_xmm_xmm_xmmm128_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.XOP8, 0x96, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX, XopOpKind.Is4X }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpmacsdql_xmm_xmm_xmmm128_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.XOP8, 0x97, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX, XopOpKind.Is4X }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpmacsdd_xmm_xmm_xmmm128_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.XOP8, 0x9E, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX, XopOpKind.Is4X }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpmacsdqh_xmm_xmm_xmmm128_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.XOP8, 0x9F, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX, XopOpKind.Is4X }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpcmov_xmm_xmm_xmmm128_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.XOP8, 0xA2, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX, XopOpKind.Is4X }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpcmov_ymm_ymm_ymmm256_ymm)], MandatoryPrefix.PNP, OpCodeTableKind.XOP8, 0xA2, -1, XopVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VY, XopOpKind.HY, XopOpKind.WY, XopOpKind.Is4Y }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpcmov_xmm_xmm_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.XOP8, 0xA2, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.Is4X, XopOpKind.WX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpcmov_ymm_ymm_ymm_ymmm256)], MandatoryPrefix.PNP, OpCodeTableKind.XOP8, 0xA2, -1, XopVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { XopOpKind.VY, XopOpKind.HY, XopOpKind.Is4Y, XopOpKind.WY }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpperm_xmm_xmm_xmmm128_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.XOP8, 0xA3, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX, XopOpKind.Is4X }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpperm_xmm_xmm_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.XOP8, 0xA3, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.Is4X, XopOpKind.WX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpmadcsswd_xmm_xmm_xmmm128_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.XOP8, 0xA6, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX, XopOpKind.Is4X }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpmadcswd_xmm_xmm_xmmm128_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.XOP8, 0xB6, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX, XopOpKind.Is4X }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vprotb_xmm_xmmm128_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.XOP8, 0xC0, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX, XopOpKind.Ib }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vprotw_xmm_xmmm128_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.XOP8, 0xC1, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX, XopOpKind.Ib }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vprotd_xmm_xmmm128_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.XOP8, 0xC2, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX, XopOpKind.Ib }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vprotq_xmm_xmmm128_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.XOP8, 0xC3, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX, XopOpKind.Ib }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpcomb_xmm_xmm_xmmm128_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.XOP8, 0xCC, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX, XopOpKind.Ib }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpcomw_xmm_xmm_xmmm128_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.XOP8, 0xCD, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX, XopOpKind.Ib }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpcomd_xmm_xmm_xmmm128_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.XOP8, 0xCE, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX, XopOpKind.Ib }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpcomq_xmm_xmm_xmmm128_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.XOP8, 0xCF, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX, XopOpKind.Ib }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpcomub_xmm_xmm_xmmm128_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.XOP8, 0xEC, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX, XopOpKind.Ib }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpcomuw_xmm_xmm_xmmm128_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.XOP8, 0xED, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX, XopOpKind.Ib }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpcomud_xmm_xmm_xmmm128_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.XOP8, 0xEE, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX, XopOpKind.Ib }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpcomuq_xmm_xmm_xmmm128_imm8)], MandatoryPrefix.PNP, OpCodeTableKind.XOP8, 0xEF, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX, XopOpKind.Ib }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Blcfill_r32_rm32)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x01, 1, XopVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { XopOpKind.Hd, XopOpKind.Ed }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Blcfill_r64_rm64)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x01, 1, XopVectorLength.L0, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { XopOpKind.Hq, XopOpKind.Eq }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Blsfill_r32_rm32)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x01, 2, XopVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { XopOpKind.Hd, XopOpKind.Ed }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Blsfill_r64_rm64)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x01, 2, XopVectorLength.L0, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { XopOpKind.Hq, XopOpKind.Eq }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Blcs_r32_rm32)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x01, 3, XopVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { XopOpKind.Hd, XopOpKind.Ed }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Blcs_r64_rm64)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x01, 3, XopVectorLength.L0, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { XopOpKind.Hq, XopOpKind.Eq }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Tzmsk_r32_rm32)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x01, 4, XopVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { XopOpKind.Hd, XopOpKind.Ed }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Tzmsk_r64_rm64)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x01, 4, XopVectorLength.L0, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { XopOpKind.Hq, XopOpKind.Eq }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Blcic_r32_rm32)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x01, 5, XopVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { XopOpKind.Hd, XopOpKind.Ed }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Blcic_r64_rm64)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x01, 5, XopVectorLength.L0, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { XopOpKind.Hq, XopOpKind.Eq }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Blsic_r32_rm32)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x01, 6, XopVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { XopOpKind.Hd, XopOpKind.Ed }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Blsic_r64_rm64)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x01, 6, XopVectorLength.L0, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { XopOpKind.Hq, XopOpKind.Eq }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_T1mskc_r32_rm32)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x01, 7, XopVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { XopOpKind.Hd, XopOpKind.Ed }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_T1mskc_r64_rm64)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x01, 7, XopVectorLength.L0, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { XopOpKind.Hq, XopOpKind.Eq }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Blcmsk_r32_rm32)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x02, 1, XopVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { XopOpKind.Hd, XopOpKind.Ed }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Blcmsk_r64_rm64)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x02, 1, XopVectorLength.L0, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { XopOpKind.Hq, XopOpKind.Eq }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Blci_r32_rm32)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x02, 6, XopVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { XopOpKind.Hd, XopOpKind.Ed }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Blci_r64_rm64)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x02, 6, XopVectorLength.L0, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { XopOpKind.Hq, XopOpKind.Eq }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Llwpcb_r32)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x12, 0, XopVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { XopOpKind.Rd }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Llwpcb_r64)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x12, 0, XopVectorLength.L0, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { XopOpKind.Rq }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Slwpcb_r32)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x12, 1, XopVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { XopOpKind.Rd }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Slwpcb_r64)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x12, 1, XopVectorLength.L0, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { XopOpKind.Rq }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vfrczps_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x80, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vfrczps_ymm_ymmm256)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x80, -1, XopVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VY, XopOpKind.WY }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vfrczpd_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x81, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vfrczpd_ymm_ymmm256)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x81, -1, XopVectorLength.L256, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VY, XopOpKind.WY }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vfrczss_xmm_xmmm32)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x82, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vfrczsd_xmm_xmmm64)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x83, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vprotb_xmm_xmmm128_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x90, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX, XopOpKind.HX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vprotb_xmm_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x90, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vprotw_xmm_xmmm128_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x91, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX, XopOpKind.HX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vprotw_xmm_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x91, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vprotd_xmm_xmmm128_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x92, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX, XopOpKind.HX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vprotd_xmm_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x92, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vprotq_xmm_xmmm128_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x93, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX, XopOpKind.HX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vprotq_xmm_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x93, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpshlb_xmm_xmmm128_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x94, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX, XopOpKind.HX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpshlb_xmm_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x94, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpshlw_xmm_xmmm128_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x95, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX, XopOpKind.HX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpshlw_xmm_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x95, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpshld_xmm_xmmm128_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x96, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX, XopOpKind.HX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpshld_xmm_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x96, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpshlq_xmm_xmmm128_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x97, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX, XopOpKind.HX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpshlq_xmm_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x97, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpshab_xmm_xmmm128_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x98, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX, XopOpKind.HX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpshab_xmm_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x98, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpshaw_xmm_xmmm128_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x99, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX, XopOpKind.HX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpshaw_xmm_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x99, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpshad_xmm_xmmm128_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x9A, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX, XopOpKind.HX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpshad_xmm_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x9A, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpshaq_xmm_xmmm128_xmm)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x9B, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX, XopOpKind.HX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vpshaq_xmm_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0x9B, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { XopOpKind.VX, XopOpKind.HX, XopOpKind.WX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vphaddbw_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0xC1, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vphaddbd_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0xC2, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vphaddbq_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0xC3, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vphaddwd_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0xC6, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vphaddwq_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0xC7, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vphadddq_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0xCB, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vphaddubw_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0xD1, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vphaddubd_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0xD2, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vphaddubq_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0xD3, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vphadduwd_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0xD6, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vphadduwq_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0xD7, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vphaddudq_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0xDB, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vphsubbw_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0xE1, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vphsubwd_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0xE2, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Vphsubdq_xmm_xmmm128)], MandatoryPrefix.PNP, OpCodeTableKind.XOP9, 0xE3, -1, XopVectorLength.L128, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64, new[] { XopOpKind.VX, XopOpKind.WX }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Bextr_r32_rm32_imm32)], MandatoryPrefix.PNP, OpCodeTableKind.XOPA, 0x10, -1, XopVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { XopOpKind.Gd, XopOpKind.Ed, XopOpKind.Id }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Bextr_r64_rm64_imm32)], MandatoryPrefix.PNP, OpCodeTableKind.XOPA, 0x10, -1, XopVectorLength.L0, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { XopOpKind.Gq, XopOpKind.Eq, XopOpKind.Id }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Lwpins_r32_rm32_imm32)], MandatoryPrefix.PNP, OpCodeTableKind.XOPA, 0x12, 0, XopVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { XopOpKind.Hd, XopOpKind.Ed, XopOpKind.Id }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Lwpins_r64_rm32_imm32)], MandatoryPrefix.PNP, OpCodeTableKind.XOPA, 0x12, 0, XopVectorLength.L0, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { XopOpKind.Hq, XopOpKind.Ed, XopOpKind.Id }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Lwpval_r32_rm32_imm32)], MandatoryPrefix.PNP, OpCodeTableKind.XOPA, 0x12, 1, XopVectorLength.L0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64 | OpCodeFlags.WIG32, new[] { XopOpKind.Hd, XopOpKind.Ed, XopOpKind.Id }),
				new XopOpCodeInfo(CodeEnum.Instance[nameof(Code.XOP_Lwpval_r64_rm32_imm32)], MandatoryPrefix.PNP, OpCodeTableKind.XOPA, 0x12, 1, XopVectorLength.L0, OpCodeFlags.Mode64 | OpCodeFlags.W, new[] { XopOpKind.Hq, XopOpKind.Ed, XopOpKind.Id }),
				new D3nowOpCodeInfo(CodeEnum.Instance[nameof(Code.D3NOW_Pi2fw_mm_mmm64)], 0x0C, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64),
				new D3nowOpCodeInfo(CodeEnum.Instance[nameof(Code.D3NOW_Pi2fd_mm_mmm64)], 0x0D, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64),
				new D3nowOpCodeInfo(CodeEnum.Instance[nameof(Code.D3NOW_Pf2iw_mm_mmm64)], 0x1C, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64),
				new D3nowOpCodeInfo(CodeEnum.Instance[nameof(Code.D3NOW_Pf2id_mm_mmm64)], 0x1D, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64),
				new D3nowOpCodeInfo(CodeEnum.Instance[nameof(Code.D3NOW_Pfrcpv_mm_mmm64)], 0x86, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64),
				new D3nowOpCodeInfo(CodeEnum.Instance[nameof(Code.D3NOW_Pfrsqrtv_mm_mmm64)], 0x87, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64),
				new D3nowOpCodeInfo(CodeEnum.Instance[nameof(Code.D3NOW_Pfnacc_mm_mmm64)], 0x8A, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64),
				new D3nowOpCodeInfo(CodeEnum.Instance[nameof(Code.D3NOW_Pfpnacc_mm_mmm64)], 0x8E, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64),
				new D3nowOpCodeInfo(CodeEnum.Instance[nameof(Code.D3NOW_Pfcmpge_mm_mmm64)], 0x90, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64),
				new D3nowOpCodeInfo(CodeEnum.Instance[nameof(Code.D3NOW_Pfmin_mm_mmm64)], 0x94, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64),
				new D3nowOpCodeInfo(CodeEnum.Instance[nameof(Code.D3NOW_Pfrcp_mm_mmm64)], 0x96, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64),
				new D3nowOpCodeInfo(CodeEnum.Instance[nameof(Code.D3NOW_Pfrsqrt_mm_mmm64)], 0x97, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64),
				new D3nowOpCodeInfo(CodeEnum.Instance[nameof(Code.D3NOW_Pfsub_mm_mmm64)], 0x9A, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64),
				new D3nowOpCodeInfo(CodeEnum.Instance[nameof(Code.D3NOW_Pfadd_mm_mmm64)], 0x9E, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64),
				new D3nowOpCodeInfo(CodeEnum.Instance[nameof(Code.D3NOW_Pfcmpgt_mm_mmm64)], 0xA0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64),
				new D3nowOpCodeInfo(CodeEnum.Instance[nameof(Code.D3NOW_Pfmax_mm_mmm64)], 0xA4, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64),
				new D3nowOpCodeInfo(CodeEnum.Instance[nameof(Code.D3NOW_Pfrcpit1_mm_mmm64)], 0xA6, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64),
				new D3nowOpCodeInfo(CodeEnum.Instance[nameof(Code.D3NOW_Pfrsqit1_mm_mmm64)], 0xA7, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64),
				new D3nowOpCodeInfo(CodeEnum.Instance[nameof(Code.D3NOW_Pfsubr_mm_mmm64)], 0xAA, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64),
				new D3nowOpCodeInfo(CodeEnum.Instance[nameof(Code.D3NOW_Pfacc_mm_mmm64)], 0xAE, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64),
				new D3nowOpCodeInfo(CodeEnum.Instance[nameof(Code.D3NOW_Pfcmpeq_mm_mmm64)], 0xB0, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64),
				new D3nowOpCodeInfo(CodeEnum.Instance[nameof(Code.D3NOW_Pfmul_mm_mmm64)], 0xB4, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64),
				new D3nowOpCodeInfo(CodeEnum.Instance[nameof(Code.D3NOW_Pfrcpit2_mm_mmm64)], 0xB6, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64),
				new D3nowOpCodeInfo(CodeEnum.Instance[nameof(Code.D3NOW_Pmulhrw_mm_mmm64)], 0xB7, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64),
				new D3nowOpCodeInfo(CodeEnum.Instance[nameof(Code.D3NOW_Pswapd_mm_mmm64)], 0xBB, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64),
				new D3nowOpCodeInfo(CodeEnum.Instance[nameof(Code.D3NOW_Pavgusb_mm_mmm64)], 0xBF, OpCodeFlags.Mode16 | OpCodeFlags.Mode32 | OpCodeFlags.Mode64),
			};
			if (data.Length != CodeEnum.Instance.Values.Length)
				throw new InvalidOperationException();
			if (data.Length != new HashSet<EnumValue>(data.Select(a => a.Code)).Count)
				throw new InvalidOperationException();
			return data.OrderBy(a => a.Code.Value).ToArray();
		}
	}
}
